/*
 * Part of the ROBOID project - http://hamster.school
 * Copyright (C) 2016 Kwang-Hyun Park (akaii@kw.ac.kr)
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
*/

#include <process.h>
#include <stdio.h>
#include <string.h>
#include <windows.h>
#include <signal.h>
#include <sys/timeb.h>
#include "roboid.h"

#ifdef _MSC_VER
#define _STRCPY(dst, sz, src) strcpy_s((dst), (sz), (src))
#define _STRCAT(dst, sz, src) strcat_s((dst), (sz), (src))
#define _STRNCPY(dst, sz, src, n) strncpy_s((dst), (sz), (src), (n))
#define _STRICMP(s1, s2) _stricmp((s1), (s2))
#else
#define _STRCPY(dst, sz, src) strcpy((dst), (src))
#define _STRCAT(dst, sz, src) strcat((dst), (src))
#define _STRNCPY(dst, sz, src, n) strncpy((dst), (src), (n))
#define _STRICMP(s1, s2) stricmp((s1), (s2))
#endif

#ifdef _WIN64
#define _LONG long long
#else
#define _LONG long
#endif

/*------------------------------
  SERIAL
------------------------------*/

#define _SERIAL_BAUDRATE_110 110
#define _SERIAL_BAUDRATE_300 300
#define _SERIAL_BAUDRATE_600 600
#define _SERIAL_BAUDRATE_1200 1200
#define _SERIAL_BAUDRATE_4800 4800
#define _SERIAL_BAUDRATE_9600 9600
#define _SERIAL_BAUDRATE_14400 14400
#define _SERIAL_BAUDRATE_19200 19200
#define _SERIAL_BAUDRATE_38400 38400
#define _SERIAL_BAUDRATE_57600 57600
#define _SERIAL_BAUDRATE_115200 115200
#define _SERIAL_BAUDRATE_128000 128000
#define _SERIAL_BAUDRATE_256000 256000

#define _SERIAL_DATABITS_5 5
#define _SERIAL_DATABITS_6 6
#define _SERIAL_DATABITS_7 7
#define _SERIAL_DATABITS_8 8

#define _SERIAL_STOPBITS_1 1
#define _SERIAL_STOPBITS_2 2
#define _SERIAL_STOPBITS_1_5 3

#define _SERIAL_PARITY_NONE 0
#define _SERIAL_PARITY_ODD 1
#define _SERIAL_PARITY_EVEN 2
#define _SERIAL_PARITY_MARK 3
#define _SERIAL_PARITY_SPACE 4

#define _SERIAL_PURGE_RXABORT 0x0002
#define _SERIAL_PURGE_RXCLEAR 0x0008
#define _SERIAL_PURGE_TXABORT 0x0001
#define _SERIAL_PURGE_TXCLEAR 0x0004

#define _SERIAL_MASK_RXCHAR 1
#define _SERIAL_MASK_RXFLAG 2
#define _SERIAL_MASK_TXEMPTY 4
#define _SERIAL_MASK_CTS 8
#define _SERIAL_MASK_DSR 16
#define _SERIAL_MASK_RLSD 32
#define _SERIAL_MASK_BREAK 64
#define _SERIAL_MASK_ERR 128
#define _SERIAL_MASK_RING 256

#define _SERIAL_FLOWCONTROL_NONE 0
#define _SERIAL_FLOWCONTROL_RTSCTS_IN 1
#define _SERIAL_FLOWCONTROL_RTSCTS_OUT 2
#define _SERIAL_FLOWCONTROL_XONXOFF_IN 4
#define _SERIAL_FLOWCONTROL_XONXOFF_OUT 8

#define _SERIAL_ERROR_FRAME 0x0008
#define _SERIAL_ERROR_OVERRUN 0x0002
#define _SERIAL_ERROR_PARITY 0x0004

#define _SERIAL_ERROR_PORT_BUSY -1
#define _SERIAL_ERROR_PORT_NOT_FOUND -2
#define _SERIAL_ERROR_PERMISSION_DENIED -3
#define _SERIAL_ERROR_INCORRECT_SERIAL_PORT -4

#define _SERIAL_PARAMS_FLAG_IGNPAR 1
#define _SERIAL_PARAMS_FLAG_PARMRK 2

#define _TEMP_CHAR_BUFFER_SIZE 256
#define _SERIAL_BUFFER_SIZE 32768

struct _serial {
	_LONG port_handle;
	int port_opened;
	char* buffer;
	int buffer_size;
	int offset;
};

char** _serial_window_get_serial_port_names(const char* keyword, int* count);
_LONG _serial_window_open_port(const char* port_name);
int _serial_window_close_port(_LONG port_handle);
int _serial_window_set_params(_LONG port_handle, int baud_rate, int byte_size, int stop_bits, int parity, int set_rts, int set_dtr, int flags);
int _serial_window_set_flow_control_mode(_LONG port_handle, int mask);
int _serial_window_purge_port(_LONG port_handle, int flags);
int _serial_window_count_read_bytes(_LONG port_handle);
int _serial_window_read_bytes(_LONG port_handle, unsigned char* buffer, int buffer_size);
int _serial_window_write_bytes(_LONG port_handle, const unsigned char* buffer, int buffer_size);

struct _serial* _serial_create(void);
void _serial_dispose(struct _serial* serial);
int _serial_open(struct _serial* serial, const char* port_name, int baud_rate, int flow_control);
void _serial_close(struct _serial* serial);
void _serial_clear(struct _serial* serial);
int _serial_read_string_until(struct _serial* serial, char* buffer, int buffer_size, char delimiter);
int _serial_write(const struct _serial* serial, const char* buffer, int buffer_size);

char** _serial_window_get_serial_port_names(const char* keyword, int* count) {
	HKEY result;
	LPCSTR sub_key = "HARDWARE\\DEVICEMAP\\SERIALCOMM\\";
	char** names = NULL;
	int names_count = 0;
	
	if(RegOpenKeyExA(HKEY_LOCAL_MACHINE, sub_key, 0, KEY_READ, &result) == ERROR_SUCCESS) {
		int has_more_elements = 1;
		char value_name[_TEMP_CHAR_BUFFER_SIZE];
		DWORD keys_count = 0, value_name_size, enum_result;
		
		while(has_more_elements == 1) {
			value_name_size = _TEMP_CHAR_BUFFER_SIZE;
			enum_result = RegEnumValueA(result, keys_count, value_name, &value_name_size, NULL, NULL, NULL, NULL);
			if(enum_result == ERROR_SUCCESS) {
				++ keys_count;
				if(keyword == NULL || strstr(value_name, keyword) != NULL) {
					++ names_count;
				}
			} else if(enum_result == ERROR_NO_MORE_ITEMS) {
				has_more_elements = 0;
			} else {
				has_more_elements = 0;
			}
		}
		if(names_count > 0) {
			byte data[_TEMP_CHAR_BUFFER_SIZE];
			DWORD data_size;
			int names_index = 0, i;
			DWORD j;
			
			names = (char**)malloc(sizeof(char*) * names_count);
			for(i = 0; i < names_count; ++i) {
				names[i] = (char*)malloc(sizeof(char) * _TEMP_CHAR_BUFFER_SIZE);
			}
			for(j = 0; j < keys_count; ++j) {
				value_name_size = _TEMP_CHAR_BUFFER_SIZE;
				data_size = _TEMP_CHAR_BUFFER_SIZE;
				enum_result = RegEnumValueA(result, j, value_name, &value_name_size, NULL, NULL, data, &data_size);
				if(enum_result == ERROR_SUCCESS) {
					if(keyword == NULL || strstr(value_name, keyword) != NULL) {
						_STRCPY(names[names_index], _TEMP_CHAR_BUFFER_SIZE, (const char*)data);
						++ names_index;
					}
				}
			}
		}
		CloseHandle(result);
	}
	*count = names_count;
	return names;
}

_LONG _serial_window_open_port(const char* port_name) {
	const char* prefix= "\\\\.\\";
	char port_full_name[_TEMP_CHAR_BUFFER_SIZE];
	HANDLE comm;

	_STRCPY(port_full_name, _TEMP_CHAR_BUFFER_SIZE, prefix);
	_STRCAT(port_full_name, _TEMP_CHAR_BUFFER_SIZE, port_name);

	comm = CreateFileA((LPCSTR)port_full_name,
		GENERIC_READ | GENERIC_WRITE,
		0,
		NULL,
		OPEN_EXISTING,
		FILE_FLAG_OVERLAPPED,
		0);

	if(comm != INVALID_HANDLE_VALUE) {
		DCB dcb = { 0 };
		dcb.DCBlength = sizeof(DCB);
		if(!GetCommState(comm, &dcb)) {
			CloseHandle(comm);
			comm = (HANDLE)_SERIAL_ERROR_INCORRECT_SERIAL_PORT;
		}
	} else {
		DWORD error_value = GetLastError();
		if(error_value == ERROR_ACCESS_DENIED) {
			comm = (HANDLE)_SERIAL_ERROR_PORT_BUSY;
		} else if(error_value == ERROR_FILE_NOT_FOUND) {
			comm = (HANDLE)_SERIAL_ERROR_PORT_NOT_FOUND;
		}
	}
	return (_LONG)comm;
}

int _serial_window_close_port(_LONG port_handle) {
	return CloseHandle((HANDLE)port_handle) ? 1 : 0;
}

int _serial_window_set_params(_LONG port_handle, int baud_rate, int byte_size, int stop_bits, int parity, int set_rts, int set_dtr, int flags) {
	HANDLE comm = (HANDLE)port_handle;
	int return_value = 0;
	DCB dcb = { 0 };

	dcb.DCBlength = sizeof(DCB);
	if(GetCommState(comm, &dcb)) {
		dcb.BaudRate = baud_rate;
		dcb.ByteSize = byte_size;
		dcb.StopBits = stop_bits;
		dcb.Parity = parity;

		if(set_rts == 1) {
			dcb.fRtsControl = RTS_CONTROL_ENABLE;
		} else {
			dcb.fRtsControl = RTS_CONTROL_DISABLE;
		}
		if(set_dtr == 1) {
			dcb.fDtrControl = DTR_CONTROL_ENABLE;
		} else {
			dcb.fDtrControl = DTR_CONTROL_DISABLE;
		}
		dcb.fOutxCtsFlow = FALSE;
		dcb.fOutxDsrFlow = FALSE;
		dcb.fDsrSensitivity = FALSE;
		dcb.fTXContinueOnXoff = TRUE;
		dcb.fOutX = FALSE;
		dcb.fInX = FALSE;
		dcb.fErrorChar = FALSE;
		dcb.fNull = FALSE;
		dcb.fAbortOnError = FALSE;
		dcb.XonLim = 2048;
		dcb.XoffLim = 512;
		dcb.XonChar = (char)17; //DC1
		dcb.XoffChar = (char)19; //DC3

		if(SetCommState(comm, &dcb)) {
			COMMTIMEOUTS comm_timeouts = { 0 };
			comm_timeouts.ReadIntervalTimeout = 0;
			comm_timeouts.ReadTotalTimeoutConstant = 100;
			comm_timeouts.ReadTotalTimeoutMultiplier = 0;
			comm_timeouts.WriteTotalTimeoutConstant = 0;
			comm_timeouts.WriteTotalTimeoutMultiplier = 0;
			if(SetCommTimeouts(comm, &comm_timeouts)) {
				return_value = 1;
			}
		}
	}
	return return_value;
}

int _serial_window_set_flow_control_mode(_LONG port_handle, int mask) {
	HANDLE comm = (HANDLE)port_handle;
	int return_value = 0;
	DCB dcb = { 0 };

	dcb.DCBlength = sizeof(DCB);
	if(GetCommState(comm, &dcb)) {
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
		dcb.fOutxCtsFlow = FALSE;
		dcb.fOutX = FALSE;
		dcb.fInX = FALSE;
		if(mask != _SERIAL_FLOWCONTROL_NONE) {
			if((mask & _SERIAL_FLOWCONTROL_RTSCTS_IN) == _SERIAL_FLOWCONTROL_RTSCTS_IN) {
				dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;
			}
			if((mask & _SERIAL_FLOWCONTROL_RTSCTS_OUT) == _SERIAL_FLOWCONTROL_RTSCTS_OUT) {
				dcb.fOutxCtsFlow = TRUE;
			}
			if((mask & _SERIAL_FLOWCONTROL_XONXOFF_IN) == _SERIAL_FLOWCONTROL_XONXOFF_IN) {
				dcb.fInX = TRUE;
			}
			if((mask & _SERIAL_FLOWCONTROL_XONXOFF_OUT) == _SERIAL_FLOWCONTROL_XONXOFF_OUT) {
				dcb.fOutX = TRUE;
			}
		}
		if(SetCommState(comm, &dcb)) {
			return_value = 1;
		}
	}
	return return_value;
}

int _serial_window_purge_port(_LONG port_handle, int flags) {
	return PurgeComm((HANDLE)port_handle, (DWORD)flags) ? 1 : 0;
}

int _serial_window_count_read_bytes(_LONG port_handle) {
	HANDLE comm = (HANDLE)port_handle;
	DWORD errors;
	COMSTAT comstat = { 0 };
	int return_value = -1;

	if(ClearCommError(comm, &errors, &comstat)) {
		return_value = comstat.cbInQue;
	} else {
		return_value = -1;
	}
	return return_value;
}

int _serial_window_read_bytes(_LONG port_handle, unsigned char* buffer, int buffer_size) {
	HANDLE comm = (HANDLE)port_handle;
	DWORD number_of_bytes_transferred;
	DWORD number_of_bytes_read;
	OVERLAPPED overlapped = { 0 };
	int return_value = 0;

	overlapped.hEvent = CreateEventA(NULL, 1, 0, NULL);
	if(ReadFile(comm, buffer, (DWORD)buffer_size, &number_of_bytes_read, &overlapped)) {
		return_value = number_of_bytes_read;
	} else if(GetLastError() == ERROR_IO_PENDING) {
		if(WaitForSingleObject(overlapped.hEvent, INFINITE) == WAIT_OBJECT_0) {
			if(GetOverlappedResult(comm, &overlapped, &number_of_bytes_transferred, 0)) {
				return_value = number_of_bytes_transferred;
			}
		}
	}
	CloseHandle(overlapped.hEvent);
	return return_value;
}

int _serial_window_write_bytes(_LONG port_handle, const unsigned char* buffer, int buffer_size) {
	HANDLE comm = (HANDLE)port_handle;
	DWORD number_of_bytes_transferred;
	DWORD number_of_bytes_written;
	OVERLAPPED overlapped = { 0 };
	int return_value = 0;

	overlapped.hEvent = CreateEventA(NULL, 1, 0, NULL);
	if(WriteFile(comm, buffer, buffer_size, &number_of_bytes_written, &overlapped)) {
		return_value = 1;
	} else if(GetLastError() == ERROR_IO_PENDING) {
		if(WaitForSingleObject(overlapped.hEvent, INFINITE) == WAIT_OBJECT_0) {
			if(GetOverlappedResult(comm, &overlapped, &number_of_bytes_transferred, 0)) {
				return_value = 1;
			}
		}
	}
	CloseHandle(overlapped.hEvent);
	return return_value;
}

struct _serial* _serial_create(void) {
	struct _serial* serial = (struct _serial*)malloc(sizeof(struct _serial));
	serial->port_handle = 0;
	serial->port_opened = 0;
	serial->buffer = NULL;
	serial->buffer_size = 0;
	serial->offset = 0;
	return serial;
}

void _serial_dispose(struct _serial* serial) {
	if(serial == NULL) return;
	_serial_close(serial); // free buffer
	free(serial);
}

int _serial_open(struct _serial* serial, const char* port_name, int baud_rate, int flow_control) {
	_LONG port_handle;

	if(serial == NULL) return 0;
	if(serial->port_opened == 1) return 0;
	
	port_handle = _serial_window_open_port(port_name);
	if(port_handle == _SERIAL_ERROR_PORT_BUSY) return 0;
	else if(port_handle == _SERIAL_ERROR_PORT_NOT_FOUND) return 0;
	else if(port_handle == _SERIAL_ERROR_PERMISSION_DENIED) return 0;
	else if(port_handle == _SERIAL_ERROR_INCORRECT_SERIAL_PORT) return 0;
	
	serial->port_handle = port_handle;
	serial->port_opened = 1;
	serial->offset = 0;
	
	if(serial->buffer != NULL) {
		free(serial->buffer);
		serial->buffer = NULL;
	}
	serial->buffer_size = _SERIAL_BUFFER_SIZE;
	serial->buffer = (char*)malloc(sizeof(char) * _SERIAL_BUFFER_SIZE);
	_serial_window_set_params(port_handle, baud_rate, _SERIAL_DATABITS_8, 0, _SERIAL_PARITY_NONE, 1, 1, 0);
	_serial_window_set_flow_control_mode(port_handle, flow_control);
	return 1;
}

void _serial_close(struct _serial* serial) {
	if(serial == NULL) return;
	if(serial->port_opened == 0) return;

	serial->buffer_size = 0;
	if(serial->buffer != NULL) {
		free(serial->buffer);
		serial->buffer = NULL;
	}
	if(_serial_window_close_port(serial->port_handle) == 1) {
		serial->port_opened = 0;
	}
	serial->offset = 0;
	serial->port_handle = 0;
}

void _serial_clear(struct _serial* serial) {
	if(serial == NULL) return;
	if(serial->port_opened == 0) return;
	
	serial->offset = 0;
	_serial_window_purge_port(serial->port_handle, _SERIAL_PURGE_RXCLEAR | _SERIAL_PURGE_RXABORT | _SERIAL_PURGE_TXCLEAR | _SERIAL_PURGE_TXABORT);
}

int _serial_read_string_until(struct _serial* serial, char* buffer, int buffer_size, char delimiter) {
	_LONG port_handle;
	int to_read, read_bytes;
	int size, found, i;
	char* temp;

	if(serial == NULL) return 0;
	if(serial->port_opened == 0) return 0;

	port_handle = serial->port_handle;
	to_read = _serial_window_count_read_bytes(port_handle);
	while(to_read > 0) {
		if(serial->buffer_size < serial->offset + to_read) {
			size = (serial->offset + to_read) * 2;
			temp = (char*)malloc(sizeof(char) * size);
			_STRNCPY(temp, size, serial->buffer, serial->buffer_size);
			free(serial->buffer);
			serial->buffer = temp;
			serial->buffer_size = size;
		}
		read_bytes = _serial_window_read_bytes(port_handle, (unsigned char*)(serial->buffer + serial->offset), to_read);
		serial->offset += read_bytes;

		found = -1;
		for(i = 0; i < serial->offset; ++i) {
			if(serial->buffer[i] == delimiter) {
				found = i;
				break;
			}
		}
		if(found != -1) {
			int to_copy = found + 1;
			if(to_copy < buffer_size) {
				_STRNCPY(buffer, buffer_size, serial->buffer, to_copy);
				_STRNCPY(serial->buffer, serial->buffer_size, serial->buffer + to_copy, serial->offset - to_copy + 1);
				serial->offset -= to_copy;
				return to_copy;
			}
		}
		to_read = _serial_window_count_read_bytes(port_handle);
	}
	return 0;
}

int _serial_write(const struct _serial* serial, const char* buffer, int buffer_size) {
	if(serial == NULL) return 0;
	if(serial->port_opened == 0) return 0;
	return _serial_window_write_bytes(serial->port_handle, (const unsigned char*)buffer, buffer_size);
}

/*------------------------------
  CONNECTOR
------------------------------*/

#define _CONNECTION_STATE_NONE 0
#define _CONNECTION_STATE_CONNECTING 1
#define _CONNECTION_STATE_CONNECTED 2
#define _CONNECTION_STATE_CONNECTION_LOST 3
#define _CONNECTION_STATE_DISCONNECTED 4
#define _CONNECTION_STATE_DISPOSED 5
#define _CONNECTION_RESULT_FOUND 1
#define _CONNECTION_RESULT_NOT_CONNECTED 2
#define _CONNECTION_RESULT_NOT_AVAILABLE 3

#define _CONNECTOR_INFO_BUFFER_SIZE 20
#define _CONNECTOR_BUFFER_SIZE 256
#define _TIMEOUT 100 // milliseconds
#define _RETRY 10

#define _VALID_PACKET_LENGTH 54
#define _MOTORING_PACKET_LENGTH 54
#define _CR 13

#define _DEFAULT_ADDRESS "000000000000"

struct _connector;
typedef int (*_CHECK_CONNECTION)(struct _connector* connector, struct _serial* serial);

struct _connector {
	struct _serial* serial;
	char* tag;
	int index;
	int packet_length;
	char delimiter;
	char* address;
	char* port_name;
	int found;
	int connected;
	int checking_timeout;
	double timestamp;
	char* buffer;
	void* arg;
	_CHECK_CONNECTION check_connection;
};

struct _connector* _connector_create(const char* tag, int index, int packet_length, char delimiter);
void _connector_dispose(struct _connector* connector);
int _connector_open(struct _connector* connector, const char* port_name, int baud_rate, int flow_control);
int _connector_check_port(struct _connector* connector, struct _serial* serial);
int _connector_open_port(struct _connector* connector, const char* port_name, int baud_rate, int flow_control);
void _connector_close(struct _connector* connector);
int _connector_is_connected(const struct _connector* connector);
const char* _connector_get_port_name(const struct _connector* connector);
const char* _connector_get_address(const struct _connector* connector);
void _connector_set_address(const struct _connector* connector, const char* address);
void _connector_set_connection_state(struct _connector* connector, int state);
int _connector_read_packet(const struct _connector* connector, struct _serial* serial, const char* start_bytes);
void _connector_write(const struct _connector* connector, const char* buffer, int buffer_size);
int _connector_read(struct _connector* _connector);
void _connector_print_state(const struct _connector* connector, int state);
void _connector_print_error(const struct _connector* connector, int error_code);

double _get_timestamp(struct timeb* t) {
	ftime(t);
	return (double)t->time + (double)t->millitm / 1000.0;
}

struct _connector* _connector_create(const char* tag, int index, int packet_length, char delimiter) {
	struct _connector* connector = (struct _connector*)malloc(sizeof(struct _connector));
	
	connector->serial = NULL;
	connector->index = index;
	connector->packet_length = packet_length;
	connector->delimiter = delimiter;
	
	connector->tag = (char*)malloc(sizeof(char) * _CONNECTOR_INFO_BUFFER_SIZE);
	connector->address = (char*)malloc(sizeof(char) * _CONNECTOR_INFO_BUFFER_SIZE);
	connector->port_name = (char*)malloc(sizeof(char) * _CONNECTOR_INFO_BUFFER_SIZE);
	_STRCPY(connector->tag, _CONNECTOR_INFO_BUFFER_SIZE, tag);
	_STRCPY(connector->address, _CONNECTOR_INFO_BUFFER_SIZE, _DEFAULT_ADDRESS);
	_STRCPY(connector->port_name, _CONNECTOR_INFO_BUFFER_SIZE, "");
	
	connector->found = 0;
	connector->connected = 0;
	connector->checking_timeout = 0;
	connector->timestamp = 0;
	
	connector->buffer = (char*)malloc(sizeof(char) * _CONNECTOR_BUFFER_SIZE);
	connector->check_connection = NULL;

	return connector;
}

void _connector_dispose(struct _connector* connector) {
	if(connector == NULL) return;

	_connector_close(connector); // close serial
	if(connector->tag != NULL) {
		free(connector->tag);
		connector->tag = NULL;
	}
	if(connector->address != NULL) {
		free(connector->address);
		connector->address = NULL;
	}
	if(connector->port_name != NULL) {
		free(connector->port_name);
		connector->port_name = NULL;
	}
	if(connector->buffer != NULL) {
		free(connector->buffer);
		connector->buffer = NULL;
	}
	connector->check_connection = NULL;
	free(connector);
}

int _connector_open(struct _connector* connector, const char* port_name, int baud_rate, int flow_control) {
	int result = _CONNECTION_RESULT_NOT_AVAILABLE;

	if(connector == NULL) return _CONNECTION_RESULT_NOT_AVAILABLE;
	if(port_name == NULL) {
		int port_count = 0;
		char** port_names = _serial_window_get_serial_port_names(NULL, &port_count);
		if(port_count > 0 && port_names != NULL) {
			int i;
	
			for(i = 0; i < port_count; ++i) {
				result = _connector_open_port(connector, port_names[i], baud_rate, flow_control);
				if(result != _CONNECTION_RESULT_NOT_AVAILABLE) {
					break;
				}
			}
			for(i = 0; i < port_count; ++i) {
				free(port_names[i]);
			}
			free(port_names);
		}
	} else {
		result = _connector_open_port(connector, port_name, baud_rate, flow_control);
	}
	if(result == _CONNECTION_RESULT_NOT_AVAILABLE) {
		_connector_print_error(connector, result);
	}
	return result;
}

int _connector_check_port(struct _connector* connector, struct _serial* serial) {
	int read_bytes1, read_bytes2;

	_connector_read_packet(connector, serial, NULL);
	read_bytes1 = _connector_read_packet(connector, serial, NULL);
	read_bytes2 = _connector_read_packet(connector, serial, NULL);

	if(read_bytes2 != 0) {
		if(read_bytes2 == connector->packet_length) {
			if(connector->check_connection != NULL) {
				return connector->check_connection(connector, serial);
			}
		} else if(read_bytes1 != 0 && read_bytes2 == 2) {
			_connector_print_error(connector, _CONNECTION_RESULT_NOT_CONNECTED);
			return _CONNECTION_RESULT_NOT_CONNECTED;
		}
	}
	return _CONNECTION_RESULT_NOT_AVAILABLE;
}

int _connector_open_port(struct _connector* connector, const char* port_name, int baud_rate, int flow_control) {
	struct _serial* serial = _serial_create();
	if(_serial_open(serial, port_name, baud_rate, flow_control) == 1) {
		int result;
		
		_serial_clear(serial);
		_STRCPY(connector->port_name, _CONNECTOR_INFO_BUFFER_SIZE, port_name);
		
		result = _connector_check_port(connector, serial);
		if(result != _CONNECTION_RESULT_NOT_AVAILABLE) {
			connector->serial = serial;
			return result;
		}
		_serial_close(serial);
	}
	_serial_dispose(serial);
	serial = NULL;
	
	return _CONNECTION_RESULT_NOT_AVAILABLE;
}

void _connector_close(struct _connector* connector) {
	if(connector == NULL) return;
	if(connector->serial != NULL) {
		_serial_dispose(connector->serial);
		connector->serial = NULL;
	}
	connector->connected = 0;
	_connector_print_state(connector, _CONNECTION_STATE_DISPOSED);
}

int _connector_is_connected(const struct _connector* connector) {
	if(connector == NULL) return 0;
	return connector->connected;
}

const char* _connector_get_port_name(const struct _connector* connector) {
	if(connector == NULL) return "";
	return connector->port_name;
}

const char* _connector_get_address(const struct _connector* connector) {
	if(connector == NULL) return _DEFAULT_ADDRESS;
	return connector->address;
}

void _connector_set_address(const struct _connector* connector, const char* address) {
	if(connector == NULL) return;
	_STRCPY(connector->address, _CONNECTOR_INFO_BUFFER_SIZE, address);
}

void _connector_set_connection_state(struct _connector* connector, int state) {
	if(connector == NULL) return;
	connector->connected = (state == _CONNECTION_STATE_CONNECTED) ? 1 : 0;
	if(connector->found == 0 && connector->connected == 1) {
		connector->found = 1;
	}
	if(connector->found == 1) {
		_connector_print_state(connector, state);
	}
}

int _connector_read_packet(const struct _connector* connector, struct _serial* serial, const char* start_bytes) {
	if(connector != NULL) {
		char* buffer = connector->buffer;
		int read_bytes = 0, i;

		for(i = 0; i < _RETRY; ++i) {
			Sleep(10);
			if((read_bytes = _serial_read_string_until(serial, buffer, _CONNECTOR_BUFFER_SIZE, connector->delimiter)) != 0) {
				if(start_bytes == NULL) {
					return read_bytes;
				}
				if(buffer[0] == start_bytes[0] && buffer[1] == start_bytes[1]) {
					return read_bytes;
				}
			}
		}
	}
	return 0;
}

void _connector_write(const struct _connector* connector, const char* buffer, int buffer_size) {
	if(connector == NULL || connector->serial == NULL) return;
	_serial_write(connector->serial, buffer, buffer_size);
}

int _connector_read(struct _connector* connector) {
	int read_bytes;

	if(connector == NULL || connector->serial == NULL) return 0;
	read_bytes = _serial_read_string_until(connector->serial, connector->buffer, _CONNECTOR_BUFFER_SIZE, connector->delimiter);
	if(read_bytes == connector->packet_length) {
		if(connector->found == 0) {
			if(connector->check_connection != NULL) {
				connector->check_connection(connector, connector->serial);
			}
		} else if(connector->connected == 0) {
			_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTED);
		}
		connector->checking_timeout = 0;
		connector->timestamp = 0;
		return read_bytes;
	} else if(connector->connected == 1) {
		struct timeb time;
		double t;

		t = _get_timestamp(&time);
		if(connector->checking_timeout == 0) {
			connector->checking_timeout = 1;
			connector->timestamp = t;
		} else if(t - connector->timestamp > _TIMEOUT) {
			_serial_clear(connector->serial);
			_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTION_LOST);
		}
	}
	return 0;
}

void _connector_print_state(const struct _connector* connector, int state) {
	switch(state) {
		case _CONNECTION_STATE_CONNECTED:
			printf("%s[%d] Connected: %s\n", connector->tag, connector->index, connector->port_name);
			break;
		case _CONNECTION_STATE_CONNECTION_LOST:
			printf("%s[%d] Connection lost\n", connector->tag, connector->index);
			break;
		case _CONNECTION_STATE_DISCONNECTED:
			printf("%s[%d] Disconnected\n", connector->tag, connector->index);
			break;
		case _CONNECTION_STATE_DISPOSED:
			printf("%s[%d] Disposed\n", connector->tag, connector->index);
			break;
	}
}

void _connector_print_error(const struct _connector* connector, int error_code) {
	switch(error_code) {
		case _CONNECTION_RESULT_NOT_AVAILABLE:
			printf("%s[%d] No available USB to BLE bridge\n", connector->tag, connector->index);
			break;
		case _CONNECTION_RESULT_NOT_CONNECTED:
			printf("%s[%d] Not connected\n", connector->tag, connector->index);
			break;
	}
}

/*------------------------------
  DEVICE
------------------------------*/

#define MIN(a, b) ((a) < (b) ? (a) : (b))

struct _device {
	char* name;
	int id;
	int device_type;
	int data_type;
	int data_size;
	int data_len;
	void* data;
	float min_value;
	float max_value;
	float initial_value;
	int event;
	int fired;
	int written;
	void (*reset)(struct _device* device);
	int (*read)(const struct _device* device);
	int (*read_at)(const struct _device* device, int index);
	int (*read_array)(const struct _device* device, int* data, int length);
	float (*read_float)(const struct _device* device);
	float (*read_float_at)(const struct _device* device, int index);
	int (*read_float_array)(const struct _device* device, float* data, int length);
	int (*write)(struct _device* device, int data);
	int (*write_at)(struct _device* device, int index, int data);
	int (*write_array)(struct _device* device, const int* data, int length);
	int (*write_float)(struct _device* device, float data);
	int (*write_float_at)(struct _device* device, int index, float data);
	int (*write_float_array)(struct _device* device, const float* data, int length);
	int (*put)(struct _device* device, int data, int fired);
	int (*put_at)(struct _device* device, int index, int data, int fired);
	int (*put_array)(struct _device* device, const int* data, int length, int fired);
	int (*put_float)(struct _device* device, float data, int fired);
	int (*put_float_at)(struct _device* device, int index, float data, int fired);
	int (*put_float_array)(struct _device* device, const float* data, int length, int fired);
};

struct _device* _device_create(int id, const char* name, int device_type, int data_type, int data_size, float min_value, float max_value, float initial_value);
void _device_dispose(struct _device* device);
const char* _device_get_name(const struct _device* device);
void _device_set_name(struct _device* device, const char* name);
int _device_get_id(const struct _device* device);
int _device_get_device_type(const struct _device* device);
int _device_get_data_type(const struct _device* device);
int _device_get_data_size(const struct _device* device);
int _device_e(const struct _device* device);
void _device_reset(struct _device* device);
int _device_read(const struct _device* device);
int _device_read_at(const struct _device* device, int index);
int _device_read_array(const struct _device* device, int* data, int length);
float _device_read_float(const struct _device* device);
float _device_read_float_at(const struct _device* device, int index);
int _device_read_float_array(const struct _device* device, float* data, int length);
int _device_write(struct _device* device, int data);
int _device_write_at(struct _device* device, int index, int data);
int _device_write_array(struct _device* device, const int* data, int length);
int _device_write_float(struct _device* device, float data);
int _device_write_float_at(struct _device* device, int index, float data);
int _device_write_float_array(struct _device* device, const float* data, int length);
int _device_put(struct _device* device, int data, int fired);
int _device_put_at(struct _device* device, int index, int data, int fired);
int _device_put_array(struct _device* device, const int* data, int length, int fired);
int _device_put_float(struct _device* device, float data, int fired);
int _device_put_float_at(struct _device* device, int index, float data, int fired);
int _device_put_float_array(struct _device* device, const float* data, int length, int fired);
void _device_update_device_state(struct _device* device);

void _int_device_reset(struct _device* device) {
	int* this_data = (int*)device->data;
	
	if(this_data != NULL) {
		int this_len = device->data_len, i;
		int initial_value = (int)device->initial_value;
		
		for(i = 0; i < this_len; ++i) {
			this_data[i] = initial_value;
		}
	}
	device->event = 0;
	device->fired = 0;
	device->written = 0;
}

int _int_device_read(const struct _device* device) {
	int* this_data;
	
	if(device->data_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	return this_data[0];
}

int _int_device_read_at(const struct _device* device, int index) {
	int* this_data;

	if(index < 0 || index >= device->data_len) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	return this_data[index];
}

int _int_device_read_array(const struct _device* device, int* data, int length) {
	int* this_data;
	int this_len, len;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	memcpy(data, this_data, sizeof(int) * len);
	return len;
}

float _int_device_read_float(const struct _device* device) {
	return (float)_int_device_read(device);
}

float _int_device_read_float_at(const struct _device* device, int index) {
	return (float)_int_device_read_at(device, index);
}

int _int_device_read_float_array(const struct _device* device, float* data, int length) {
	int* this_data;
	int this_len, len, i;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	for(i = 0; i < len; ++i) {
		data[i] = (float)this_data[i];
	}
	return len;
}

int _int_device_write(struct _device* device, int data) {
	int* this_data;
	int min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(device->data_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	min_value = (int)device->min_value;
	max_value = (int)device->max_value;
	if(data < min_value) data = min_value;
	else if(data > max_value) data = max_value;
	
	this_data[0] = data;
	device->fired = 1;
	device->written = 1;
	return 1;
}

int _int_device_write_at(struct _device* device, int index, int data) {
	int* this_data;
	int min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(index < 0 || index >= device->data_len) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	min_value = (int)device->min_value;
	max_value = (int)device->max_value;
	if(data < min_value) data = min_value;
	else if(data > max_value) data = max_value;
	
	this_data[index] = data;
	device->fired = 1;
	device->written = 1;
	return 1;
}

int _int_device_write_array(struct _device* device, const int* data, int length) {
	int* this_data;
	int this_len, len, i;
	int value, min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	min_value = (int)device->min_value;
	max_value = (int)device->max_value;
	for(i = 0; i < len; ++i) {
		value = data[i];
		if(value < min_value) value = min_value;
		else if(value > max_value) value = max_value;
		this_data[i] = value;
	}
	device->fired = 1;
	device->written = 1;
	return len;
}

int _int_device_write_float(struct _device* device, float data) {
	return _int_device_write(device, (int)data);
}

int _int_device_write_float_at(struct _device* device, int index, float data) {
	return _int_device_write_at(device, index, (int)data);
}

int _int_device_write_float_array(struct _device* device, const float* data, int length) {
	int* this_data;
	int this_len, len, i;
	int value, min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	min_value = (int)device->min_value;
	max_value = (int)device->max_value;
	for(i = 0; i < len; ++i) {
		value = (int)data[i];
		if(value < min_value) value = min_value;
		else if(value > max_value) value = max_value;
		this_data[i] = value;
	}
	device->fired = 1;
	device->written = 1;
	return len;
}

int _int_device_put(struct _device* device, int data, int fired) {
	int* this_data;

	if(device->data_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	this_data[0] = data;
	if(fired == 1) device->fired = 1;
	return 1;
}

int _int_device_put_at(struct _device* device, int index, int data, int fired) {
	int* this_data;

	if(index < 0 || index >= device->data_len) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	this_data[index] = data;
	if(fired == 1) device->fired = 1;
	return 1;
}

int _int_device_put_array(struct _device* device, const int* data, int length, int fired) {
	int* this_data;
	int this_len, len;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	memcpy(this_data, data, sizeof(int) * len);
	if(fired == 1) device->fired = 1;
	return len;
}

int _int_device_put_float(struct _device* device, float data, int fired) {
	return _int_device_put(device, (int)data, fired);
}

int _int_device_put_float_at(struct _device* device, int index, float data, int fired) {
	return _int_device_put_at(device, index, (int)data, fired);
}

int _int_device_put_float_array(struct _device* device, const float* data, int length, int fired) {
	int* this_data;
	int this_len, len, i;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (int*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	for(i = 0; i < len; ++i) {
		this_data[i] = (int)data[i];
	}
	if(fired == 1) device->fired = 1;
	return len;
}

void _float_device_reset(struct _device* device) {
	float* this_data = (float*)device->data;
	
	if(this_data != NULL) {
		int this_len = device->data_len, i;
		float initial_value = device->initial_value;
		
		for(i = 0; i < this_len; ++i) {
			this_data[i] = initial_value;
		}
	}
	device->event = 0;
	device->fired = 0;
	device->written = 0;
}

int _float_device_read_array(const struct _device* device, int* data, int length) {
	float* this_data;
	int this_len, len, i;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	for(i = 0; i < len; ++i) {
		data[i] = (int)this_data[i];
	}
	return len;
}

float _float_device_read_float(const struct _device* device) {
	float* this_data;

	if(device->data_len <= 0) return 0.0f;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0.0f;
	
	return this_data[0];
}

float _float_device_read_float_at(const struct _device* device, int index) {
	float* this_data;

	if(index < 0 || index >= device->data_len) return 0.0f;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0.0f;
	
	return this_data[index];
}

int _float_device_read(const struct _device* device) {
	return (int)_float_device_read_float(device);
}

int _float_device_read_at(const struct _device* device, int index) {
	return (int)_float_device_read_float_at(device, index);
}

int _float_device_read_float_array(const struct _device* device, float* data, int length) {
	float* this_data;
	int this_len, len;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	memcpy(data, this_data, sizeof(float) * len);
	return len;
}

int _float_device_write_array(struct _device* device, const int* data, int length) {
	float* this_data;
	int this_len, len, i;
	float value, min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	min_value = device->min_value;
	max_value = device->max_value;
	for(i = 0; i < len; ++i) {
		value = (float)data[i];
		if(value < min_value) value = min_value;
		else if(value > max_value) value = max_value;
		this_data[i] = value;
	}
	device->fired = 1;
	device->written = 1;
	return len;
}

int _float_device_write_float(struct _device* device, float data) {
	float* this_data;
	float min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(device->data_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	min_value = device->min_value;
	max_value = device->max_value;
	if(data < min_value) data = min_value;
	else if(data > max_value) data = max_value;
	
	this_data[0] = data;
	device->fired = 1;
	device->written = 1;
	return 1;
}

int _float_device_write_float_at(struct _device* device, int index, float data) {
	float* this_data;
	float min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(index < 0 || index >= device->data_len) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	min_value = device->min_value;
	max_value = device->max_value;
	if(data < min_value) data = min_value;
	else if(data > max_value) data = max_value;
	
	this_data[index] = data;
	device->fired = 1;
	device->written = 1;
	return 1;
}

int _float_device_write(struct _device* device, int data) {
	return _float_device_write_float(device, (float)data);
}

int _float_device_write_at(struct _device* device, int index, int data) {
	return _float_device_write_float_at(device, index, (float)data);
}

int _float_device_write_float_array(struct _device* device, const float* data, int length) {
	float* this_data;
	int this_len, len, i;
	float value, min_value, max_value;

	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;
	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	min_value = device->min_value;
	max_value = device->max_value;
	for(i = 0; i < len; ++i) {
		value = data[i];
		if(value < min_value) value = min_value;
		else if(value > max_value) value = max_value;
		this_data[i] = value;
	}
	device->fired = 1;
	device->written = 1;
	return len;
}

int _float_device_put_array(struct _device* device, const int* data, int length, int fired) {
	float* this_data;
	int this_len, len, i;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	for(i = 0; i < len; ++i) {
		this_data[i] = (float)data[i];
	}
	if(fired == 1) device->fired = 1;
	return len;
}

int _float_device_put_float(struct _device* device, float data, int fired) {
	float* this_data;

	if(device->data_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	this_data[0] = data;
	if(fired == 1) device->fired = 1;
	return 1;
}

int _float_device_put_float_at(struct _device* device, int index, float data, int fired) {
	float* this_data;

	if(index < 0 || index >= device->data_len) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	this_data[index] = data;
	if(fired == 1) device->fired = 1;
	return 1;
}

int _float_device_put(struct _device* device, int data, int fired) {
	return _float_device_put_float(device, (float)data, fired);
}

int _float_device_put_at(struct _device* device, int index, int data, int fired) {
	return _float_device_put_float_at(device, index, (float)data, fired);
}

int _float_device_put_float_array(struct _device* device, const float* data, int length, int fired) {
	float* this_data;
	int this_len, len;

	if(data == NULL) return 0;
	if(length <= 0) return 0;
	
	this_len = device->data_len;
	if(this_len <= 0) return 0;
	this_data = (float*)device->data;
	if(this_data == NULL) return 0;
	
	len = MIN(this_len, length);
	memcpy(this_data, data, sizeof(float) * len);
	if(fired == 1) device->fired = 1;
	return len;
}

struct _device* _device_create(int id, const char* name, int device_type, int data_type, int data_size, float min_value, float max_value, float initial_value) {
	if(data_type == DATA_TYPE_INTEGER || data_type == DATA_TYPE_FLOAT) {
		struct _device* device = (struct _device*)malloc(sizeof(struct _device));
		int len;

		if(name == NULL) name = "";
		len = strlen(name) + 1;
		device->name = (char*)malloc(sizeof(char) * len);
		_STRCPY(device->name, len, name);

		device->id = id & 0xfff00fff;
		device->device_type = device_type;
		device->data_type = data_type;
		device->data_size = data_size;
		device->data_len = 0;
		device->data = NULL;
		device->min_value = min_value;
		device->max_value = max_value;
		device->initial_value = initial_value;
		device->event = 0;
		device->fired = 0;
		device->written = 0;

		if(data_type == DATA_TYPE_INTEGER) {
			device->reset = _int_device_reset;
			device->read = _int_device_read;
			device->read_at = _int_device_read_at;
			device->read_array = _int_device_read_array;
			device->read_float = _int_device_read_float;
			device->read_float_at = _int_device_read_float_at;
			device->read_float_array = _int_device_read_float_array;
			device->write = _int_device_write;
			device->write_at = _int_device_write_at;
			device->write_array = _int_device_write_array;
			device->write_float = _int_device_write_float;
			device->write_float_at = _int_device_write_float_at;
			device->write_float_array = _int_device_write_float_array;
			device->put = _int_device_put;
			device->put_at = _int_device_put_at;
			device->put_array = _int_device_put_array;
			device->put_float = _int_device_put_float;
			device->put_float_at = _int_device_put_float_at;
			device->put_float_array = _int_device_put_float_array;
		
			if(data_size > 0) {
				int* data;
				int i;

				data = (int*)malloc(sizeof(int) * data_size);
				for(i = 0; i < data_size; ++i) {
					data[i] = (int)initial_value;
				}
				device->data = data;
				device->data_len = data_size;
			}
		} else if(data_type == DATA_TYPE_FLOAT) {
			device->reset = _float_device_reset;
			device->read = _float_device_read;
			device->read_at = _float_device_read_at;
			device->read_array = _float_device_read_array;
			device->read_float = _float_device_read_float;
			device->read_float_at = _float_device_read_float_at;
			device->read_float_array = _float_device_read_float_array;
			device->write = _float_device_write;
			device->write_at = _float_device_write_at;
			device->write_array = _float_device_write_array;
			device->write_float = _float_device_write_float;
			device->write_float_at = _float_device_write_float_at;
			device->write_float_array = _float_device_write_float_array;
			device->put = _float_device_put;
			device->put_at = _float_device_put_at;
			device->put_array = _float_device_put_array;
			device->put_float = _float_device_put_float;
			device->put_float_at = _float_device_put_float_at;
			device->put_float_array = _float_device_put_float_array;
		
			if(data_size > 0) {
				float* data;
				int i;

				data = (float*)malloc(sizeof(float) * data_size);
				for(i = 0; i < data_size; ++i) {
					data[i] = initial_value;
				}
				device->data = data;
				device->data_len = data_size;
			}
		}
		return device;
	}
	return NULL;
}

void _device_dispose(struct _device* device) {
	if(device == NULL) return;
	
	if(device->name != NULL) {
		free(device->name);
		device->name = NULL;
	}
	if(device->data_type == DATA_TYPE_INTEGER) {
		int* data = (int*)device->data;
		if(data != NULL) {
			free(data);
			device->data = NULL;
		}
	} else if(device->data_type == DATA_TYPE_FLOAT) {
		float* data = (float*)device->data;
		if(data != NULL) {
			free(data);
			device->data = NULL;
		}
	}
	free(device);
}

const char* _device_get_name(const struct _device* device) {
	if(device == NULL) return "";
	return device->name;
}

void _device_set_name(struct _device* device, const char* name) {
	int len;

	if(device == NULL) return;
	if(name == NULL) name = "";
	if(device->name != NULL) {
		free(device->name);
	}
	len = strlen(name) + 1;
	device->name = (char*)malloc(sizeof(char) * len);
	_STRCPY(device->name, len, name);
}

int _device_get_id(const struct _device* device) {
	if(device == NULL) return 0;
	return device->id;
}

int _device_get_device_type(const struct _device* device) {
	if(device == NULL) return 0;
	return device->device_type;
}

int _device_get_data_type(const struct _device* device) {
	if(device == NULL) return 0;
	return device->data_type;
}

int _device_get_data_size(const struct _device* device) {
	if(device == NULL) return 0;
	return device->data_size;
}

int _device_e(const struct _device* device) {
	if(device == NULL) return 0;
	return device->event;
}

void _device_reset(struct _device* device) {
	if(device == NULL) return;
	device->reset(device);
}

int _device_read(const struct _device* device) {
	if(device == NULL) return 0;
	return device->read(device);
}

int _device_read_at(const struct _device* device, int index) {
	if(device == NULL) return 0;
	return device->read_at(device, index);
}

int _device_read_array(const struct _device* device, int* data, int length) {
	if(device == NULL) return 0;
	return device->read_array(device, data, length);
}

float _device_read_float(const struct _device* device) {
	if(device == NULL) return 0.0f;
	return device->read_float(device);
}

float _device_read_float_at(const struct _device* device, int index) {
	if(device == NULL) return 0.0f;
	return device->read_float_at(device, index);
}

int _device_read_float_array(const struct _device* device, float* data, int length) {
	if(device == NULL) return 0;
	return device->read_float_array(device, data, length);
}

int _device_write_empty(struct _device* device) {
	if(device == NULL) return 0;
	if(device->device_type == DEVICE_TYPE_SENSOR || device->device_type == DEVICE_TYPE_EVENT) return 0;

	device->fired = 1;
	device->written = 1;
	return 1;
}

int _device_write(struct _device* device, int data) {
	if(device == NULL) return 0;
	return device->write(device, data);
}

int _device_write_at(struct _device* device, int index, int data) {
	if(device == NULL) return 0;
	return device->write_at(device, index, data);
}

int _device_write_array(struct _device* device, const int* data, int length) {
	if(device == NULL) return 0;
	return device->write_array(device, data, length);
}

int _device_write_float(struct _device* device, float data) {
	if(device == NULL) return 0;
	return device->write_float(device, data);
}

int _device_write_float_at(struct _device* device, int index, float data) {
	if(device == NULL) return 0;
	return device->write_float_at(device, index, data);
}

int _device_write_float_array(struct _device* device, const float* data, int length) {
	if(device == NULL) return 0;
	return device->write_float_array(device, data, length);
}

int _device_put_empty(struct _device* device, int fired) {
	if(device == NULL) return 0;
	if(fired == 1) device->fired = 1;
	return 1;
}

int _device_put(struct _device* device, int data, int fired) {
	if(device == NULL) return 0;
	return device->put(device, data, fired);
}

int _device_put_at(struct _device* device, int index, int data, int fired) {
	if(device == NULL) return 0;
	return device->put_at(device, index, data, fired);
}

int _device_put_array(struct _device* device, const int* data, int length, int fired) {
	if(device == NULL) return 0;
	return device->put_array(device, data, length, fired);
}

int _device_put_float(struct _device* device, float data, int fired) {
	if(device == NULL) return 0;
	return device->put_float(device, data, fired);
}

int _device_put_float_at(struct _device* device, int index, float data, int fired) {
	if(device == NULL) return 0;
	return device->put_float_at(device, index, data, fired);
}

int _device_put_float_array(struct _device* device, const float* data, int length, int fired) {
	if(device == NULL) return 0;
	return device->put_float_array(device, data, length, fired);
}

void _device_update_device_state(struct _device* device) {
	if(device == NULL) return;
	device->event = device->fired;
	device->fired = 0;
}

/*------------------------------
  UTIL
------------------------------*/

int _hex_char_to_value(char character);
int _value_to_hex(char* buffer, int index, int value, int bytes);
int _hex_to_value(const char* str, int start, int end);

char _HEX_DIGITS[] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };

int _hex_char_to_value(char character) {
	switch(character) {
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			return (int)(character - '0');
		case 'a':
		case 'b':
		case 'c':
		case 'd':
		case 'e':
		case 'f':
			return (int)(character - 'a' + 10);
		case 'A':
		case 'B':
		case 'C':
		case 'D':
		case 'E':
		case 'F':
			return (int)(character - 'A' + 10);
	}
	return ' ';
}

int _value_to_hex(char* buffer, int index, int value, int bytes) {
	int high, low, val, i;

	for(i = 0; i < bytes; ++i) {
		val = value >> ((bytes - i - 1) * 8);
		high = (val >> 4) & 0xf;
		low = val & 0xf;
		buffer[index++] = _HEX_DIGITS[high];
		buffer[index++] = _HEX_DIGITS[low];
	}
	return index;
}

int _hex_to_value(const char* str, int start, int end) {
	int result = 0, i;

	for(i = start; i < end; ++i) {
		result <<= 4;
		result += _hex_char_to_value(str[i]);
	}
	return result;
}

int _round_float_value(float value) {
	if(value < 0) return (int)(value - 0.5f);
	else return (int)(value + 0.5f);
}

int _round_double_value(double value) {
	if(value < 0) return (int)(value - 0.5);
	else return (int)(value + 0.5);
}

int _strcasecmp(const char* s1, const char* s2) {
	const unsigned char* p1 = (const unsigned char*)s1;
	const unsigned char* p2 = (const unsigned char*)s2;
	int result;
	
	if(p1 == p2) return 0;
	while((result = tolower(*p1) - tolower(*p2++)) == 0) {
		if(*p1++ == '\0') break;
	}
	return result;
}

/*------------------------------
  ROBOT
------------------------------*/

struct _robot;
typedef struct _device* (*_FIND_DEVICE)(struct _robot* robot, int device_id);
typedef void (*_REQUEST_MOTORING_DATA)(struct _robot* robot);
typedef void (*_UPDATE_SENSORY_DEVICE_STATE)(struct _robot* robot);
typedef void (*_UPDATE_MOTORING_DEVICE_STATE)(struct _robot* robot);
typedef void (*_RESET)(struct _robot* robot);
typedef void (*_DISPOSE)(struct _robot* robot);

struct _robot {
	int model_code;
	int index;
	char* name;
	char* write_buffer;
	int devices_size;
	struct _device** devices;
	struct _connector* connector;
	int alive;
	int running;
	int ready;
	int thread_alive;
	HANDLE thread_handle;
	_FIND_DEVICE find_device;
	_REQUEST_MOTORING_DATA request_motoring_data;
	_UPDATE_SENSORY_DEVICE_STATE update_sensory_device_state;
	_UPDATE_MOTORING_DEVICE_STATE update_motoring_device_state;
	_RESET reset;
	_DISPOSE dispose;
};

void _robot_init(struct _robot* robot, int index, const char* name, int write_buffer_size) {
	if(robot != NULL) {
		int len;
		robot->index = index;
		if(name == NULL) name = "";
		len = strlen(name) + 1;
		robot->name = (char*)malloc(sizeof(char) * len);
		_STRCPY(robot->name, len, name);
		robot->write_buffer = (char*)malloc(sizeof(char) * write_buffer_size);
		robot->alive = 0;
		robot->running = 0;
		robot->ready = 0;
		robot->thread_alive = 0;
		robot->thread_handle = 0;
	}
}

void _robot_dispose(struct _robot* robot) {
	if(robot != NULL) {
		struct _device** devices = robot->devices;
		int size = robot->devices_size, i;

		if(robot->connector != NULL) {
			_connector_dispose(robot->connector);
			robot->connector = NULL;
		}
		if(robot->name != NULL) {
			free(robot->name);
			robot->name = NULL;
		}
		if(robot->write_buffer != NULL) {
			free(robot->write_buffer);
			robot->write_buffer = NULL;
		}
		for(i = 0; i < size; ++i) {
			_device_dispose(devices[i]);
			devices[i] = NULL;
		}
		free(devices);
		robot->devices = NULL;
		robot->devices_size = 0;
	}
}

const char* _robot_get_name(struct _robot* robot) {
	if(robot == NULL) return "";
	return robot->name;
}

void _robot_set_name(struct _robot* robot, const char* name) {
	if(robot != NULL) {
		int len;
		if(robot->name != NULL) {
			free(robot->name);
		}
		if(name == NULL) name = "";
		len = strlen(name) + 1;
		robot->name = (char*)malloc(sizeof(char) * len);
		_STRCPY(robot->name, len, name);
	}
}

int _robot_get_index(struct _robot* robot) {
	if(robot == NULL) return 0;
	return robot->index;
}

void _robot_reset(struct _robot* robot) {
	if(robot != NULL) {
		struct _device** devices = robot->devices;
		int size = robot->devices_size, i;
		for(i = 0; i < size; ++i) {
			_device_reset(devices[i]);
		}
	}
}

void _robot_clear_written(struct _robot* robot) {
	if(robot != NULL) {
		struct _device** devices = robot->devices;
		int size = robot->devices_size, i;
		for(i = 0; i < size; ++i) {
			devices[i]->written = 0;
		}
	}
}

struct _device* _robot_find_device(struct _robot* robot, int device_id) {
	if(robot == NULL) return NULL;
	return robot->find_device(robot, device_id);
}

int _robot_e(struct _robot* robot, int device_id) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_e(device);
}

int _robot_read(struct _robot* robot, int device_id) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read(device);
}

int _robot_read_at(struct _robot* robot, int device_id, int index) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read_at(device, index);
}

int _robot_read_array(struct _robot* robot, int device_id, int* data, int length) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read_array(device, data, length);
}

float _robot_read_float(struct _robot* robot, int device_id) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read_float(device);
}

float _robot_read_float_at(struct _robot* robot, int device_id, int index) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read_float_at(device, index);
}

int _robot_read_float_array(struct _robot* robot, int device_id, float* data, int length) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_read_float_array(device, data, length);
}

int _robot_write_empty(struct _robot* robot, int device_id) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_empty(device);
}

int _robot_write(struct _robot* robot, int device_id, int data) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write(device, data);
}

int _robot_write_at(struct _robot* robot, int device_id, int index, int data) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_at(device, index, data);
}

int _robot_write_array(struct _robot* robot, int device_id, const int* data, int length) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_array(device, data, length);
}

int _robot_write_float(struct _robot* robot, int device_id, float data) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_float(device, data);
}

int _robot_write_float_at(struct _robot* robot, int device_id, int index, float data) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_float_at(device, index, data);
}

int _robot_write_float_array(struct _robot* robot, int device_id, const float* data, int length) {
	struct _device* device = _robot_find_device(robot, device_id);
	return _device_write_float_array(device, data, length);
}

#define _NUM_ROBOT_GROUPS 3

struct _robot_group {
	int robots_size;
	int robots_count;
	struct _robot** robots;
};

struct _robot_group** _robot_groups = NULL;

void _robot_group_dispose_all(void) {
	if(_robot_groups != NULL) {
		struct _robot_group* group;
		int i;
		for(i = 0; i < _NUM_ROBOT_GROUPS; ++i) {
			group = _robot_groups[i];
			if(group != NULL) {
				if(group->robots != NULL) {
					free(group->robots);
					group->robots = NULL;
				}
				free(group);
				_robot_groups[i] = NULL;
			}
		}
		free(_robot_groups);
		_robot_groups = NULL;
	}
}

struct _robot_group* _robot_group_get(int group_index) {
	struct _robot_group* group;
	
	if(group_index < 0 || group_index >= _NUM_ROBOT_GROUPS) return NULL;
	if(_robot_groups == NULL) {
		int i;
		_robot_groups = (struct _robot_group**)malloc(sizeof(struct _robot_group*) * _NUM_ROBOT_GROUPS);
		for(i = 0; i < _NUM_ROBOT_GROUPS; ++i) {
			_robot_groups[i] = NULL;
		}
	}
	group = _robot_groups[group_index];
	if(group == NULL) {
		group = (struct _robot_group*)malloc(sizeof(struct _robot_group));
		group->robots_size = 1;
		group->robots_count = 0;
		group->robots = (struct _robot**)malloc(sizeof(struct _robot*));
		group->robots[0] = NULL;
		_robot_groups[group_index] = group;
	}
	return group;
}

struct _robot* _robot_group_get_robot(int group_index, int robot_index) {
	struct _robot_group* group = _robot_group_get(group_index);
	if(group == NULL) return NULL;
	if(robot_index < 0 || robot_index >= group->robots_count) return NULL;
	return group->robots[robot_index];
}

void _robot_group_add_robot(int group_index, struct _robot* robot) {
	if(robot != NULL) {
		struct _robot_group* group = _robot_group_get(group_index);
		if(group != NULL) {
			if(group->robots_size <= group->robots_count) {
				int size = group->robots_size * 2, i;
				struct _robot** temp = (struct _robot**)malloc(sizeof(struct _robot*) * size);
				for(i = 0; i < size; ++i) {
					temp[i] = NULL;
				}
				memcpy(temp, group->robots, sizeof(struct _robot*) * group->robots_size);
				free(group->robots);
				group->robots = temp;
				group->robots_size = size;
			}
			group->robots[group->robots_count] = robot;
			group->robots_count ++;
		}
	}
}

void _robot_group_remove_robot(int group_index, struct _robot* robot) {
	if(robot != NULL) {
		struct _robot_group* group = _robot_group_get(group_index);
		if(group == NULL) return;
		if(robot->index < 0 || robot->index >= group->robots_size) return;
		group->robots[robot->index] = NULL;
	}
}

int _robot_group_count_robots(int group_index) {
	struct _robot_group* group = _robot_group_get(group_index);
	if(group == NULL) return 0;
	return group->robots_count;
}

const char* _robot_group_get_name(int group_index, int robot_index) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_get_name(robot);
}

void _robot_group_set_name(int group_index, int robot_index, const char* name) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	_robot_set_name(robot, name);
}

void _robot_group_reset(int group_index, int robot_index) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	if(robot != NULL) {
		robot->reset(robot);
	}
}

void _robot_group_dispose(int group_index, int robot_index) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	if(robot != NULL) {
		robot->alive = 0;
		robot->reset(robot);
		robot->dispose(robot);
	}
}

int _robot_group_e(int group_index, int robot_index, int device_id) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_e(robot, device_id);
}

int _robot_group_read(int group_index, int robot_index, int device_id) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read(robot, device_id);
}

int _robot_group_read_at(int group_index, int robot_index, int device_id, int index) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read_at(robot, device_id, index);
}

int _robot_group_read_array(int group_index, int robot_index, int device_id, int* data, int length) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read_array(robot, device_id, data, length);
}

float _robot_group_read_float(int group_index, int robot_index, int device_id) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read_float(robot, device_id);
}

float _robot_group_read_float_at(int group_index, int robot_index, int device_id, int index) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read_float_at(robot, device_id, index);
}

int _robot_group_read_float_array(int group_index, int robot_index, int device_id, float* data, int length) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_read_float_array(robot, device_id, data, length);
}

int _robot_group_write_empty(int group_index, int robot_index, int device_id) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_empty(robot, device_id);
}

int _robot_group_write(int group_index, int robot_index, int device_id, int data) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write(robot, device_id, data);
}

int _robot_group_write_at(int group_index, int robot_index, int device_id, int index, int data) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_at(robot, device_id, index, data);
}

int _robot_group_write_array(int group_index, int robot_index, int device_id, const int* data, int length) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_array(robot, device_id, data, length);
}

int _robot_group_write_float(int group_index, int robot_index, int device_id, float data) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_float(robot, device_id, data);
}

int _robot_group_write_float_at(int group_index, int robot_index, int device_id, int index, float data) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_float_at(robot, device_id, index, data);
}

int _robot_group_write_float_array(int group_index, int robot_index, int device_id, const float* data, int length) {
	struct _robot* robot = _robot_group_get_robot(group_index, robot_index);
	return _robot_write_float_array(robot, device_id, data, length);
}

/*------------------------------
  RUNNER
------------------------------*/

typedef void (*_EXECUTE)(void* arg);
typedef int (*_EVALUATE)(void* arg);

struct _runner {
	int started;
	int connection_required;
	int connection_checked;
	int robots_count;
	int robots_size;
	struct _robot** robots;
	_EXECUTE execute;
	void* execute_arg;
	_EVALUATE evaluate;
	void* evaluate_arg;
	int evaluate_result;
	int running;
	int thread_alive;
	HANDLE thread_handle;
};

struct _runner* _runner = NULL;

void _runner_create(void);
void _runner_shutdown(void);
void _runner_start(void);
void _runner_register_required(void);
void _runner_register_checked(void);
int _runner_is_all_checked(void);
void _runner_add_robot(struct _robot* robot);

void _runner_create(void) {
	if(_runner != NULL) return;
	_runner = (struct _runner*)malloc(sizeof(struct _runner));
	_runner->started = 0;
	_runner->connection_required = 0;
	_runner->connection_checked = 0;
	_runner->robots_count = 0;
	_runner->robots_size = 1;
	_runner->robots = (struct _robot**)malloc(sizeof(struct _robot*));
	_runner->robots[0] = NULL;
	_runner->execute = NULL;
	_runner->execute_arg = NULL;
	_runner->evaluate = NULL;
	_runner->evaluate_arg = NULL;
	_runner->evaluate_result = 0;
	_runner->running = 0;
	_runner->thread_alive = 0;
	_runner->thread_handle = 0;
}

void _runner_shutdown(void) {
	if(_runner != NULL) {
		struct _robot** robots = _runner->robots;
		int count = _runner->robots_count, i;
		struct _robot* robot;
		
		if(robots != NULL) {
			for(i = 0; i < count; ++i) {
				robot = robots[i];
				if(robot != NULL) {
					robot->alive = 0;
					robot->reset(robot);
				}
			}
		}
		
		_runner->running = 0;
		if(_runner->thread_alive == 1) {
			WaitForSingleObject(_runner->thread_handle, INFINITE);
		}
		
		if(robots != NULL) {
			for(i = 0; i < count; ++i) {
				robot = robots[i];
				if(robot != NULL) {
					robot->dispose(robot);
					robots[i] = NULL;
				}
			}
			free(robots);
			_runner->robots = NULL;
		}
		free(_runner);
		_runner = NULL;
	}
}

unsigned WINAPI _runner_thread_proc(void* arg) {
	struct _runner* runner = (struct _runner*)arg;
	struct _robot** robots;
	struct _robot* robot;
	int count, i;
	struct timeb time;
	double target_time;

	if(runner == NULL) return 0;
	
	runner->thread_alive = 1;
	target_time = _get_timestamp(&time);
	while(runner->running == 1) {
		if(_get_timestamp(&time) > target_time) {
			robots = runner->robots;
			count = runner->robots_count;
	
			for(i = 0; i < count; ++i) {
				robot = robots[i];
				if(robot != NULL && robot->alive == 1) {
					robot->update_sensory_device_state(robot);
				}
			}
			
			if(runner->evaluate != NULL) {
				runner->evaluate_result = runner->evaluate(runner->evaluate_arg);
				if(runner->evaluate_result == 1) {
					runner->evaluate = NULL;
				}
			}
			
			if(runner->execute != NULL) {
				runner->execute(runner->execute_arg);
			}
			
			for(i = 0; i < count; ++i) {
				robot = robots[i];
				if(robot != NULL && robot->alive == 1) {
					robot->request_motoring_data(robot);
				}
			}
			for(i = 0; i < count; ++i) {
				robot = robots[i];
				if(robot != NULL && robot->alive == 1) {
					robot->update_motoring_device_state(robot);
				}
			}
			target_time += 0.02;
		}
		Sleep(5);
	}
	runner->thread_alive = 0;
	return 0;
}

void _runner_signal_handler(int sig) {
	dispose_all();
	exit(1);
}

void _runner_start(void) {
	if(_runner == NULL) return;
	if(_runner->started == 0) {
		unsigned int thread_id;
		
		_runner->started = 1;
		_runner->running = 1;
		signal(SIGINT, _runner_signal_handler);
		_runner->thread_handle = (HANDLE)_beginthreadex(NULL,
			0,
			_runner_thread_proc,
			_runner,
			CREATE_SUSPENDED,
			&thread_id);
		if(_runner->thread_handle != 0) {
			ResumeThread(_runner->thread_handle);
		}
	}
}

void _runner_register_required(void) {
	if(_runner == NULL) {
		_runner_create();
	}
	_runner->connection_required ++;
}

void _runner_register_checked(void) {
	if(_runner == NULL) {
		_runner_create();
	}
	_runner->connection_checked ++;
}

int _runner_is_all_checked(void) {
	if(_runner == NULL) {
		_runner_create();
	}
	return (_runner->connection_checked >= _runner->connection_required) ? 1 : 0;
}

void _runner_add_robot(struct _robot* robot) {
	if(robot == NULL) return;
	if(_runner == NULL) {
		_runner_create();
	}
	if(_runner->robots_size <= _runner->robots_count) {
		int size = _runner->robots_size * 2, i;
		struct _robot** temp = (struct _robot**)malloc(sizeof(struct _robot*) * size);
		for(i = 0; i < size; ++i) {
			temp[i] = 0;
		}
		memcpy(temp, _runner->robots, sizeof(struct _robot*) * _runner->robots_size);
		free(_runner->robots);
		_runner->robots = temp;
		_runner->robots_size = size;
	}
	_runner->robots[_runner->robots_count] = robot;
	_runner->robots_count ++;
}

void _runner_remove_robot(struct _robot* robot) {
	struct _robot** robots;
	int count, i;
	
	if(robot == NULL) return;
	if(_runner == NULL) return;
	count = _runner->robots_count;
	robots = _runner->robots;
	for(i = 0; i < count; ++i) {
		if(robots[i] == robot) {
			robots[i] = NULL;
		}
	}
}

/*------------------------------
  COMMON
------------------------------*/

void scan(void) {
	int port_count = 0;
	char** port_names;
	
	printf("Serial ports:\n");

	port_names = _serial_window_get_serial_port_names(NULL, &port_count);
	if(port_count > 0 && port_names != NULL) {
		int i;
		for(i = 0; i < port_count; ++i) {
			printf("%s\n", port_names[i]);
			free(port_names[i]);
		}
		free(port_names);
	} else {
		printf("No available serial port\n");
	}
}

void set_executable(void (*execute)(void* arg), void* arg) {
	if(_runner == NULL) {
		_runner_create();
	}
	_runner->execute_arg = arg;
	_runner->execute = execute;
}

void wait(int milliseconds) {
	if(milliseconds > 0) {
		struct timeb time;
		double timeout;
		
		timeout = _get_timestamp(&time) + (double)milliseconds / 1000.0;
		while(1) {
			if(_get_timestamp(&time) >= timeout) {
				break;
			}
			Sleep(1);
		}
	}
}

void wait_until(int (*evaluate)(void* arg), void* arg) {
	if(_runner == NULL) {
		_runner_create();
	}
	_runner->evaluate_arg = arg;
	_runner->evaluate_result = 0;
	_runner->evaluate = evaluate;
	while(_runner->evaluate_result == 0) {
		Sleep(10);
	}
}

void wait_until_ready(void) {
	while(_runner_is_all_checked() == 0) {
		Sleep(10);
	}
}

void dispose_all(void) {
	_runner_shutdown();
	_robot_group_dispose_all();
}

/*------------------------------
  QUEUE
------------------------------*/

#define _ROBOID_QUEUE_SIZE 64
#define _ROBOID_QUEUE_MASK (_ROBOID_QUEUE_SIZE - 1)

struct _write_queue {
	char buffer[_ROBOID_QUEUE_SIZE];
	int output[19];
	int provider;
	int consumer;
};

void _write_queue_reset(struct _write_queue* queue) {
	if(queue == NULL) return;
	queue->provider = 0;
	queue->consumer = 0;
}

void _write_queue_push(struct _write_queue* queue, const char* text, int len, int line) {
	int i;

	if(queue == NULL || text == NULL) return;
	if(len > 0) {
		for(i = 0; i < len; ++i) {
			if(((queue->provider - queue->consumer) & _ROBOID_QUEUE_MASK) == _ROBOID_QUEUE_MASK) { // full
				queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
			}
			queue->buffer[queue->provider] = text[i];
			queue->provider = (queue->provider + 1) & _ROBOID_QUEUE_MASK;
		}
	}
	if(line == 1) {
		if(((queue->provider - queue->consumer) & _ROBOID_QUEUE_MASK) == _ROBOID_QUEUE_MASK) { // full
			queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
		}
		queue->buffer[queue->provider] = 0x0D;
		queue->provider = (queue->provider + 1) & _ROBOID_QUEUE_MASK;
	}
}

int* _write_queue_pop(struct _write_queue* queue) {
	int len, i;
	
	if(queue == NULL || queue->provider == queue->consumer) return NULL;
	len = (queue->provider - queue->consumer) & _ROBOID_QUEUE_MASK;
	if(len > 18) len = 18;
	
	queue->output[0] = len;
	i = 1;
	for(; i <= len && queue->consumer != queue->provider; ++i) {
		queue->output[i] = queue->buffer[queue->consumer];
		queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
	}
	for(; i <= 18; ++i) {
		queue->output[i] = 0;
	}
	return queue->output;
}

struct _read_queue {
	int buffer[_ROBOID_QUEUE_SIZE];
	int provider;
	int consumer;
};

void _read_queue_reset(struct _read_queue* queue) {
	if(queue == NULL) return;
	queue->provider = 0;
	queue->consumer = 0;
}

void _read_queue_push(struct _read_queue* queue, int* packet, int offset) {
	int len, i, end;
	
	if(queue == NULL || packet == NULL) return;
	len = packet[offset];
	if(len > 0) {
		if(len > 18) len = 18;
		
		end = len + offset;
		for(i = 1 + offset; i <= end; ++i) {
			if(((queue->provider - queue->consumer) & _ROBOID_QUEUE_MASK) == _ROBOID_QUEUE_MASK) { // full
				queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
			}
			queue->buffer[queue->provider] = packet[i];
			queue->provider = (queue->provider + 1) & _ROBOID_QUEUE_MASK;
		}
	}
}

int _read_queue_pop(struct _read_queue* queue, char* text, int len, char delimiter) {
	int index = 0, limit = len - 1, found = -1;
	
	if(queue == NULL || text == NULL || queue->provider == queue->consumer) return 0; // empty
	if(delimiter == 0) {
		while(queue->consumer != queue->provider && index < limit) {
			text[index++] = queue->buffer[queue->consumer];
			queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
		}
		text[index] = '\0';
	} else {
		while(queue->consumer != queue->provider) {
			if(queue->buffer[queue->consumer] == delimiter) {
				found = queue->consumer;
				break;
			}
			queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
		}
		if(found >= 0) {
			while(queue->consumer != found && index < limit) {
				text[index++] = queue->buffer[queue->consumer];
				queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
			}
			queue->consumer = (queue->consumer + 1) & _ROBOID_QUEUE_MASK;
			text[index] = '\0';
		}
	}
	return index;
}

/*------------------------------
  HAMSTER
------------------------------*/

#define _HAMSTER_LEFT_WHEEL_INDEX 0
#define _HAMSTER_RIGHT_WHEEL_INDEX 1
#define _HAMSTER_BUZZER_INDEX 2
#define _HAMSTER_OUTPUT_A_INDEX 3
#define _HAMSTER_OUTPUT_B_INDEX 4
#define _HAMSTER_TOPOLOGY_INDEX 5
#define _HAMSTER_LEFT_LED_INDEX 6
#define _HAMSTER_RIGHT_LED_INDEX 7
#define _HAMSTER_NOTE_INDEX 8
#define _HAMSTER_LINE_TRACER_MODE_INDEX 9
#define _HAMSTER_LINE_TRACER_SPEED_INDEX 10
#define _HAMSTER_IO_MODE_A_INDEX 11
#define _HAMSTER_IO_MODE_B_INDEX 12
#define _HAMSTER_CONFIG_PROXIMITY_INDEX 13
#define _HAMSTER_CONFIG_GRAVITY_INDEX 14
#define _HAMSTER_CONFIG_BAND_WIDTH_INDEX 15
#define _HAMSTER_SIGNAL_STRENGTH_INDEX 16
#define _HAMSTER_LEFT_PROXIMITY_INDEX 17
#define _HAMSTER_RIGHT_PROXIMITY_INDEX 18
#define _HAMSTER_LEFT_FLOOR_INDEX 19
#define _HAMSTER_RIGHT_FLOOR_INDEX 20
#define _HAMSTER_ACCELERATION_INDEX 21
#define _HAMSTER_LIGHT_INDEX 22
#define _HAMSTER_TEMPERATURE_INDEX 23
#define _HAMSTER_INPUT_A_INDEX 24
#define _HAMSTER_INPUT_B_INDEX 25
#define _HAMSTER_LINE_TRACER_STATE_INDEX 26
#define _HAMSTER_TILT_INDEX 27
#define _HAMSTER_BATTERY_STATE_INDEX 28

#define _GROUP_HAMSTER 0

#define _HAMSTER_LED_OFF 0
#define _HAMSTER_LED_BLUE 1
#define _HAMSTER_LED_GREEN 2
#define _HAMSTER_LED_SKY_BLUE 3
#define _HAMSTER_LED_RED 4
#define _HAMSTER_LED_PURPLE 5
#define _HAMSTER_LED_YELLOW 6
#define _HAMSTER_LED_WHITE 7

#define _HAMSTER_IO_MODE_ANALOG_INPUT 0
#define _HAMSTER_IO_MODE_DIGITAL_INPUT 1
#define _HAMSTER_IO_MODE_SERVO_OUTPUT 8
#define _HAMSTER_IO_MODE_PWM_OUTPUT 9
#define _HAMSTER_IO_MODE_DIGITAL_OUTPUT 10

struct _hamster_robot {
	struct _robot robot;
	int left_wheel;
	int right_wheel;
	float buzzer;
	int output_a;
	int output_b;
	int topology;
	int left_led;
	int right_led;
	int note;
	int line_tracer_mode_written;
	int line_tracer_mode;
	int line_tracer_speed;
	int io_mode_a;
	int io_mode_b;
	int config_proximity;
	int config_gravity;
	int config_band_width;
	int light;
	int temperature;
	int line_tracer_flag;
	int line_tracer_event;
	int line_tracer_state;
	int line_tracer_count;
	int event_tilt;
	int event_battery_state;
	double bpm;
	int board_count;
	int board_state;
};

int _hamster_check_connection(struct _connector* connector, struct _serial* serial) {
	int i, read_bytes;
	struct _robot* robot;

	for(i = 0; i < _RETRY; ++i) {
		_serial_write(serial, "FF\r", 3);
		read_bytes = _connector_read_packet(connector, serial, "FF");
		if(read_bytes != 0) {
			char tmp[5][256];
			int index = 0, len;
			char* buf;
			char* comma;

			buf = connector->buffer;
			comma = strchr(buf, ',');
			while(comma != NULL && index < 5) {
				len = comma - buf;
				_STRNCPY(tmp[index], 256, buf, len);
				tmp[index][len] = '\0';
				index ++;
				buf = comma + 1;
				comma = strchr(buf, ',');
			}
			if(index >= 4 && buf != NULL && strlen(buf) >= 12) {
				robot = (struct _robot*)connector->arg;
				if(_STRICMP(tmp[1], "Hamster") == 0 && _STRICMP(tmp[2], "04") == 0) {
					buf[12] = '\0';
					robot->model_code = 0x04;
					_connector_set_address(connector, buf);
					_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTED);
					return _CONNECTION_RESULT_FOUND;
				} else if(_STRICMP(tmp[2], "0E") == 0) {
					buf[12] = '\0';
					robot->model_code = 0x0E;
					_connector_set_address(connector, buf);
					_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTED);
					return _CONNECTION_RESULT_FOUND;
				}
			}
			return _CONNECTION_RESULT_NOT_AVAILABLE;
		}
	}
	return _CONNECTION_RESULT_NOT_AVAILABLE;
}

struct _device* _hamster_find_device(struct _robot* robot, int device_id) {
	if(robot == NULL) return NULL;
	if(robot->devices == NULL) return NULL;
	switch(device_id) {
		case HAMSTER_LEFT_WHEEL: return robot->devices[_HAMSTER_LEFT_WHEEL_INDEX];
		case HAMSTER_RIGHT_WHEEL: return robot->devices[_HAMSTER_RIGHT_WHEEL_INDEX];
		case HAMSTER_BUZZER: return robot->devices[_HAMSTER_BUZZER_INDEX];
		case HAMSTER_OUTPUT_A: return robot->devices[_HAMSTER_OUTPUT_A_INDEX];
		case HAMSTER_OUTPUT_B: return robot->devices[_HAMSTER_OUTPUT_B_INDEX];
		case HAMSTER_TOPOLOGY: return robot->devices[_HAMSTER_TOPOLOGY_INDEX];
		case HAMSTER_LEFT_LED: return robot->devices[_HAMSTER_LEFT_LED_INDEX];
		case HAMSTER_RIGHT_LED: return robot->devices[_HAMSTER_RIGHT_LED_INDEX];
		case HAMSTER_NOTE: return robot->devices[_HAMSTER_NOTE_INDEX];
		case HAMSTER_LINE_TRACER_MODE: return robot->devices[_HAMSTER_LINE_TRACER_MODE_INDEX];
		case HAMSTER_LINE_TRACER_SPEED: return robot->devices[_HAMSTER_LINE_TRACER_SPEED_INDEX];
		case HAMSTER_IO_MODE_A: return robot->devices[_HAMSTER_IO_MODE_A_INDEX];
		case HAMSTER_IO_MODE_B: return robot->devices[_HAMSTER_IO_MODE_B_INDEX];
		case HAMSTER_CONFIG_PROXIMITY: return robot->devices[_HAMSTER_CONFIG_PROXIMITY_INDEX];
		case HAMSTER_CONFIG_GRAVITY: return robot->devices[_HAMSTER_CONFIG_GRAVITY_INDEX];
		case HAMSTER_CONFIG_BAND_WIDTH: return robot->devices[_HAMSTER_CONFIG_BAND_WIDTH_INDEX];
		case HAMSTER_SIGNAL_STRENGTH: return robot->devices[_HAMSTER_SIGNAL_STRENGTH_INDEX];
		case HAMSTER_LEFT_PROXIMITY: return robot->devices[_HAMSTER_LEFT_PROXIMITY_INDEX];
		case HAMSTER_RIGHT_PROXIMITY: return robot->devices[_HAMSTER_RIGHT_PROXIMITY_INDEX];
		case HAMSTER_LEFT_FLOOR: return robot->devices[_HAMSTER_LEFT_FLOOR_INDEX];
		case HAMSTER_RIGHT_FLOOR: return robot->devices[_HAMSTER_RIGHT_FLOOR_INDEX];
		case HAMSTER_ACCELERATION: return robot->devices[_HAMSTER_ACCELERATION_INDEX];
		case HAMSTER_LIGHT: return robot->devices[_HAMSTER_LIGHT_INDEX];
		case HAMSTER_TEMPERATURE: return robot->devices[_HAMSTER_TEMPERATURE_INDEX];
		case HAMSTER_INPUT_A: return robot->devices[_HAMSTER_INPUT_A_INDEX];
		case HAMSTER_INPUT_B: return robot->devices[_HAMSTER_INPUT_B_INDEX];
		case HAMSTER_LINE_TRACER_STATE: return robot->devices[_HAMSTER_LINE_TRACER_STATE_INDEX];
		case HAMSTER_TILT: return robot->devices[_HAMSTER_TILT_INDEX];
		case HAMSTER_BATTERY_STATE: return robot->devices[_HAMSTER_BATTERY_STATE_INDEX];
	}
	return NULL;
}

void _hamster_request_motoring_data(struct _robot* robot) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	struct _device** devices = robot->devices;
	
	hamster->left_wheel = _device_read(devices[_HAMSTER_LEFT_WHEEL_INDEX]);
	hamster->right_wheel = _device_read(devices[_HAMSTER_RIGHT_WHEEL_INDEX]);
	hamster->buzzer = _device_read_float(devices[_HAMSTER_BUZZER_INDEX]);
	hamster->output_a = _device_read(devices[_HAMSTER_OUTPUT_A_INDEX]);
	hamster->output_b = _device_read(devices[_HAMSTER_OUTPUT_B_INDEX]);
	if(devices[_HAMSTER_TOPOLOGY_INDEX]->written == 1) {
		hamster->topology = _device_read(devices[_HAMSTER_TOPOLOGY_INDEX]);
	}
	if(devices[_HAMSTER_LEFT_LED_INDEX]->written == 1) {
		hamster->left_led = _device_read(devices[_HAMSTER_LEFT_LED_INDEX]);
	}
	if(devices[_HAMSTER_RIGHT_LED_INDEX]->written == 1) {
		hamster->right_led = _device_read(devices[_HAMSTER_RIGHT_LED_INDEX]);
	}
	if(devices[_HAMSTER_NOTE_INDEX]->written == 1) {
		hamster->note = _device_read(devices[_HAMSTER_NOTE_INDEX]);
	}
	if(devices[_HAMSTER_LINE_TRACER_MODE_INDEX]->written == 1) {
		hamster->line_tracer_mode = _device_read(devices[_HAMSTER_LINE_TRACER_MODE_INDEX]);
		hamster->line_tracer_mode_written = 1;
	}
	if(devices[_HAMSTER_LINE_TRACER_SPEED_INDEX]->written == 1) {
		hamster->line_tracer_speed = _device_read(devices[_HAMSTER_LINE_TRACER_SPEED_INDEX]);
	}
	if(devices[_HAMSTER_IO_MODE_A_INDEX]->written == 1) {
		hamster->io_mode_a = _device_read(devices[_HAMSTER_IO_MODE_A_INDEX]);
	}
	if(devices[_HAMSTER_IO_MODE_B_INDEX]->written == 1) {
		hamster->io_mode_b = _device_read(devices[_HAMSTER_IO_MODE_B_INDEX]);
	}
	if(devices[_HAMSTER_CONFIG_PROXIMITY_INDEX]->written == 1) {
		hamster->config_proximity = _device_read(devices[_HAMSTER_CONFIG_PROXIMITY_INDEX]);
	}
	if(devices[_HAMSTER_CONFIG_GRAVITY_INDEX]->written == 1) {
		hamster->config_gravity = _device_read(devices[_HAMSTER_CONFIG_GRAVITY_INDEX]);
	}
	if(devices[_HAMSTER_CONFIG_BAND_WIDTH_INDEX]->written == 1) {
		hamster->config_band_width = _device_read(devices[_HAMSTER_CONFIG_BAND_WIDTH_INDEX]);
	}
	_robot_clear_written(robot);
}

int _HAMSTER_COLOR_TO_RGB[][3] = {
	{ 0, 0, 0 },
	{ 0, 0, 255 },
	{ 0, 255, 0 },
	{ 0, 255, 255 },
	{ 255, 0, 0 },
	{ 255, 0, 255 },
	{ 255, 255, 0 },
	{ 255, 255, 255 }
};

int* _hamster_color_to_rgb(int color) {
	if(color > 7) color = 7;
	else if(color < 0) color = 0;
	return _HAMSTER_COLOR_TO_RGB[color];
}

int _hamster_speed_to_gain(int speed) {
	switch(speed) {
		case 1:
		case 2: return 6;
		case 3:
		case 4: return 5;
		case 5:
		case 6: return 4;
		case 7:
		case 8: return 3;
		case 9:
		case 10: return 2;
	}
	if(speed < 1) return 6;
	else return 2;
}

void _hamster_encode_motoring_packet(struct _robot* robot) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	char* buffer = robot->write_buffer;
	int index = 0, temp, i;
	int* rgb;
	const char* address;

	if(robot->model_code == 0x0E) {
		buffer[index++] = '1';
		buffer[index++] = '0';
		index = _value_to_hex(buffer, index, hamster->left_wheel, 1);
		index = _value_to_hex(buffer, index, hamster->right_wheel, 1);
		rgb = _hamster_color_to_rgb(hamster->left_led);
		index = _value_to_hex(buffer, index, rgb[0], 1);
		index = _value_to_hex(buffer, index, rgb[1], 1);
		index = _value_to_hex(buffer, index, rgb[2], 1);
		rgb = _hamster_color_to_rgb(hamster->right_led);
		index = _value_to_hex(buffer, index, rgb[0], 1);
		index = _value_to_hex(buffer, index, rgb[1], 1);
		index = _value_to_hex(buffer, index, rgb[2], 1);
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		temp = hamster->line_tracer_mode & 0x0f;
		if(temp > 7) temp ++;
		if(hamster->line_tracer_mode_written == 1) {
			hamster->line_tracer_count = 0;
			if(temp > 0) {
				hamster->line_tracer_flag = (hamster->line_tracer_flag % 15) + 1;
				hamster->line_tracer_event = 1;
			} else {
				hamster->line_tracer_event = 0;
			}
			hamster->line_tracer_mode_written = 0;
		}
		temp |= (hamster->line_tracer_flag & 0x0f) << 4;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->line_tracer_speed & 0x0f) << 4;
		temp |= _hamster_speed_to_gain(hamster->line_tracer_speed) & 0x0f;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->config_proximity & 0x07) << 5;
		temp |= (hamster->config_band_width & 0x07) << 2;
		temp |= hamster->config_gravity & 0x03;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->io_mode_a & 0x0f) << 4;
		temp |= hamster->io_mode_b & 0x0f;
		index = _value_to_hex(buffer, index, temp, 1);
		index = _value_to_hex(buffer, index, hamster->output_a, 1);
		index = _value_to_hex(buffer, index, hamster->output_b, 1);
		if(hamster->note > 0) {
			buffer[index++] = '0';
			buffer[index++] = '1';
			index = _value_to_hex(buffer, index, hamster->note, 1);
		} else {
			index = _value_to_hex(buffer, index, _round_float_value(hamster->buzzer * 10) + 512, 2);
		}
	} else {
		index = _value_to_hex(buffer, index, hamster->topology & 0x0f, 1);
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '1';
		buffer[index++] = '0';
		index = _value_to_hex(buffer, index, hamster->left_wheel, 1);
		index = _value_to_hex(buffer, index, hamster->right_wheel, 1);
		index = _value_to_hex(buffer, index, hamster->left_led, 1);
		index = _value_to_hex(buffer, index, hamster->right_led, 1);
		index = _value_to_hex(buffer, index, _round_float_value(hamster->buzzer * 100), 3);
		index = _value_to_hex(buffer, index, hamster->note, 1);
		if(hamster->line_tracer_mode_written == 1) {
			if(hamster->line_tracer_mode > 0) {
				hamster->line_tracer_flag ^= 0x80;
				hamster->line_tracer_event = 1;
			}
			hamster->line_tracer_mode_written = 0;
		}
		temp = (hamster->line_tracer_mode & 0x0f) << 3;
		temp |= (hamster->line_tracer_speed - 1) & 0x07;
		temp |= hamster->line_tracer_flag & 0x80;
		index = _value_to_hex(buffer, index, temp, 1);
		index = _value_to_hex(buffer, index, hamster->config_proximity, 1);
		temp = (hamster->config_gravity & 0x0f) << 4;
		temp |= hamster->config_band_width & 0x0f;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->io_mode_a & 0x0f) << 4;
		temp |= hamster->io_mode_b & 0x0f;
		index = _value_to_hex(buffer, index, temp, 1);
		index = _value_to_hex(buffer, index, hamster->output_a, 1);
		index = _value_to_hex(buffer, index, hamster->output_b, 1);
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
		buffer[index++] = '0';
	}
	buffer[index++] = '-';
	address = _connector_get_address(robot->connector);
	for(i = 0; i < 12; ++i) {
		buffer[index++] = address[i];
	}
	buffer[index++] = '\r';
	_connector_write(robot->connector, buffer, _MOTORING_PACKET_LENGTH);
}

int _hamster_decode_sensory_packet(struct _robot* robot, char* packet) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	struct _device** devices = robot->devices;
	char* buffer = packet;
	int value, value2, acc_x, acc_y, acc_z;
	
	if(robot->model_code == 0x0E) {
		value = _hex_to_value(buffer, 0, 1);
		if(value != 1) return 0;
		_device_put(devices[_HAMSTER_LEFT_PROXIMITY_INDEX], _hex_to_value(buffer, 6, 8), 1); // left proximity
		_device_put(devices[_HAMSTER_RIGHT_PROXIMITY_INDEX], _hex_to_value(buffer, 8, 10), 1); // right proximity
		value2 = _hex_to_value(buffer, 38, 40);
		if((value2 & 0x01) == 0) {
			hamster->light = _hex_to_value(buffer, 10, 14);
		} else {
			value = _hex_to_value(buffer, 10, 12);
			if(value > 0x7F) value -= 0x100;
			hamster->temperature = _round_float_value(value / 2.0f + 23);
		}
		_device_put(devices[_HAMSTER_LIGHT_INDEX], hamster->light, 1);
		_device_put(devices[_HAMSTER_TEMPERATURE_INDEX], hamster->temperature, 1);
		_device_put(devices[_HAMSTER_LEFT_FLOOR_INDEX], _hex_to_value(buffer, 14, 16), 1); // left floor
		_device_put(devices[_HAMSTER_RIGHT_FLOOR_INDEX], _hex_to_value(buffer, 16, 18), 1); // right floor
		acc_x = _hex_to_value(buffer, 18, 22);
		if(acc_x > 0x7FFF) acc_x -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 0, acc_x, 1);
		acc_y = _hex_to_value(buffer, 22, 26);
		if(acc_y > 0x7FFF) acc_y -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 1, acc_y, 1);
		acc_z = _hex_to_value(buffer, 26, 30);
		if(acc_z > 0x7FFF) acc_z -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 2, acc_z, 1);
		if(acc_z < 8192 && acc_x > 8192 && acc_y > -4096 && acc_y < 4096) value = 1;
		else if(acc_z < 8192 && acc_x < -8192 && acc_y > -4096 && acc_y < 4096) value = -1;
		else if(acc_z < 8192 && acc_y > 8192 && acc_x > -4096 && acc_x < 4096) value = 2;
		else if(acc_z < 8192 && acc_y < -8192 && acc_x > -4096 && acc_x < 4096) value = -2;
		else if(acc_z > 12288 && acc_x > -8192 && acc_x < 8192 && acc_y > -8192 && acc_y < 8192) value = 3;
		else if(acc_z < -12288 && acc_x > -4096 && acc_x < 4096 && acc_y > -4096 && acc_y < 4096) value = -3;
		else value = 0;
		if(value != hamster->event_tilt) {
			_device_put(devices[_HAMSTER_TILT_INDEX], value, (hamster->event_tilt == -4) ? 0 : 1);
			hamster->event_tilt = value;
		}
		_device_put(devices[_HAMSTER_INPUT_A_INDEX], _hex_to_value(buffer, 30, 32), 1);
		_device_put(devices[_HAMSTER_INPUT_B_INDEX], _hex_to_value(buffer, 32, 34), 1);
		value = _hex_to_value(buffer, 36, 38);
		value -= 0x100;
		_device_put(devices[_HAMSTER_SIGNAL_STRENGTH_INDEX], value, 1);
		value = (value2 >> 6) & 0x03;
		if((value & 0x02) != 0) {
			if(hamster->line_tracer_event == 1) {
				if(value == 0x02) {
					if(++hamster->line_tracer_count > 5) hamster->line_tracer_event = 2;
				} else {
					hamster->line_tracer_event = 2;
				}
			}
			if(hamster->line_tracer_event == 2) {
				if(value != hamster->line_tracer_state || hamster->line_tracer_count > 5) {
					hamster->line_tracer_state = value;
					_device_put(devices[_HAMSTER_LINE_TRACER_STATE_INDEX], value << 5, 1);
					if(value == 0x02) {
						hamster->line_tracer_event = 0;
						hamster->line_tracer_count = 0;
					}
				}
			}
		}
		value = (value2 >> 1) & 0x03;
		if(value == 0) value = 2;
		else if(value >= 2) value = 0;
		if(value != hamster->event_battery_state) {
			_device_put(devices[_HAMSTER_BATTERY_STATE_INDEX], value, (hamster->event_battery_state == -1) ? 0 : 1);
			hamster->event_battery_state = value;
		}
	} else {
		value = _hex_to_value(buffer, 4, 5);
		if(value != 1) return 0;
		value = _hex_to_value(buffer, 6, 8);
		value -= 0x100;
		_device_put(devices[_HAMSTER_SIGNAL_STRENGTH_INDEX], value, 1);
		_device_put(devices[_HAMSTER_LEFT_PROXIMITY_INDEX], _hex_to_value(buffer, 8, 10), 1); // left proximity
		_device_put(devices[_HAMSTER_RIGHT_PROXIMITY_INDEX], _hex_to_value(buffer, 10, 12), 1); // right proximity
		_device_put(devices[_HAMSTER_LEFT_FLOOR_INDEX], _hex_to_value(buffer, 12, 14), 1); // left floor
		_device_put(devices[_HAMSTER_RIGHT_FLOOR_INDEX], _hex_to_value(buffer, 14, 16), 1); // right floor
		acc_x = _hex_to_value(buffer, 16, 20);
		if(acc_x > 0x7FFF) acc_x -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 0, acc_x, 1);
		acc_y = _hex_to_value(buffer, 20, 24);
		if(acc_y > 0x7FFF) acc_y -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 1, acc_y, 1);
		acc_z = _hex_to_value(buffer, 24, 28);
		if(acc_z > 0x7FFF) acc_z -= 0x10000;
		_device_put_at(devices[_HAMSTER_ACCELERATION_INDEX], 2, acc_z, 1);
		if(acc_z < 8192 && acc_x > 8192 && acc_y > -4096 && acc_y < 4096) value = 1;
		else if(acc_z < 8192 && acc_x < -8192 && acc_y > -4096 && acc_y < 4096) value = -1;
		else if(acc_z < 8192 && acc_y > 8192 && acc_x > -4096 && acc_x < 4096) value = 2;
		else if(acc_z < 8192 && acc_y < -8192 && acc_x > -4096 && acc_x < 4096) value = -2;
		else if(acc_z > 12288 && acc_x > -8192 && acc_x < 8192 && acc_y > -8192 && acc_y < 8192) value = 3;
		else if(acc_z < -12288 && acc_x > -4096 && acc_x < 4096 && acc_y > -4096 && acc_y < 4096) value = -3;
		else value = 0;
		if(value != hamster->event_tilt) {
			_device_put(devices[_HAMSTER_TILT_INDEX], value, (hamster->event_tilt == -4) ? 0 : 1);
			hamster->event_tilt = value;
		}
		value = _hex_to_value(buffer, 28, 30);
		if(value == 0) {
			hamster->light = _hex_to_value(buffer, 30, 34);
		} else {
			value = _hex_to_value(buffer, 30, 32);
			if(value > 0x7F) value -= 0x100;
			hamster->temperature = _round_float_value(value / 2.0f + 24);
			value = _hex_to_value(buffer, 32, 34) + 200;
			if(value < 360) value = 0;
			else if(value <= 370) value = 1;
			else value = 2;
			if(value != hamster->event_battery_state) {
				_device_put(devices[_HAMSTER_BATTERY_STATE_INDEX], value, (hamster->event_battery_state == -1) ? 0 : 1);
				hamster->event_battery_state = value;
			}
		}
		_device_put(devices[_HAMSTER_LIGHT_INDEX], hamster->light, 1);
		_device_put(devices[_HAMSTER_TEMPERATURE_INDEX], hamster->temperature, 1);
		_device_put(devices[_HAMSTER_INPUT_A_INDEX], _hex_to_value(buffer, 34, 36), 1);
		_device_put(devices[_HAMSTER_INPUT_B_INDEX], _hex_to_value(buffer, 36, 38), 1);
		value = _hex_to_value(buffer, 38, 40);
		if((value & 0x40) != 0) {
			if(hamster->line_tracer_event == 1) {
				if(value != 0x40) {
					hamster->line_tracer_event = 2;
				}
			}
			if(hamster->line_tracer_event == 2) {
				if(value != hamster->line_tracer_state) {
					hamster->line_tracer_state = value;
					_device_put(devices[_HAMSTER_LINE_TRACER_STATE_INDEX], value, 1);
					if(value == 0x40) {
						hamster->line_tracer_event = 0;
					}
				}
			}
		}
	}
	return 1;
}

int _hamster_receive(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		int read_bytes = _connector_read(connector);
		if(read_bytes != 0) {
			if(_hamster_decode_sensory_packet(robot, connector->buffer) == 1) {
				if(robot->ready == 0) {
					robot->ready = 1;
					_runner_register_checked();
				}
			}
			return 1;
		}
	}
	return 0;
}

void _hamster_send(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		_hamster_encode_motoring_packet(robot);
	}
}

void _hamster_update_sensory_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _HAMSTER_SIGNAL_STRENGTH_INDEX; i <= _HAMSTER_BATTERY_STATE_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _hamster_update_motoring_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _HAMSTER_LEFT_WHEEL_INDEX; i <= _HAMSTER_CONFIG_BAND_WIDTH_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _hamster_reset(struct _robot* robot) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_reset(robot);
	
	hamster->left_wheel = 0;
	hamster->right_wheel = 0;
	hamster->buzzer = 0;
	hamster->output_a = 0;
	hamster->output_b = 0;
	hamster->topology = 0;
	hamster->left_led = 0;
	hamster->right_led = 0;
	hamster->note = 0;
	hamster->line_tracer_mode_written = 0;
	hamster->line_tracer_mode = 0;
	hamster->line_tracer_speed = 5;
	hamster->io_mode_a = 0;
	hamster->io_mode_b = 0;
	hamster->config_proximity = 2;
	hamster->config_gravity = 0;
	hamster->config_band_width = 3;
	
	hamster->light = 0;
	hamster->temperature = 0;
	
//	hamster->line_tracer_flag = 0;
	hamster->line_tracer_event = 0;
	hamster->line_tracer_state = 0;
	hamster->line_tracer_count = 0;
	
	hamster->event_tilt = -4;
	hamster->event_battery_state = -1;
	
	hamster->bpm = 60;
	hamster->board_count = 0;
	hamster->board_state = 0;
}

void _hamster_dispose(struct _robot* robot) {
	_robot_group_remove_robot(_GROUP_HAMSTER, robot);
	_runner_remove_robot(robot);

	robot->running = 0;
	if(robot->thread_alive == 1) {
		WaitForSingleObject(robot->thread_handle, 1000);
	}
	Sleep(100);
	_robot_dispose(robot);
	free((struct _hamster_robot*)robot);
}

unsigned WINAPI _hamster_thread_proc(void* arg) {
	struct _robot* robot = (struct _robot*)arg;
	int shutdown = 0, shutdown_count = 0;
	
	if(robot == NULL) return 0;

	robot->thread_alive = 1;
	while(1) {
		if(_hamster_receive(robot) == 1) {
			_hamster_send(robot);
			if(shutdown == 1) {
				break;
			}
		}
		if(shutdown == 1) {
			if(++shutdown_count > 5) {
				break;
			}
		}
		if(robot->running == 0) {
			shutdown = 1;
		}
		Sleep(5);
	}
	robot->thread_alive = 0;
	return 0;
}

struct _hamster_robot* _hamster_create(const char* port_name) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)malloc(sizeof(struct _hamster_robot));
	struct _robot* robot = (struct _robot*)hamster;
	struct _device** devices;
	struct _connector* connector;
	unsigned int thread_id;
	
	_robot_init(robot, _robot_group_count_robots(_GROUP_HAMSTER), "Hamster", 55);
	
	hamster->left_wheel = 0;
	hamster->right_wheel = 0;
	hamster->buzzer = 0;
	hamster->output_a = 0;
	hamster->output_b = 0;
	hamster->topology = 0;
	hamster->left_led = 0;
	hamster->right_led = 0;
	hamster->note = 0;
	hamster->line_tracer_mode_written = 0;
	hamster->line_tracer_mode = 0;
	hamster->line_tracer_speed = 5;
	hamster->io_mode_a = 0;
	hamster->io_mode_b = 0;
	hamster->config_proximity = 2;
	hamster->config_gravity = 0;
	hamster->config_band_width = 3;
	
	hamster->light = 0;
	hamster->temperature = 0;
	
	hamster->line_tracer_flag = 0;
	hamster->line_tracer_event = 0;
	hamster->line_tracer_state = 0;
	hamster->line_tracer_count = 0;
	
	hamster->event_tilt = -4;
	hamster->event_battery_state = -1;
	
	hamster->bpm = 60;
	hamster->board_count = 0;
	hamster->board_state = 0;
	
	robot->find_device = _hamster_find_device;
	robot->request_motoring_data = _hamster_request_motoring_data;
	robot->update_sensory_device_state = _hamster_update_sensory_device_state;
	robot->update_motoring_device_state = _hamster_update_motoring_device_state;
	robot->reset = _hamster_reset;
	robot->dispose = _hamster_dispose;
	
	robot->devices_size = 29;
	devices = (struct _device**)malloc(sizeof(struct _device*) * 29);
	devices[0] = _device_create(HAMSTER_LEFT_WHEEL, "LeftWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, -100, 100, 0);
	devices[1] = _device_create(HAMSTER_RIGHT_WHEEL, "RightWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, -100, 100, 0);
	devices[2] = _device_create(HAMSTER_BUZZER, "Buzzer", DEVICE_TYPE_EFFECTOR, DATA_TYPE_FLOAT, 1, 0, 167772.15f, 0);
	devices[3] = _device_create(HAMSTER_OUTPUT_A, "OutputA", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[4] = _device_create(HAMSTER_OUTPUT_B, "OutputB", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[5] = _device_create(HAMSTER_TOPOLOGY, "Topology", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 15, 0);
	devices[6] = _device_create(HAMSTER_LEFT_LED, "LeftLed", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 7, 0);
	devices[7] = _device_create(HAMSTER_RIGHT_LED, "RightLed", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 7, 0);
	devices[8] = _device_create(HAMSTER_NOTE, "Note", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 88, 0);
	devices[9] = _device_create(HAMSTER_LINE_TRACER_MODE, "LineTracerMode", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 15, 0);
	devices[10] = _device_create(HAMSTER_LINE_TRACER_SPEED, "LineTracerSpeed", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 8, 5);
	devices[11] = _device_create(HAMSTER_IO_MODE_A, "IoModeA", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 15, 0);
	devices[12] = _device_create(HAMSTER_IO_MODE_B, "IoModeB", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 15, 0);
	devices[13] = _device_create(HAMSTER_CONFIG_PROXIMITY, "ConfigProximity", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 7, 2);
	devices[14] = _device_create(HAMSTER_CONFIG_GRAVITY, "ConfigGravity", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[15] = _device_create(HAMSTER_CONFIG_BAND_WIDTH, "ConfigBandWidth", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 8, 3);
	devices[16] = _device_create(HAMSTER_SIGNAL_STRENGTH, "SignalStrength", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -128, 0, 0);
	devices[17] = _device_create(HAMSTER_LEFT_PROXIMITY, "LeftProximity", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[18] = _device_create(HAMSTER_RIGHT_PROXIMITY, "RightProximity", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[19] = _device_create(HAMSTER_LEFT_FLOOR, "LeftFloor", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[20] = _device_create(HAMSTER_RIGHT_FLOOR, "RightFloor", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[21] = _device_create(HAMSTER_ACCELERATION, "Acceleration", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 3, -32768, 32767, 0);
	devices[22] = _device_create(HAMSTER_LIGHT, "Light", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[23] = _device_create(HAMSTER_TEMPERATURE, "Temperature", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -40, 88, 0);
	devices[24] = _device_create(HAMSTER_INPUT_A, "inputA", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[25] = _device_create(HAMSTER_INPUT_B, "inputB", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[26] = _device_create(HAMSTER_LINE_TRACER_STATE, "LineTracerState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[27] = _device_create(HAMSTER_TILT, "Tilt", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, -3, 3, 0);
	devices[28] = _device_create(HAMSTER_BATTERY_STATE, "BatteryState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 2, 2);
	robot->devices = devices;
	robot->connector = NULL;
	robot->alive = 1;

	_robot_group_add_robot(_GROUP_HAMSTER, robot);
	_runner_add_robot(robot);
	_runner_start();
	
	_runner_register_required();
	robot->running = 1;
	robot->thread_handle = (HANDLE)_beginthreadex(NULL,
		0,
		_hamster_thread_proc,
		robot,
		CREATE_SUSPENDED,
		&thread_id);
	if(robot->thread_handle != 0) {
		ResumeThread(robot->thread_handle);
	}

	connector = _connector_create("Hamster", robot->index, _VALID_PACKET_LENGTH, _CR);
	robot->connector = connector;
	if(connector != NULL) {
		int result;
		connector->arg = (void*)robot;
		connector->check_connection = _hamster_check_connection;
		result = _connector_open(connector, port_name, _SERIAL_BAUDRATE_115200, _SERIAL_FLOWCONTROL_RTSCTS_IN | _SERIAL_FLOWCONTROL_RTSCTS_OUT);
		if(result == _CONNECTION_RESULT_FOUND) {
			while(robot->ready == 0 && robot->running == 1) {
				Sleep(10);
			}
		} else if(result == _CONNECTION_RESULT_NOT_AVAILABLE) {
			_runner_register_checked();
		}
	}
	return hamster;
}

const char* _hamster_get_id(void) {
	return HAMSTER_ID;
}

void _hamster_wheels(int hamster_index, double left_velocity, double right_velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, _round_double_value(left_velocity));
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, _round_double_value(right_velocity));
}

void _hamster_left_wheel(int hamster_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, _round_double_value(velocity));
}

void _hamster_right_wheel(int hamster_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, _round_double_value(velocity));
}

void _hamster_stop(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
}

void _hamster_motion_sec(int hamster_index, double sec, double left_velocity, double right_velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int vel;
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	if(sec < 0) {
		sec = -sec;
		left_velocity = -left_velocity;
		right_velocity = -right_velocity;
	}
	if(sec > 0) {
		_robot_write(robot, HAMSTER_LEFT_WHEEL, _round_double_value(left_velocity));
		_robot_write(robot, HAMSTER_RIGHT_WHEEL, _round_double_value(right_velocity));
		wait(_round_double_value(sec * 1000));
	}
	_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
}

void _hamster_move_forward(int hamster_index, double sec, double velocity) {
	_hamster_motion_sec(hamster_index, sec, velocity, velocity);
}

void _hamster_move_backward(int hamster_index, double sec, double velocity) {
	_hamster_motion_sec(hamster_index, sec, -velocity, -velocity);
}

void _hamster_turn_left(int hamster_index, double sec, double velocity) {
	_hamster_motion_sec(hamster_index, sec, -velocity, velocity);
}

void _hamster_turn_right(int hamster_index, double sec, double velocity) {
	_hamster_motion_sec(hamster_index, sec, velocity, -velocity);
}

int _hamster_line_tracer_mode_callback(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, HAMSTER_LINE_TRACER_STATE) == 1 && _robot_read(robot, HAMSTER_LINE_TRACER_STATE) == 0x40) ? 1 : 0;
}

void _hamster_line_tracer_mode(int hamster_index, int mode, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, mode);
	
	switch(mode) {
		case HAMSTER_LINE_TRACER_MODE_BLACK_MOVE_FORWARD:
		case HAMSTER_LINE_TRACER_MODE_BLACK_TURN_LEFT:
		case HAMSTER_LINE_TRACER_MODE_BLACK_TURN_RIGHT:
		case HAMSTER_LINE_TRACER_MODE_BLACK_UTURN:
		case HAMSTER_LINE_TRACER_MODE_WHITE_MOVE_FORWARD:
		case HAMSTER_LINE_TRACER_MODE_WHITE_TURN_LEFT:
		case HAMSTER_LINE_TRACER_MODE_WHITE_TURN_RIGHT:
		case HAMSTER_LINE_TRACER_MODE_WHITE_UTURN: {
			wait_until(callback, NULL);
			_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
			break;
		}
	}
}

void _hamster_line_left(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_LEFT_SENSOR, NULL);
}

void _hamster_line_right(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_RIGHT_SENSOR, NULL);
}

void _hamster_line_both(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_BOTH_SENSORS, NULL);
}

void _hamster_cross_forward(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_MOVE_FORWARD, callback);
}

void _hamster_cross_left(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_TURN_LEFT, callback);
}

void _hamster_cross_right(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_TURN_RIGHT, callback);
}

void _hamster_cross_uturn(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_BLACK_UTURN, callback);
}

void _hamster_white_line_left(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_LEFT_SENSOR, NULL);
}

void _hamster_white_line_right(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_RIGHT_SENSOR, NULL);
}

void _hamster_white_line_both(int hamster_index) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_BOTH_SENSORS, NULL);
}

void _hamster_white_cross_forward(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_MOVE_FORWARD, callback);
}

void _hamster_white_cross_left(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_TURN_LEFT, callback);
}

void _hamster_white_cross_right(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_TURN_RIGHT, callback);
}

void _hamster_white_cross_uturn(int hamster_index, int (*callback)(void* arg)) {
	_hamster_line_tracer_mode(hamster_index, HAMSTER_LINE_TRACER_MODE_WHITE_UTURN, callback);
}

void _hamster_line_speed(int hamster_index, double speed) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_SPEED, _round_double_value(speed));
}

void _hamster_line_tracer_speed(int hamster_index, double speed) {
	_hamster_line_speed(hamster_index, speed);
}

int _hamster_board_forward_callback(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				int left_floor = _robot_read(robot, HAMSTER_LEFT_FLOOR);
				int right_floor = _robot_read(robot, HAMSTER_RIGHT_FLOOR);
				int diff = _round_float_value((left_floor - right_floor) * 0.25f);
				if(left_floor < 50 && right_floor < 50) {
					hamster->board_count ++;
				} else {
					hamster->board_count = 0;
				}
				_robot_write(robot, HAMSTER_LEFT_WHEEL, 45 + diff);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, 45 - diff);
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(hamster->board_count < 10) {
				int diff = _round_float_value((_robot_read(robot, HAMSTER_LEFT_FLOOR) - _robot_read(robot, HAMSTER_RIGHT_FLOOR)) * 0.25f);
				hamster->board_count ++;
				_robot_write(robot, HAMSTER_LEFT_WHEEL, 45 + diff);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, 45 - diff);
			} else {
				_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
				return 1;
			}
			break;
		}
	}
	return 0;
}

void _hamster_board_forward(int hamster_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, 45);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, 45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

int _hamster_board_left_callback(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_LEFT_FLOOR) > 50) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(_robot_read(robot, HAMSTER_LEFT_FLOOR) < 20) {
				hamster->board_state = 3;
			}
			break;
		}
		case 3: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_LEFT_FLOOR) < 20) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 4;
			}
			break;
		}
		case 4: {
			if(_robot_read(robot, HAMSTER_LEFT_FLOOR) > 50) {
				hamster->board_state = 5;
			}
			break;
		}
		case 5: {
			int diff = _robot_read(robot, HAMSTER_LEFT_FLOOR) - _robot_read(robot, HAMSTER_RIGHT_FLOOR);
			if(diff > -15) {
				_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
				return 1;
			} else {
				diff = _round_float_value(diff * 0.5f);
				_robot_write(robot, HAMSTER_LEFT_WHEEL, diff);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, -diff);
			}
			break;
		}
	}
	return 0;
}

void _hamster_board_left(int hamster_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, -45);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, 45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

int _hamster_board_right_callback(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_RIGHT_FLOOR) > 50) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(_robot_read(robot, HAMSTER_RIGHT_FLOOR) < 20) {
				hamster->board_state = 3;
			}
			break;
		}
		case 3: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_RIGHT_FLOOR) < 20) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 4;
			}
			break;
		}
		case 4: {
			if(_robot_read(robot, HAMSTER_RIGHT_FLOOR) > 50) {
				hamster->board_state = 5;
			}
			break;
		}
		case 5: {
			int diff = _robot_read(robot, HAMSTER_RIGHT_FLOOR) - _robot_read(robot, HAMSTER_LEFT_FLOOR);
			if(diff > -15) {
				_robot_write(robot, HAMSTER_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, 0);
				return 1;
			} else {
				diff = _round_float_value(diff * 0.5f);
				_robot_write(robot, HAMSTER_LEFT_WHEEL, -diff);
				_robot_write(robot, HAMSTER_RIGHT_WHEEL, diff);
			}
			break;
		}
	}
	return 0;
}

void _hamster_board_right(int hamster_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_LINE_TRACER_MODE, HAMSTER_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_LEFT_WHEEL, 45);
	_robot_write(robot, HAMSTER_RIGHT_WHEEL, -45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

int _hamster_get_color_value(const char* color) {
	if(color == NULL) return -1;
	if(_STRICMP(color, HAMSTER_LED_OFF) == 0) return _HAMSTER_LED_OFF;
	if(_STRICMP(color, HAMSTER_LED_RED) == 0) return _HAMSTER_LED_RED;
	if(_STRICMP(color, HAMSTER_LED_YELLOW) == 0) return _HAMSTER_LED_YELLOW;
	if(_STRICMP(color, HAMSTER_LED_GREEN) == 0) return _HAMSTER_LED_GREEN;
	if(_STRICMP(color, HAMSTER_LED_SKY_BLUE) == 0) return _HAMSTER_LED_SKY_BLUE;
	if(_STRICMP(color, HAMSTER_LED_BLUE) == 0) return _HAMSTER_LED_BLUE;
	if(_STRICMP(color, HAMSTER_LED_PURPLE) == 0) return _HAMSTER_LED_PURPLE;
	if(_STRICMP(color, HAMSTER_LED_WHITE) == 0) return _HAMSTER_LED_WHITE;
	if(_STRICMP(color, HAMSTER_LED_CYAN) == 0) return _HAMSTER_LED_SKY_BLUE;
	if(_STRICMP(color, HAMSTER_LED_MAGENTA) == 0) return _HAMSTER_LED_PURPLE;
	if(_STRICMP(color, "SKY_BLUE") == 0) return _HAMSTER_LED_SKY_BLUE;
	if(_STRICMP(color, "SKYBLUE") == 0) return _HAMSTER_LED_SKY_BLUE;
	return -1;
}

void _hamster_leds(int hamster_index, const char* left_color, const char* right_color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_get_color_value(left_color);
	if(value >= 0) _robot_write(robot, HAMSTER_LEFT_LED, value);
	value = _hamster_get_color_value(right_color);
	if(value >= 0) _robot_write(robot, HAMSTER_RIGHT_LED, value);
}

void _hamster_left_led(int hamster_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_get_color_value(color);
	if(value >= 0) _robot_write(robot, HAMSTER_LEFT_LED, value);
}

void _hamster_right_led(int hamster_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_get_color_value(color);
	if(value >= 0) _robot_write(robot, HAMSTER_RIGHT_LED, value);
}

void _hamster_beep(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_NOTE, 0);
	_robot_write(robot, HAMSTER_BUZZER, 440);
	wait(100);
	_robot_write(robot, HAMSTER_BUZZER, 0);
	wait(100);
}

void _hamster_buzzer(int hamster_index, double hz) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_NOTE, 0);
	_robot_write_float(robot, HAMSTER_BUZZER, (float)hz);
}

void _hamster_tempo(int hamster_index, double bpm) {
	struct _hamster_robot* hamster = (struct _hamster_robot*)_robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(hamster == NULL) return;
	if(bpm > 0) hamster->bpm = bpm;
}

int _get_pitch_value(const char* pitch) {
	int len;
	
	if(pitch == NULL) return -1;
	if(_STRICMP(pitch, "OFF") == 0) return 0;
	
	len = strlen(pitch);
	if(len > 0) {
		char ch1 = tolower(pitch[0]);
		int value = 0;
		
		switch(ch1) {
			case 'c': value = 4; break;
			case 'd': value = 6; break;
			case 'e': value = 8; break;
			case 'f': value = 9; break;
			case 'g': value = 11; break;
			case 'a': value = 13; break;
			case 'b': value = 15; break;
			default: return -1;
		}
		if(len > 1) {
			char ch2 = tolower(pitch[1]);
			int octave = 4;
			
			if(len > 2) {
				if(ch2 == '#') value ++;
				else if(ch2 == 'b') value --;
				octave = (int)pitch[2] - 48;
			} else {
				if(ch2 == '#') value ++;
				else if(ch2 == 'b') value --;
				else octave = (int)ch2 - 48;
			}
			return value + (octave - 1) * 12;
		}
		return value + 3 * 12;
	}
	return -1;
}

void _hamster_pitch(int hamster_index, const char* pitch) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int _pitch;

	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_BUZZER, 0);
	_pitch = _get_pitch_value(pitch);
	if(_pitch >= 0) _robot_write(robot, HAMSTER_NOTE, _pitch);
}

void _hamster_note(int hamster_index, const char* pitch, double beats) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	struct _hamster_robot* hamster = (struct _hamster_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_BUZZER, 0);
	if(beats > 0 && hamster->bpm > 0) {
		int _pitch = _get_pitch_value(pitch);
		if(_pitch == 0) {
			_robot_write(robot, HAMSTER_NOTE, 0);
			wait((int)(beats * 60 * 1000 / hamster->bpm));
		} else if(_pitch > 0) {
			int timeout = (int)(beats * 60 * 1000 / hamster->bpm);
			int tail = 0;
			if(timeout > 100) {
				tail = 100;
			}
			_robot_write(robot, HAMSTER_NOTE, _pitch);
			wait(timeout - tail);
			_robot_write(robot, HAMSTER_NOTE, 0);
			if(tail > 0) {
				wait(tail);
			}
		}
	} else {
		_robot_write(robot, HAMSTER_NOTE, 0);
	}
}

int _hamster_get_io_mode_value(const char* mode) {
	if(mode == NULL) return -1;
	if(_STRICMP(mode, HAMSTER_IO_MODE_ANALOG_INPUT) == 0) return _HAMSTER_IO_MODE_ANALOG_INPUT;
	if(_STRICMP(mode, HAMSTER_IO_MODE_DIGITAL_INPUT) == 0) return _HAMSTER_IO_MODE_DIGITAL_INPUT;
	if(_STRICMP(mode, HAMSTER_IO_MODE_SERVO_OUTPUT) == 0) return _HAMSTER_IO_MODE_SERVO_OUTPUT;
	if(_STRICMP(mode, HAMSTER_IO_MODE_PWM_OUTPUT) == 0) return _HAMSTER_IO_MODE_PWM_OUTPUT;
	if(_STRICMP(mode, HAMSTER_IO_MODE_DIGITAL_OUTPUT) == 0) return _HAMSTER_IO_MODE_DIGITAL_OUTPUT;
	if(_STRICMP(mode, "ANALOG_INPUT") == 0) return _HAMSTER_IO_MODE_ANALOG_INPUT;
	if(_STRICMP(mode, "DIGITAL_INPUT") == 0) return _HAMSTER_IO_MODE_DIGITAL_INPUT;
	if(_STRICMP(mode, "SERVO_OUTPUT") == 0) return _HAMSTER_IO_MODE_SERVO_OUTPUT;
	if(_STRICMP(mode, "PWM_OUTPUT") == 0) return _HAMSTER_IO_MODE_PWM_OUTPUT;
	if(_STRICMP(mode, "DIGITAL_OUTPUT") == 0) return _HAMSTER_IO_MODE_DIGITAL_OUTPUT;
	return -1;
}

void _hamster_io_mode_a(int hamster_index, const char* mode) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_get_io_mode_value(mode);
	if(value >= 0) _robot_write(robot, HAMSTER_IO_MODE_A, value);
}

void _hamster_io_mode_b(int hamster_index, const char* mode) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_get_io_mode_value(mode);
	if(value >= 0) _robot_write(robot, HAMSTER_IO_MODE_B, value);
}

void _hamster_output_a(int hamster_index, double value) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_OUTPUT_A, _round_double_value(value));
}

void _hamster_output_b(int hamster_index, double value) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_OUTPUT_B, _round_double_value(value));
}

void _hamster_open_gripper(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_IO_MODE_A, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_IO_MODE_B, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_OUTPUT_A, 1);
	_robot_write(robot, HAMSTER_OUTPUT_B, 0);
	wait(500);
}

void _hamster_close_gripper(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_IO_MODE_A, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_IO_MODE_B, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_OUTPUT_A, 0);
	_robot_write(robot, HAMSTER_OUTPUT_B, 1);
	wait(500);
}

void _hamster_release_gripper(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_IO_MODE_A, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_IO_MODE_B, _HAMSTER_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_OUTPUT_A, 0);
	_robot_write(robot, HAMSTER_OUTPUT_B, 0);
}

int _hamster_signal_strength(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_SIGNAL_STRENGTH);
}

int _hamster_left_proximity(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_LEFT_PROXIMITY);
}

int _hamster_right_proximity(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_RIGHT_PROXIMITY);
}

int _hamster_left_floor(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_LEFT_FLOOR);
}

int _hamster_right_floor(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_RIGHT_FLOOR);
}

int _hamster_acceleration_x(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_ACCELERATION, 0);
}

int _hamster_acceleration_y(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_ACCELERATION, 1);
}

int _hamster_acceleration_z(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_ACCELERATION, 2);
}

int _hamster_light(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_LIGHT);
}

int _hamster_temperature(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_TEMPERATURE);
}

int _hamster_input_a(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_INPUT_A);
}

int _hamster_input_b(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_INPUT_B);
}

int _hamster_tilt(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_TILT);
}

int _hamster_battery_state(int hamster_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER, hamster_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_BATTERY_STATE);
}

#define _MAX_NUM_ROBOTS 10

#define _FUNCTION_HAMSTER(n) \
	__inline const char* _hamster_get_name_##n(void) { return _robot_group_get_name(_GROUP_HAMSTER, n); } \
	__inline void _hamster_set_name_##n(const char* name) { _robot_group_set_name(_GROUP_HAMSTER, n, name); } \
	__inline const char* _hamster_get_id_##n(void) { return HAMSTER_ID; } \
	__inline int _hamster_get_index_##n(void) { return n; } \
	__inline int _hamster_e_##n(int device_id) { return _robot_group_e(_GROUP_HAMSTER, n, device_id); } \
	__inline int _hamster_read_##n(int device_id) { return _robot_group_read(_GROUP_HAMSTER, n, device_id); } \
	__inline int _hamster_read_at_##n(int device_id, int index) { return _robot_group_read_at(_GROUP_HAMSTER, n, device_id, index); } \
	__inline int _hamster_read_array_##n(int device_id, int* data, int length) { return _robot_group_read_array(_GROUP_HAMSTER, n, device_id, data, length); } \
	__inline float _hamster_read_float_##n(int device_id) { return _robot_group_read_float(_GROUP_HAMSTER, n, device_id); } \
	__inline float _hamster_read_float_at_##n(int device_id, int index) { return _robot_group_read_float_at(_GROUP_HAMSTER, n, device_id, index); } \
	__inline int _hamster_read_float_array_##n(int device_id, float* data, int length) { return _robot_group_read_float_array(_GROUP_HAMSTER, n, device_id, data, length); } \
	__inline int _hamster_write_empty_##n(int device_id) { return _robot_group_write_empty(_GROUP_HAMSTER, n, device_id); } \
	__inline int _hamster_write_##n(int device_id, int data) { return _robot_group_write(_GROUP_HAMSTER, n, device_id, data); } \
	__inline int _hamster_write_at_##n(int device_id, int index, int data) { return _robot_group_write_at(_GROUP_HAMSTER, n, device_id, index, data); } \
	__inline int _hamster_write_array_##n(int device_id, const int* data, int length) { return _robot_group_write_array(_GROUP_HAMSTER, n, device_id, data, length); } \
	__inline int _hamster_write_float_##n(int device_id, float data) { return _robot_group_write_float(_GROUP_HAMSTER, n, device_id, data); } \
	__inline int _hamster_write_float_at_##n(int device_id, int index, float data) { return _robot_group_write_float_at(_GROUP_HAMSTER, n, device_id, index, data); } \
	__inline int _hamster_write_float_array_##n(int device_id, const float* data, int length) { return _robot_group_write_float_array(_GROUP_HAMSTER, n, device_id, data, length); } \
	__inline void _hamster_reset_##n(void) { _robot_group_reset(_GROUP_HAMSTER, n); } \
	__inline void _hamster_dispose_##n(void) { _robot_group_dispose(_GROUP_HAMSTER, n); } \
	__inline void _hamster_wheels_##n(double left_velocity, double right_velocity) { _hamster_wheels(n, left_velocity, right_velocity); } \
	__inline void _hamster_left_wheel_##n(double velocity) { _hamster_left_wheel(n, velocity); } \
	__inline void _hamster_right_wheel_##n(double velocity) { _hamster_right_wheel(n, velocity); } \
	__inline void _hamster_stop_##n(void) { _hamster_stop(n); } \
	__inline void _hamster_move_forward_##n(double sec, double velocity) { _hamster_move_forward(n, sec, velocity); } \
	__inline void _hamster_move_backward_##n(double sec, double velocity) { _hamster_move_backward(n, sec, velocity); } \
	__inline void _hamster_turn_left_##n(double sec, double velocity) { _hamster_turn_left(n, sec, velocity); } \
	__inline void _hamster_turn_right_##n(double sec, double velocity) { _hamster_turn_right(n, sec, velocity); } \
	__inline int _hamster_line_tracer_mode_callback_##n(void* arg) { return _hamster_line_tracer_mode_callback(n); } \
	__inline void _hamster_line_tracer_mode_##n(int mode) { _hamster_line_tracer_mode(n, mode, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_line_left_##n(void) { _hamster_line_left(n); } \
	__inline void _hamster_line_right_##n(void) { _hamster_line_right(n); } \
	__inline void _hamster_line_both_##n(void) { _hamster_line_both(n); } \
	__inline void _hamster_cross_forward_##n(void) { _hamster_cross_forward(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_cross_left_##n(void) { _hamster_cross_left(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_cross_right_##n(void) { _hamster_cross_right(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_cross_uturn_##n(void) { _hamster_cross_uturn(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_white_line_left_##n(void) { _hamster_white_line_left(n); } \
	__inline void _hamster_white_line_right_##n(void) { _hamster_white_line_right(n); } \
	__inline void _hamster_white_line_both_##n(void) { _hamster_white_line_both(n); } \
	__inline void _hamster_white_cross_forward_##n(void) { _hamster_white_cross_forward(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_white_cross_left_##n(void) { _hamster_white_cross_left(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_white_cross_right_##n(void) { _hamster_white_cross_right(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_white_cross_uturn_##n(void) { _hamster_white_cross_uturn(n, _hamster_line_tracer_mode_callback_##n); } \
	__inline void _hamster_line_speed_##n(double speed) { _hamster_line_speed(n, speed); } \
	__inline void _hamster_line_tracer_speed_##n(double speed) { _hamster_line_tracer_speed(n, speed); } \
	__inline int _hamster_board_forward_callback_##n(void* arg) { return _hamster_board_forward_callback(n); } \
	__inline void _hamster_board_forward_##n(void) { _hamster_board_forward(n, _hamster_board_forward_callback_##n); } \
	__inline int _hamster_board_left_callback_##n(void* arg) { return _hamster_board_left_callback(n); } \
	__inline void _hamster_board_left_##n(void) { _hamster_board_left(n, _hamster_board_left_callback_##n); } \
	__inline int _hamster_board_right_callback_##n(void* arg) { return _hamster_board_right_callback(n); } \
	__inline void _hamster_board_right_##n(void) { _hamster_board_right(n, _hamster_board_right_callback_##n); } \
	__inline void _hamster_leds_##n(const char* left_color, const char* right_color) { _hamster_leds(n, left_color, right_color); } \
	__inline void _hamster_left_led_##n(const char* color) { _hamster_left_led(n, color); } \
	__inline void _hamster_right_led_##n(const char* color) { _hamster_right_led(n, color); } \
	__inline void _hamster_beep_##n(void) { _hamster_beep(n); } \
	__inline void _hamster_buzzer_##n(double hz) { _hamster_buzzer(n, hz); } \
	__inline void _hamster_tempo_##n(double bpm) { _hamster_tempo(n, bpm); } \
	__inline void _hamster_pitch_##n(const char* pitch) { _hamster_pitch(n, pitch); } \
	__inline void _hamster_note_##n(const char* pitch, double beats) { _hamster_note(n, pitch, beats); } \
	__inline void _hamster_io_mode_a_##n(const char* mode) { _hamster_io_mode_a(n, mode); } \
	__inline void _hamster_io_mode_b_##n(const char* mode) { _hamster_io_mode_b(n, mode); } \
	__inline void _hamster_output_a_##n(double value) { _hamster_output_a(n, value); } \
	__inline void _hamster_output_b_##n(double value) { _hamster_output_b(n, value); } \
	__inline void _hamster_open_gripper_##n(void) { _hamster_open_gripper(n); } \
	__inline void _hamster_close_gripper_##n(void) { _hamster_close_gripper(n); } \
	__inline void _hamster_release_gripper_##n(void) { _hamster_release_gripper(n); } \
	__inline int _hamster_signal_strength_##n(void) { return _hamster_signal_strength(n); } \
	__inline int _hamster_left_proximity_##n(void) { return _hamster_left_proximity(n); } \
	__inline int _hamster_right_proximity_##n(void) { return _hamster_right_proximity(n); } \
	__inline int _hamster_left_floor_##n(void) { return _hamster_left_floor(n); } \
	__inline int _hamster_right_floor_##n(void) { return _hamster_right_floor(n); } \
	__inline int _hamster_acceleration_x_##n(void) { return _hamster_acceleration_x(n); } \
	__inline int _hamster_acceleration_y_##n(void) { return _hamster_acceleration_y(n); } \
	__inline int _hamster_acceleration_z_##n(void) { return _hamster_acceleration_z(n); } \
	__inline int _hamster_light_##n(void) { return _hamster_light(n); } \
	__inline int _hamster_temperature_##n(void) { return _hamster_temperature(n); } \
	__inline int _hamster_input_a_##n(void) { return _hamster_input_a(n); } \
	__inline int _hamster_input_b_##n(void) { return _hamster_input_b(n); } \
	__inline int _hamster_tilt_##n(void) { return _hamster_tilt(n); } \
	__inline int _hamster_battery_state_##n(void) { return _hamster_battery_state(n); }

#define _FUNCTION_PTR_HAMSTER(name, n) \
	name->get_name = _hamster_get_name_##n; \
	name->set_name = _hamster_set_name_##n; \
	name->get_id = _hamster_get_id_##n; \
	name->get_index = _hamster_get_index_##n; \
	name->e = _hamster_e_##n; \
	name->read = _hamster_read_##n; \
	name->read_at = _hamster_read_at_##n; \
	name->read_array = _hamster_read_array_##n; \
	name->read_float = _hamster_read_float_##n; \
	name->read_float_at = _hamster_read_float_at_##n; \
	name->read_float_array = _hamster_read_float_array_##n; \
	name->write_empty = _hamster_write_empty_##n; \
	name->write = _hamster_write_##n; \
	name->write_at = _hamster_write_at_##n; \
	name->write_array = _hamster_write_array_##n; \
	name->write_float = _hamster_write_float_##n; \
	name->write_float_at = _hamster_write_float_at_##n; \
	name->write_float_array = _hamster_write_float_array_##n; \
	name->reset = _hamster_reset_##n; \
	name->dispose = _hamster_dispose_##n; \
	name->wheels = _hamster_wheels_##n; \
	name->left_wheel = _hamster_left_wheel_##n; \
	name->right_wheel = _hamster_right_wheel_##n; \
	name->stop = _hamster_stop_##n; \
	name->move_forward = _hamster_move_forward_##n; \
	name->move_backward = _hamster_move_backward_##n; \
	name->turn_left = _hamster_turn_left_##n; \
	name->turn_right = _hamster_turn_right_##n; \
	name->line_tracer_mode = _hamster_line_tracer_mode_##n; \
	name->line_left = _hamster_line_left_##n; \
	name->line_right = _hamster_line_right_##n; \
	name->line_both = _hamster_line_both_##n; \
	name->cross_forward = _hamster_cross_forward_##n; \
	name->cross_left = _hamster_cross_left_##n; \
	name->cross_right = _hamster_cross_right_##n; \
	name->cross_uturn = _hamster_cross_uturn_##n; \
	name->white_line_left = _hamster_white_line_left_##n; \
	name->white_line_right = _hamster_white_line_right_##n; \
	name->white_line_both = _hamster_white_line_both_##n; \
	name->white_cross_forward = _hamster_white_cross_forward_##n; \
	name->white_cross_left = _hamster_white_cross_left_##n; \
	name->white_cross_right = _hamster_white_cross_right_##n; \
	name->white_cross_uturn = _hamster_white_cross_uturn_##n; \
	name->line_speed = _hamster_line_speed_##n; \
	name->line_tracer_speed = _hamster_line_tracer_speed_##n; \
	name->board_forward = _hamster_board_forward_##n; \
	name->board_left = _hamster_board_left_##n; \
	name->board_right = _hamster_board_right_##n; \
	name->leds = _hamster_leds_##n; \
	name->left_led = _hamster_left_led_##n; \
	name->right_led = _hamster_right_led_##n; \
	name->beep = _hamster_beep_##n; \
	name->buzzer = _hamster_buzzer_##n; \
	name->tempo = _hamster_tempo_##n; \
	name->pitch = _hamster_pitch_##n; \
	name->note = _hamster_note_##n; \
	name->io_mode_a = _hamster_io_mode_a_##n; \
	name->io_mode_b = _hamster_io_mode_b_##n; \
	name->output_a = _hamster_output_a_##n; \
	name->output_b = _hamster_output_b_##n; \
	name->open_gripper = _hamster_open_gripper_##n; \
	name->close_gripper = _hamster_close_gripper_##n; \
	name->release_gripper = _hamster_release_gripper_##n; \
	name->signal_strength = _hamster_signal_strength_##n; \
	name->left_proximity = _hamster_left_proximity_##n; \
	name->right_proximity = _hamster_right_proximity_##n; \
	name->left_floor = _hamster_left_floor_##n; \
	name->right_floor = _hamster_right_floor_##n; \
	name->acceleration_x = _hamster_acceleration_x_##n; \
	name->acceleration_y = _hamster_acceleration_y_##n; \
	name->acceleration_z = _hamster_acceleration_z_##n; \
	name->light = _hamster_light_##n; \
	name->temperature = _hamster_temperature_##n; \
	name->input_a = _hamster_input_a_##n; \
	name->input_b = _hamster_input_b_##n; \
	name->tilt = _hamster_tilt_##n; \
	name->battery_state = _hamster_battery_state_##n;

_FUNCTION_HAMSTER(0);
_FUNCTION_HAMSTER(1);
_FUNCTION_HAMSTER(2);
_FUNCTION_HAMSTER(3);
_FUNCTION_HAMSTER(4);
_FUNCTION_HAMSTER(5);
_FUNCTION_HAMSTER(6);
_FUNCTION_HAMSTER(7);
_FUNCTION_HAMSTER(8);
_FUNCTION_HAMSTER(9);

Hamster* hamster_create_port(const char* port_name) {
	if(_robot_group_count_robots(_GROUP_HAMSTER) < _MAX_NUM_ROBOTS) {
		struct hamster* hamster = (struct hamster*)malloc(sizeof(struct hamster));
		struct _robot* robot = (struct _robot*)_hamster_create(port_name);
	
		switch(robot->index) {
			case 0: _FUNCTION_PTR_HAMSTER(hamster, 0); break;
			case 1: _FUNCTION_PTR_HAMSTER(hamster, 1); break;
			case 2: _FUNCTION_PTR_HAMSTER(hamster, 2); break;
			case 3: _FUNCTION_PTR_HAMSTER(hamster, 3); break;
			case 4: _FUNCTION_PTR_HAMSTER(hamster, 4); break;
			case 5: _FUNCTION_PTR_HAMSTER(hamster, 5); break;
			case 6: _FUNCTION_PTR_HAMSTER(hamster, 6); break;
			case 7: _FUNCTION_PTR_HAMSTER(hamster, 7); break;
			case 8: _FUNCTION_PTR_HAMSTER(hamster, 8); break;
			case 9: _FUNCTION_PTR_HAMSTER(hamster, 9); break;
		}
		return hamster;
	}
	return NULL;
}

Hamster* hamster_create(void) {
	return hamster_create_port(NULL);
}

const char* hamster_get_name(void) {
	return _robot_group_get_name(_GROUP_HAMSTER, 0);
}

void hamster_set_name(const char* name) {
	_robot_group_set_name(_GROUP_HAMSTER, 0, name);
}

const char* hamster_get_id(void) {
	return HAMSTER_ID;
}

int hamster_e(int device_id) {
	return _robot_group_e(_GROUP_HAMSTER, 0, device_id);
}

int hamster_read(int device_id) {
	return _robot_group_read(_GROUP_HAMSTER, 0, device_id);
}

int hamster_read_at(int device_id, int index) {
	return _robot_group_read_at(_GROUP_HAMSTER, 0, device_id, index);
}

int hamster_read_array(int device_id, int* data, int length) {
	return _robot_group_read_array(_GROUP_HAMSTER, 0, device_id, data, length);
}

float hamster_read_float(int device_id) {
	return _robot_group_read_float(_GROUP_HAMSTER, 0, device_id);
}

float hamster_read_float_at(int device_id, int index) {
	return _robot_group_read_float_at(_GROUP_HAMSTER, 0, device_id, index);
}

int hamster_read_float_array(int device_id, float* data, int length) {
	return _robot_group_read_float_array(_GROUP_HAMSTER, 0, device_id, data, length);
}

int hamster_write_empty(int device_id) {
	return _robot_group_write_empty(_GROUP_HAMSTER, 0, device_id);
}

int hamster_write(int device_id, int data) {
	return _robot_group_write(_GROUP_HAMSTER, 0, device_id, data);
}

int hamster_write_at(int device_id, int index, int data) {
	return _robot_group_write_at(_GROUP_HAMSTER, 0, device_id, index, data);
}

int hamster_write_array(int device_id, const int* data, int length) {
	return _robot_group_write_array(_GROUP_HAMSTER, 0, device_id, data, length);
}

int hamster_write_float(int device_id, float data) {
	return _robot_group_write_float(_GROUP_HAMSTER, 0, device_id, data);
}

int hamster_write_float_at(int device_id, int index, float data) {
	return _robot_group_write_float_at(_GROUP_HAMSTER, 0, device_id, index, data);
}

int hamster_write_float_array(int device_id, const float* data, int length) {
	return _robot_group_write_float_array(_GROUP_HAMSTER, 0, device_id, data, length);
}

void hamster_reset(void) {
	_robot_group_reset(_GROUP_HAMSTER, 0);
}

void hamster_dispose(void) {
	_robot_group_dispose(_GROUP_HAMSTER, 0);
}

void hamster_wheels(double left_velocity, double right_velocity) {
	_hamster_wheels(0, left_velocity, right_velocity);
}

void hamster_left_wheel(double velocity) {
	_hamster_left_wheel(0, velocity);
}

void hamster_right_wheel(double velocity) {
	_hamster_right_wheel(0, velocity);
}

void hamster_stop(void) {
	_hamster_stop(0);
}

void hamster_move_forward(double sec, double velocity) {
	_hamster_move_forward(0, sec, velocity);
}

void hamster_move_backward(double sec, double velocity) {
	_hamster_move_backward(0, sec, velocity);
}

void hamster_turn_left(double sec, double velocity) {
	_hamster_turn_left(0, sec, velocity);
}

void hamster_turn_right(double sec, double velocity) {
	_hamster_turn_right(0, sec, velocity);
}

int _hamster_line_tracer_mode_evaluate(void* arg) {
	return _hamster_line_tracer_mode_callback(0);
}

void hamster_line_tracer_mode(int mode) {
	_hamster_line_tracer_mode(0, mode, _hamster_line_tracer_mode_evaluate);
}

void hamster_line_left(void) {
	_hamster_line_left(0);
}

void hamster_line_right(void) {
	_hamster_line_right(0);
}

void hamster_line_both(void) {
	_hamster_line_both(0);
}

void hamster_cross_forward(void) {
	_hamster_cross_forward(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_cross_left(void) {
	_hamster_cross_left(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_cross_right(void) {
	_hamster_cross_right(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_cross_uturn(void) {
	_hamster_cross_uturn(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_white_line_left(void) {
	_hamster_white_line_left(0);
}

void hamster_white_line_right(void) {
	_hamster_white_line_right(0);
}

void hamster_white_line_both(void) {
	_hamster_white_line_both(0);
}

void hamster_white_cross_forward(void) {
	_hamster_white_cross_forward(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_white_cross_left(void) {
	_hamster_white_cross_left(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_white_cross_right(void) {
	_hamster_white_cross_right(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_white_cross_uturn(void) {
	_hamster_white_cross_uturn(0, _hamster_line_tracer_mode_evaluate);
}

void hamster_line_speed(double speed) {
	_hamster_line_speed(0, speed);
}

void hamster_line_tracer_speed(double speed) {
	_hamster_line_tracer_speed(0, speed);
}

int _hamster_board_forward_evaluate(void* arg) {
	return _hamster_board_forward_callback(0);
}

void hamster_board_forward(void) {
	_hamster_board_forward(0, _hamster_board_forward_evaluate);
}

int _hamster_board_left_evaluate(void* arg) {
	return _hamster_board_left_callback(0);
}

void hamster_board_left(void) {
	_hamster_board_left(0, _hamster_board_left_evaluate);
}

int _hamster_board_right_evaluate(void* arg) {
	return _hamster_board_right_callback(0);
}

void hamster_board_right(void) {
	_hamster_board_right(0, _hamster_board_right_evaluate);
}

void hamster_leds(const char* left_color, const char* right_color) {
	_hamster_leds(0, left_color, right_color);
}

void hamster_left_led(const char* color) {
	_hamster_left_led(0, color);
}

void hamster_right_led(const char* color) {
	_hamster_right_led(0, color);
}

void hamster_beep(void) {
	_hamster_beep(0);
}

void hamster_buzzer(double hz) {
	_hamster_buzzer(0, hz);
}

void hamster_tempo(double bpm) {
	_hamster_tempo(0, bpm);
}

void hamster_pitch(const char* pitch) {
	_hamster_pitch(0, pitch);
}

void hamster_note(const char* pitch, double beats) {
	_hamster_note(0, pitch, beats);
}

void hamster_io_mode_a(const char* mode) {
	_hamster_io_mode_a(0, mode);
}

void hamster_io_mode_b(const char* mode) {
	_hamster_io_mode_b(0, mode);
}

void hamster_output_a(double value) {
	_hamster_output_a(0, value);
}

void hamster_output_b(double value) {
	_hamster_output_b(0, value);
}

void hamster_open_gripper(void) {
	_hamster_open_gripper(0);
}

void hamster_close_gripper(void) {
	_hamster_close_gripper(0);
}

void hamster_release_gripper(void) {
	_hamster_release_gripper(0);
}

int hamster_signal_strength(void) {
	return _hamster_signal_strength(0);
}

int hamster_left_proximity(void) {
	return _hamster_left_proximity(0);
}

int hamster_right_proximity(void) {
	return _hamster_right_proximity(0);
}

int hamster_left_floor(void) {
	return _hamster_left_floor(0);
}

int hamster_right_floor(void) {
	return _hamster_right_floor(0);
}

int hamster_acceleration_x(void) {
	return _hamster_acceleration_x(0);
}

int hamster_acceleration_y(void) {
	return _hamster_acceleration_y(0);
}

int hamster_acceleration_z(void) {
	return _hamster_acceleration_z(0);
}

int hamster_light(void) {
	return _hamster_light(0);
}

int hamster_temperature(void) {
	return _hamster_temperature(0);
}

int hamster_input_a(void) {
	return _hamster_input_a(0);
}

int hamster_input_b(void) {
	return _hamster_input_b(0);
}

int hamster_tilt(void) {
	return _hamster_tilt(0);
}

int hamster_battery_state(void) {
	return _hamster_battery_state(0);
}

/*------------------------------
  HAMSTER-S
------------------------------*/

#define _HAMSTER_S_LEFT_WHEEL_INDEX 0
#define _HAMSTER_S_RIGHT_WHEEL_INDEX 1
#define _HAMSTER_S_LEFT_RGB_INDEX 2
#define _HAMSTER_S_RIGHT_RGB_INDEX 3
#define _HAMSTER_S_BUZZER_INDEX 4
#define _HAMSTER_S_OUTPUT_A_INDEX 5
#define _HAMSTER_S_OUTPUT_B_INDEX 6
#define _HAMSTER_S_PULSE_INDEX 7
#define _HAMSTER_S_NOTE_INDEX 8
#define _HAMSTER_S_SOUND_INDEX 9
#define _HAMSTER_S_LINE_TRACER_MODE_INDEX 10
#define _HAMSTER_S_LINE_TRACER_GAIN_INDEX 11
#define _HAMSTER_S_LINE_TRACER_SPEED_INDEX 12
#define _HAMSTER_S_IO_MODE_A_INDEX 13
#define _HAMSTER_S_IO_MODE_B_INDEX 14
#define _HAMSTER_S_WRITE_SERIAL_INDEX 15
#define _HAMSTER_S_MOTOR_MODE_INDEX 16
#define _HAMSTER_S_CONFIG_PROXIMITY_INDEX 17
#define _HAMSTER_S_CONFIG_GRAVITY_INDEX 18
#define _HAMSTER_S_CONFIG_BAND_WIDTH_INDEX 19
#define _HAMSTER_S_SIGNAL_STRENGTH_INDEX 20
#define _HAMSTER_S_LEFT_PROXIMITY_INDEX 21
#define _HAMSTER_S_RIGHT_PROXIMITY_INDEX 22
#define _HAMSTER_S_LEFT_FLOOR_INDEX 23
#define _HAMSTER_S_RIGHT_FLOOR_INDEX 24
#define _HAMSTER_S_ACCELERATION_INDEX 25
#define _HAMSTER_S_LIGHT_INDEX 26
#define _HAMSTER_S_TEMPERATURE_INDEX 27
#define _HAMSTER_S_INPUT_A_INDEX 28
#define _HAMSTER_S_INPUT_B_INDEX 29
#define _HAMSTER_S_FREE_FALL_INDEX 30
#define _HAMSTER_S_TAP_INDEX 31
#define _HAMSTER_S_TILT_INDEX 32
#define _HAMSTER_S_READ_SERIAL_INDEX 33
#define _HAMSTER_S_PULSE_COUNT_INDEX 34
#define _HAMSTER_S_WHEEL_STATE_INDEX 35
#define _HAMSTER_S_SOUND_STATE_INDEX 36
#define _HAMSTER_S_LINE_TRACER_STATE_INDEX 37
#define _HAMSTER_S_BATTERY_STATE_INDEX 38
#define _HAMSTER_S_SERIAL_STATE_INDEX 39

#define _GROUP_HAMSTER_S 1

#define _HAMSTER_S_LINE_TRACER_MODE_OFF 0
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_LEFT_SENSOR 1
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_RIGHT_SENSOR 2
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_BOTH_SENSORS 3
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_LEFT 4
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_RIGHT 5
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_MOVE_FORWARD 6
#define _HAMSTER_S_LINE_TRACER_MODE_BLACK_UTURN 7
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_LEFT_SENSOR 8
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_RIGHT_SENSOR 9
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_BOTH_SENSORS 10
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_LEFT 11
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_RIGHT 12
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_MOVE_FORWARD 13
#define _HAMSTER_S_LINE_TRACER_MODE_WHITE_UTURN 14

#define _HAMSTER_S_IO_MODE_ANALOG_INPUT 0
#define _HAMSTER_S_IO_MODE_ANALOG_INPUT_RELATIVE 0
#define _HAMSTER_S_IO_MODE_DIGITAL_INPUT 1
#define _HAMSTER_S_IO_MODE_DIGITAL_INPUT_TRI_STATE 1
#define _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_UP 2
#define _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_DOWN 3
#define _HAMSTER_S_IO_MODE_ANALOG_INPUT_ABSOLUTE 4
#define _HAMSTER_S_IO_MODE_VOLTAGE_INPUT 5
#define _HAMSTER_S_IO_MODE_SERVO_OUTPUT 8
#define _HAMSTER_S_IO_MODE_PWM_OUTPUT 9
#define _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT 10
#define _HAMSTER_S_IO_MODE_SERIAL_9600 176
#define _HAMSTER_S_IO_MODE_SERIAL_14400 177
#define _HAMSTER_S_IO_MODE_SERIAL_19200 178
#define _HAMSTER_S_IO_MODE_SERIAL_28800 179
#define _HAMSTER_S_IO_MODE_SERIAL_38400 180
#define _HAMSTER_S_IO_MODE_SERIAL_57600 181
#define _HAMSTER_S_IO_MODE_SERIAL_76800 182
#define _HAMSTER_S_IO_MODE_SERIAL_115200 183

#define _HAMSTER_S_SOUND_OFF 0
#define _HAMSTER_S_SOUND_BEEP 1
#define _HAMSTER_S_SOUND_RANDOM_BEEP 2
#define _HAMSTER_S_SOUND_NOISE 10
#define _HAMSTER_S_SOUND_SIREN 3
#define _HAMSTER_S_SOUND_ENGINE 4
#define _HAMSTER_S_SOUND_CHOP 11
#define _HAMSTER_S_SOUND_ROBOT 5
#define _HAMSTER_S_SOUND_DIBIDIBIDIP 8
#define _HAMSTER_S_SOUND_GOOD_JOB 9
#define _HAMSTER_S_SOUND_HAPPY 12
#define _HAMSTER_S_SOUND_ANGRY 13
#define _HAMSTER_S_SOUND_SAD 14
#define _HAMSTER_S_SOUND_SLEEP 15
#define _HAMSTER_S_SOUND_MARCH 6
#define _HAMSTER_S_SOUND_BIRTHDAY 7

#define _HAMSTER_S_MOTOR_AUTO 0
#define _HAMSTER_S_MOTOR_WAVE 1
#define _HAMSTER_S_MOTOR_HALF 2
#define _HAMSTER_S_MOTOR_FULL 3

struct _hamster_s_robot {
	struct _robot robot;
	int left_wheel;
	int right_wheel;
	int left_red;
	int left_green;
	int left_blue;
	int right_red;
	int right_green;
	int right_blue;
	float buzzer;
	int output_a;
	int output_b;
	int pulse;
	int note;
	int sound;
	int line_tracer_mode;
	int line_tracer_gain;
	int line_tracer_speed;
	int io_mode_a;
	int io_mode_b;
	int write_serial_data[19];
	int read_serial_data[19];
	int motor_mode;
	int config_proximity;
	int config_gravity;
	int config_band_width;
	int light;
	int temperature;
	int wheel_id;
	int wheel_pulse;
	int wheel_pulse_prev;
	int wheel_event;
	int wheel_state;
	int wheel_count;
	int wheel_move;
	int wheel_move_count;
	int pulse_written;
	int current_sound;
	int sound_repeat;
	int sound_flag;
	int sound_event;
	int sound_state;
	int sound_count;
	int sound_written;
	int line_tracer_id;
	int line_tracer_event;
	int line_tracer_state;
	int line_tracer_count;
	int line_tracer_mode_written;
	int event_free_fall_id;
	int event_tap_id;
	int event_tilt;
	int event_serial_id;
	int event_pulse_count;
	int event_battery_state;
	int command_serial_id;
	int write_serial_written;
	int port_ack_id;
	int port_serial;
	int serial_send_id;
	int serial_send_prev_id;
	int packet_sent;
	int packet_received;
	struct _write_queue write_queue;
	struct _read_queue read_queue;
	int write_serial_id;
	double bpm;
	int gain;
	int speed;
	int board_count;
	int board_state;
	int serial_baud;
};

int _hamster_s_check_connection(struct _connector* connector, struct _serial* serial) {
	int i, read_bytes;

	for(i = 0; i < _RETRY; ++i) {
		_serial_write(serial, "FF\r", 3);
		read_bytes = _connector_read_packet(connector, serial, "FF");
		if(read_bytes != 0) {
			char tmp[5][256];
			int index = 0, len;
			char* buf;
			char* comma;

			buf = connector->buffer;
			comma = strchr(buf, ',');
			while(comma != NULL && index < 5) {
				len = comma - buf;
				_STRNCPY(tmp[index], 256, buf, len);
				tmp[index][len] = '\0';
				index ++;
				buf = comma + 1;
				comma = strchr(buf, ',');
			}
			if(index >= 4 && buf != NULL && strlen(buf) >= 12) {
				if(_STRICMP(tmp[2], "0E") == 0) {
					buf[12] = '\0';
					_connector_set_address(connector, buf);
					_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTED);
					return _CONNECTION_RESULT_FOUND;
				}
			}
			return _CONNECTION_RESULT_NOT_AVAILABLE;
		}
	}
	return _CONNECTION_RESULT_NOT_AVAILABLE;
}

struct _device* _hamster_s_find_device(struct _robot* robot, int device_id) {
	if(robot == NULL) return NULL;
	if(robot->devices == NULL) return NULL;
	switch(device_id) {
		case HAMSTER_S_LEFT_WHEEL: return robot->devices[_HAMSTER_S_LEFT_WHEEL_INDEX];
		case HAMSTER_S_RIGHT_WHEEL: return robot->devices[_HAMSTER_S_RIGHT_WHEEL_INDEX];
		case HAMSTER_S_LEFT_RGB: return robot->devices[_HAMSTER_S_LEFT_RGB_INDEX];
		case HAMSTER_S_RIGHT_RGB: return robot->devices[_HAMSTER_S_RIGHT_RGB_INDEX];
		case HAMSTER_S_BUZZER: return robot->devices[_HAMSTER_S_BUZZER_INDEX];
		case HAMSTER_S_OUTPUT_A: return robot->devices[_HAMSTER_S_OUTPUT_A_INDEX];
		case HAMSTER_S_OUTPUT_B: return robot->devices[_HAMSTER_S_OUTPUT_B_INDEX];
		case HAMSTER_S_PULSE: return robot->devices[_HAMSTER_S_PULSE_INDEX];
		case HAMSTER_S_NOTE: return robot->devices[_HAMSTER_S_NOTE_INDEX];
		case HAMSTER_S_SOUND: return robot->devices[_HAMSTER_S_SOUND_INDEX];
		case HAMSTER_S_LINE_TRACER_MODE: return robot->devices[_HAMSTER_S_LINE_TRACER_MODE_INDEX];
		case HAMSTER_S_LINE_TRACER_GAIN: return robot->devices[_HAMSTER_S_LINE_TRACER_GAIN_INDEX];
		case HAMSTER_S_LINE_TRACER_SPEED: return robot->devices[_HAMSTER_S_LINE_TRACER_SPEED_INDEX];
		case HAMSTER_S_IO_MODE_A: return robot->devices[_HAMSTER_S_IO_MODE_A_INDEX];
		case HAMSTER_S_IO_MODE_B: return robot->devices[_HAMSTER_S_IO_MODE_B_INDEX];
		case HAMSTER_S_WRITE_SERIAL: return robot->devices[_HAMSTER_S_WRITE_SERIAL_INDEX];
		case HAMSTER_S_MOTOR_MODE: return robot->devices[_HAMSTER_S_MOTOR_MODE_INDEX];
		case HAMSTER_S_CONFIG_PROXIMITY: return robot->devices[_HAMSTER_S_CONFIG_PROXIMITY_INDEX];
		case HAMSTER_S_CONFIG_GRAVITY: return robot->devices[_HAMSTER_S_CONFIG_GRAVITY_INDEX];
		case HAMSTER_S_CONFIG_BAND_WIDTH: return robot->devices[_HAMSTER_S_CONFIG_BAND_WIDTH_INDEX];
		case HAMSTER_S_SIGNAL_STRENGTH: return robot->devices[_HAMSTER_S_SIGNAL_STRENGTH_INDEX];
		case HAMSTER_S_LEFT_PROXIMITY: return robot->devices[_HAMSTER_S_LEFT_PROXIMITY_INDEX];
		case HAMSTER_S_RIGHT_PROXIMITY: return robot->devices[_HAMSTER_S_RIGHT_PROXIMITY_INDEX];
		case HAMSTER_S_LEFT_FLOOR: return robot->devices[_HAMSTER_S_LEFT_FLOOR_INDEX];
		case HAMSTER_S_RIGHT_FLOOR: return robot->devices[_HAMSTER_S_RIGHT_FLOOR_INDEX];
		case HAMSTER_S_ACCELERATION: return robot->devices[_HAMSTER_S_ACCELERATION_INDEX];
		case HAMSTER_S_LIGHT: return robot->devices[_HAMSTER_S_LIGHT_INDEX];
		case HAMSTER_S_TEMPERATURE: return robot->devices[_HAMSTER_S_TEMPERATURE_INDEX];
		case HAMSTER_S_INPUT_A: return robot->devices[_HAMSTER_S_INPUT_A_INDEX];
		case HAMSTER_S_INPUT_B: return robot->devices[_HAMSTER_S_INPUT_B_INDEX];
		case HAMSTER_S_FREE_FALL: return robot->devices[_HAMSTER_S_FREE_FALL_INDEX];
		case HAMSTER_S_TAP: return robot->devices[_HAMSTER_S_TAP_INDEX];
		case HAMSTER_S_TILT: return robot->devices[_HAMSTER_S_TILT_INDEX];
		case HAMSTER_S_READ_SERIAL: return robot->devices[_HAMSTER_S_READ_SERIAL_INDEX];
		case HAMSTER_S_PULSE_COUNT: return robot->devices[_HAMSTER_S_PULSE_COUNT_INDEX];
		case HAMSTER_S_WHEEL_STATE: return robot->devices[_HAMSTER_S_WHEEL_STATE_INDEX];
		case HAMSTER_S_SOUND_STATE: return robot->devices[_HAMSTER_S_SOUND_STATE_INDEX];
		case HAMSTER_S_LINE_TRACER_STATE: return robot->devices[_HAMSTER_S_LINE_TRACER_STATE_INDEX];
		case HAMSTER_S_BATTERY_STATE: return robot->devices[_HAMSTER_S_BATTERY_STATE_INDEX];
		case HAMSTER_S_SERIAL_STATE: return robot->devices[_HAMSTER_S_SERIAL_STATE_INDEX];
	}
	return NULL;
}

void _hamster_s_request_motoring_data(struct _robot* robot) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	struct _device** devices = robot->devices;
	
	hamster->left_wheel = _device_read(devices[_HAMSTER_S_LEFT_WHEEL_INDEX]);
	hamster->right_wheel = _device_read(devices[_HAMSTER_S_RIGHT_WHEEL_INDEX]);
	hamster->left_red = _device_read_at(devices[_HAMSTER_S_LEFT_RGB_INDEX], 0);
	hamster->left_green = _device_read_at(devices[_HAMSTER_S_LEFT_RGB_INDEX], 1);
	hamster->left_blue = _device_read_at(devices[_HAMSTER_S_LEFT_RGB_INDEX], 2);
	hamster->right_red = _device_read_at(devices[_HAMSTER_S_RIGHT_RGB_INDEX], 0);
	hamster->right_green = _device_read_at(devices[_HAMSTER_S_RIGHT_RGB_INDEX], 1);
	hamster->right_blue = _device_read_at(devices[_HAMSTER_S_RIGHT_RGB_INDEX], 2);
	hamster->buzzer = _device_read_float(devices[_HAMSTER_S_BUZZER_INDEX]);
	hamster->output_a = _device_read(devices[_HAMSTER_S_OUTPUT_A_INDEX]);
	hamster->output_b = _device_read(devices[_HAMSTER_S_OUTPUT_B_INDEX]);
	if(devices[_HAMSTER_S_PULSE_INDEX]->written == 1) {
		hamster->pulse = _device_read(devices[_HAMSTER_S_PULSE_INDEX]);
		hamster->pulse_written = 1;
	}
	if(devices[_HAMSTER_S_NOTE_INDEX]->written == 1) {
		hamster->note = _device_read(devices[_HAMSTER_S_NOTE_INDEX]);
	}
	if(devices[_HAMSTER_S_SOUND_INDEX]->written == 1) {
		hamster->sound = _device_read(devices[_HAMSTER_S_SOUND_INDEX]);
		hamster->sound_written = 1;
	}
	if(devices[_HAMSTER_S_LINE_TRACER_MODE_INDEX]->written == 1) {
		hamster->line_tracer_mode = _device_read(devices[_HAMSTER_S_LINE_TRACER_MODE_INDEX]);
		hamster->line_tracer_mode_written = 1;
	}
	if(devices[_HAMSTER_S_LINE_TRACER_GAIN_INDEX]->written == 1) {
		hamster->line_tracer_gain = _device_read(devices[_HAMSTER_S_LINE_TRACER_GAIN_INDEX]);
	}
	if(devices[_HAMSTER_S_LINE_TRACER_SPEED_INDEX]->written == 1) {
		hamster->line_tracer_speed = _device_read(devices[_HAMSTER_S_LINE_TRACER_SPEED_INDEX]);
	}
	if(devices[_HAMSTER_S_IO_MODE_A_INDEX]->written == 1) {
		hamster->io_mode_a = _device_read(devices[_HAMSTER_S_IO_MODE_A_INDEX]);
	}
	if(devices[_HAMSTER_S_IO_MODE_B_INDEX]->written == 1) {
		hamster->io_mode_b = _device_read(devices[_HAMSTER_S_IO_MODE_B_INDEX]);
	}
	if(devices[_HAMSTER_S_WRITE_SERIAL_INDEX]->written == 1) {
		_device_read_array(devices[_HAMSTER_S_WRITE_SERIAL_INDEX], hamster->write_serial_data, 19);
		hamster->write_serial_written = 1;
	}
	if(devices[_HAMSTER_S_MOTOR_MODE_INDEX]->written == 1) {
		hamster->motor_mode = _device_read(devices[_HAMSTER_S_MOTOR_MODE_INDEX]);
	}
	if(devices[_HAMSTER_S_CONFIG_PROXIMITY_INDEX]->written == 1) {
		hamster->config_proximity = _device_read(devices[_HAMSTER_S_CONFIG_PROXIMITY_INDEX]);
	}
	if(devices[_HAMSTER_S_CONFIG_GRAVITY_INDEX]->written == 1) {
		hamster->config_gravity = _device_read(devices[_HAMSTER_S_CONFIG_GRAVITY_INDEX]);
	}
	if(devices[_HAMSTER_S_CONFIG_BAND_WIDTH_INDEX]->written == 1) {
		hamster->config_band_width = _device_read(devices[_HAMSTER_S_CONFIG_BAND_WIDTH_INDEX]);
	}
	_robot_clear_written(robot);
}

#define _HAMSTER_S_WHEEL_CENTER_DISTANCE 1.685234f
#define _HAMSTER_S_DEG_TO_PULSE 3.975f
#define _HAMSTER_S_DEG_TO_PULSE_PIVOT_PEN 3.895f
#define _HAMSTER_S_CM_TO_PULSE 137.3078f
#define _HAMSTER_S_PEN_CENTER_DISTANCE 2.452266f

int _hamster_s_cm_to_pulse(double cm) {
	return _round_double_value(cm * _HAMSTER_S_CM_TO_PULSE);
}

int _hamster_s_calc_turn_pulse(double degree) {
	return _round_double_value(degree * _HAMSTER_S_DEG_TO_PULSE);
}

int _hamster_s_calc_pivot_pulse(double degree) {
	return _round_double_value(degree * _HAMSTER_S_DEG_TO_PULSE * 2);
}

double _hamster_s_calc_inner_velocity(double outer_velocity, double radius) {
	return outer_velocity * (radius - _HAMSTER_S_WHEEL_CENTER_DISTANCE) / (radius + _HAMSTER_S_WHEEL_CENTER_DISTANCE);
}

int _hamster_s_calc_circle_pulse(double degree, double radius) {
	return _round_double_value(degree * _HAMSTER_S_DEG_TO_PULSE * (radius + _HAMSTER_S_WHEEL_CENTER_DISTANCE) / _HAMSTER_S_WHEEL_CENTER_DISTANCE);
}

int _hamster_s_calc_pivot_pen_pulse(double degree, double radius) {
	return _round_double_value(degree * _HAMSTER_S_DEG_TO_PULSE_PIVOT_PEN * (radius + _HAMSTER_S_WHEEL_CENTER_DISTANCE) / _HAMSTER_S_WHEEL_CENTER_DISTANCE);
}

void _hamster_s_encode_motoring_packet(struct _robot* robot) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	char* buffer = robot->write_buffer;
	int index = 0, temp, t, i, len;
	const char* address;

	if(hamster->port_serial == 1 && hamster->write_serial_written == 1 && hamster->packet_sent != 3) {
		hamster->write_serial_written = 0;
		hamster->packet_sent = 3;
		hamster->serial_send_id = (hamster->serial_send_id % 255) + 1;
		
		len = hamster->write_serial_data[0];
		if(len > 0) {
			if(len > 18) len = 18;
			hamster->command_serial_id = (hamster->command_serial_id % 15) + 1;
			temp = 0x20 | (hamster->command_serial_id & 0x0f);
			index = _value_to_hex(buffer, index, temp, 1);
			index = _value_to_hex(buffer, index, len, 1);
			for(i = 1; i <= len; ++i) {
				index = _value_to_hex(buffer, index, hamster->write_serial_data[i], 1);
			}
			for(i = len + 1; i <= 18; ++i) {
				buffer[index++] = '0';
				buffer[index++] = '0';
			}
			buffer[index++] = '-';
			address = _connector_get_address(robot->connector);
			for(i = 0; i < 12; ++i) {
				buffer[index++] = address[i];
			}
			buffer[index++] = '\r';
			_connector_write(robot->connector, buffer, _MOTORING_PACKET_LENGTH);
		}
	} else {
		buffer[index++] = '1';
		if(hamster->motor_mode == _HAMSTER_S_MOTOR_WAVE) {
			buffer[index++] = '1';
		} else if(hamster->motor_mode == _HAMSTER_S_MOTOR_HALF) {
			buffer[index++] = '2';
		} else if(hamster->motor_mode == _HAMSTER_S_MOTOR_FULL) {
			buffer[index++] = '3';
		} else {
			buffer[index++] = '0';
		}
		if(hamster->left_wheel == 0 && hamster->right_wheel == 0) {
			if(hamster->wheel_move == 1 && ++hamster->wheel_move_count > 5) {
				hamster->wheel_move = 0;
			}
		} else {
			hamster->wheel_move = 1;
			hamster->wheel_move_count = 0;
		}
		index = _value_to_hex(buffer, index, hamster->left_wheel, 1);
		index = _value_to_hex(buffer, index, hamster->right_wheel, 1);
		index = _value_to_hex(buffer, index, hamster->left_red, 1);
		index = _value_to_hex(buffer, index, hamster->left_green, 1);
		index = _value_to_hex(buffer, index, hamster->left_blue, 1);
		index = _value_to_hex(buffer, index, hamster->right_red, 1);
		index = _value_to_hex(buffer, index, hamster->right_green, 1);
		index = _value_to_hex(buffer, index, hamster->right_blue, 1);
		hamster->wheel_pulse = hamster->pulse;
		if(hamster->pulse_written == 1) {
			if(hamster->wheel_pulse != 0 || hamster->wheel_pulse_prev != 0) {
				hamster->wheel_id = (hamster->wheel_id % 255) + 1;
			}
			hamster->wheel_count = 0;
			if(hamster->wheel_pulse > 0) {
				hamster->wheel_event = 1;
			} else {
				hamster->wheel_event = 0;
			}
			hamster->wheel_pulse_prev = hamster->wheel_pulse;
			hamster->pulse_written = 0;
		}
		index = _value_to_hex(buffer, index, hamster->wheel_id, 1);
		index = _value_to_hex(buffer, index, hamster->wheel_pulse, 2);
		temp = hamster->line_tracer_mode & 0x0f;
		if(temp > 7) temp ++;
		if(hamster->line_tracer_mode_written == 1) {
			hamster->line_tracer_count = 0;
			if(temp > 0) {
				hamster->line_tracer_id = (hamster->line_tracer_id % 15) + 1;
				hamster->line_tracer_event = 1;
			} else {
				hamster->line_tracer_event = 0;
			}
			hamster->line_tracer_mode_written = 0;
		}
		temp |= (hamster->line_tracer_id & 0x0f) << 4;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->line_tracer_speed & 0x0f) << 4;
		temp |= hamster->line_tracer_gain & 0x0f;
		index = _value_to_hex(buffer, index, temp, 1);
		temp = (hamster->config_proximity & 0x07) << 5;
		temp |= (hamster->config_band_width & 0x07) << 2;
		temp |= hamster->config_gravity & 0x03;
		index = _value_to_hex(buffer, index, temp, 1);
		if(hamster->io_mode_a > 10) {
			temp = hamster->io_mode_a;
		} else {
			t = hamster->io_mode_a;
			if(t == _HAMSTER_S_IO_MODE_VOLTAGE_INPUT) t = _HAMSTER_S_IO_MODE_ANALOG_INPUT_ABSOLUTE;
			temp = (t & 0x0f) << 4;
			t = hamster->io_mode_b;
			if(t == _HAMSTER_S_IO_MODE_VOLTAGE_INPUT) t = _HAMSTER_S_IO_MODE_ANALOG_INPUT_ABSOLUTE;
			temp |= t & 0x0f;
		}
		index = _value_to_hex(buffer, index, temp, 1);
		index = _value_to_hex(buffer, index, hamster->output_a, 1);
		index = _value_to_hex(buffer, index, hamster->output_b, 1);
		temp = 0;
		switch(hamster->sound) {
			case _HAMSTER_S_SOUND_OFF: temp = 0x00; break;
			case _HAMSTER_S_SOUND_BEEP: temp = 0x01; break;
			case _HAMSTER_S_SOUND_RANDOM_BEEP: temp = 0x05; break;
			case _HAMSTER_S_SOUND_NOISE: temp = 0x07; break;
			case _HAMSTER_S_SOUND_SIREN: temp = 0x09; break;
			case _HAMSTER_S_SOUND_ENGINE: temp = 0x0b; break;
			case _HAMSTER_S_SOUND_CHOP: temp = 0x12; break;
			case _HAMSTER_S_SOUND_ROBOT: temp = 0x20; break;
			case _HAMSTER_S_SOUND_DIBIDIBIDIP: temp = 0x21; break;
			case _HAMSTER_S_SOUND_GOOD_JOB: temp = 0x23; break;
			case _HAMSTER_S_SOUND_HAPPY: temp = 0x30; break;
			case _HAMSTER_S_SOUND_ANGRY: temp = 0x31; break;
			case _HAMSTER_S_SOUND_SAD: temp = 0x32; break;
			case _HAMSTER_S_SOUND_SLEEP: temp = 0x33; break;
			case _HAMSTER_S_SOUND_MARCH: temp = 0x34; break;
			case _HAMSTER_S_SOUND_BIRTHDAY: temp = 0x35; break;
		}
		if(hamster->sound_written == 1) {
			hamster->sound_count = 0;
			if(temp > 0) {
				hamster->sound_flag ^= 0x80;
				hamster->sound_event = 1;
			} else {
				hamster->sound_event = 0;
			}
			hamster->sound_written = 0;
		}
		if(temp > 0) {
			buffer[index++] = '0';
			buffer[index++] = '0';
			index = _value_to_hex(buffer, index, temp | hamster->sound_flag, 1);
		} else if(hamster->note > 0) {
			buffer[index++] = '0';
			buffer[index++] = '1';
			index = _value_to_hex(buffer, index, hamster->note, 1);
		} else {
			index = _value_to_hex(buffer, index, _round_float_value(hamster->buzzer * 10) + 512, 2);
		}
		buffer[index++] = '-';
		address = _connector_get_address(robot->connector);
		for(i = 0; i < 12; ++i) {
			buffer[index++] = address[i];
		}
		buffer[index++] = '\r';
		hamster->packet_sent = 1;
		_connector_write(robot->connector, buffer, _MOTORING_PACKET_LENGTH);
	}
}

void _hamster_s_run_sound(struct _robot* robot, int sound, int repeat) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	if(repeat < 0) repeat = -1;
	if(repeat != 0) {
		hamster->current_sound = sound;
		hamster->sound_repeat = repeat;
		_robot_write(robot, HAMSTER_S_SOUND, sound);
	}
}

int _hamster_s_decode_sensory_packet(struct _robot* robot, char* packet) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	struct _device** devices = robot->devices;
	char* buffer = packet;
	int* temp;
	int value, value2, acc_x, acc_y, acc_z, id, state, i, j, len;
	
	hamster->packet_received = 0;
	value = _hex_to_value(buffer, 0, 1);
	if(value == 1) {
		value = _hex_to_value(buffer, 2, 6);
		if(value != hamster->event_pulse_count) {
			_device_put(devices[_HAMSTER_S_PULSE_COUNT_INDEX], value, (hamster->event_pulse_count == -1) ? 0 : 1);
			hamster->event_pulse_count = value;
		}
		_device_put(devices[_HAMSTER_S_LEFT_PROXIMITY_INDEX], _hex_to_value(buffer, 6, 8), 1); // left proximity
		_device_put(devices[_HAMSTER_S_RIGHT_PROXIMITY_INDEX], _hex_to_value(buffer, 8, 10), 1); // right proximity
		value2 = _hex_to_value(buffer, 38, 40);
		if((value2 & 0x01) == 0) {
			hamster->light = _hex_to_value(buffer, 10, 14);
		} else {
			value = _hex_to_value(buffer, 10, 12);
			if(value > 0x7F) value -= 0x100;
			hamster->temperature = _round_float_value(value / 2.0f + 23);
		}
		_device_put(devices[_HAMSTER_S_LIGHT_INDEX], hamster->light, 1);
		_device_put(devices[_HAMSTER_S_TEMPERATURE_INDEX], hamster->temperature, 1);
		_device_put(devices[_HAMSTER_S_LEFT_FLOOR_INDEX], _hex_to_value(buffer, 14, 16), 1); // left floor
		_device_put(devices[_HAMSTER_S_RIGHT_FLOOR_INDEX], _hex_to_value(buffer, 16, 18), 1); // right floor
		acc_x = _hex_to_value(buffer, 18, 22);
		if(acc_x > 0x7FFF) acc_x -= 0x10000;
		_device_put_at(devices[_HAMSTER_S_ACCELERATION_INDEX], 0, acc_x, 1);
		acc_y = _hex_to_value(buffer, 22, 26);
		if(acc_y > 0x7FFF) acc_y -= 0x10000;
		_device_put_at(devices[_HAMSTER_S_ACCELERATION_INDEX], 1, acc_y, 1);
		acc_z = _hex_to_value(buffer, 26, 30);
		if(acc_z > 0x7FFF) acc_z -= 0x10000;
		_device_put_at(devices[_HAMSTER_S_ACCELERATION_INDEX], 2, acc_z, 1);
		if(acc_z < 8192 && acc_x > 8192 && acc_y > -4096 && acc_y < 4096) value = 1;
		else if(acc_z < 8192 && acc_x < -8192 && acc_y > -4096 && acc_y < 4096) value = -1;
		else if(acc_z < 8192 && acc_y > 8192 && acc_x > -4096 && acc_x < 4096) value = 2;
		else if(acc_z < 8192 && acc_y < -8192 && acc_x > -4096 && acc_x < 4096) value = -2;
		else if(acc_z > 12288 && acc_x > -8192 && acc_x < 8192 && acc_y > -8192 && acc_y < 8192) value = 3;
		else if(acc_z < -12288 && acc_x > -4096 && acc_x < 4096 && acc_y > -4096 && acc_y < 4096) value = -3;
		else value = 0;
		if(value != hamster->event_tilt) {
			_device_put(devices[_HAMSTER_S_TILT_INDEX], value, (hamster->event_tilt == -4) ? 0 : 1);
			hamster->event_tilt = value;
		}
		value = _hex_to_value(buffer, 30, 32);
		_device_put(devices[_HAMSTER_S_INPUT_A_INDEX], value, 1);
		value = _hex_to_value(buffer, 32, 34);
		_device_put(devices[_HAMSTER_S_INPUT_B_INDEX], value, 1);
		value = _hex_to_value(buffer, 34, 36);
		id = (value >> 6) & 0x03;
		if(id != hamster->event_free_fall_id) {
			_device_put_empty(devices[_HAMSTER_S_FREE_FALL_INDEX], (hamster->event_free_fall_id == -1) ? 0 : 1);
			hamster->event_free_fall_id = id;
		}
		id = (value >> 4) & 0x03;
		if(id != hamster->event_tap_id) {
			_device_put_empty(devices[_HAMSTER_S_TAP_INDEX], (hamster->event_tap_id == -1 || hamster->wheel_move == 1) ? 0 : 1);
			hamster->event_tap_id = id;
		}
		state = (value >> 2) & 0x03;
		if(hamster->wheel_event == 1) {
			if(state == 2) {
				if(hamster->wheel_pulse > 0 && hamster->wheel_pulse < 25) {
					if(++hamster->wheel_count > 8) hamster->wheel_event = 2;
				}
			} else if(state == 3) {
				hamster->wheel_event = 2;
			}
		}
		if(hamster->wheel_event == 2) {
			if(state != hamster->wheel_state || hamster->wheel_count > 8) {
				hamster->wheel_state = state;
				_device_put(devices[_HAMSTER_S_WHEEL_STATE_INDEX], state, 1);
				if(state == 2) {
					hamster->wheel_event = 0;
					hamster->wheel_count = 0;
				}
			}
		}
		state = (value >> 1) & 0x01;
		if(hamster->sound_event == 1) {
			if(state == 0) {
				if(++hamster->sound_count > 5) hamster->sound_event = 2;
			} else {
				hamster->sound_event = 2;
			}
		}
		if(hamster->sound_event == 2) {
			if(state != hamster->sound_state || hamster->sound_count > 5) {
				hamster->sound_state = state;
				if(state == 0) {
					hamster->sound_event = 0;
					hamster->sound_count = 0;
					if(hamster->current_sound > 0) {
						if(hamster->sound_repeat < 0) {
							_hamster_s_run_sound(robot, hamster->current_sound, -1);
						} else if(hamster->sound_repeat > 1) {
							hamster->sound_repeat --;
							_hamster_s_run_sound(robot, hamster->current_sound, hamster->sound_repeat);
						} else {
							hamster->current_sound = 0;
							hamster->sound_repeat = 1;
							_device_put(devices[_HAMSTER_S_SOUND_STATE_INDEX], state, 1);
						}
					} else {
						hamster->current_sound = 0;
						hamster->sound_repeat = 1;
					}
				}
			}
		}
		value = _hex_to_value(buffer, 36, 38);
		value -= 0x100;
		_device_put(devices[_HAMSTER_S_SIGNAL_STRENGTH_INDEX], value, 1);
		state = (value2 >> 6) & 0x03;
		if((state & 0x02) != 0) {
			if(hamster->line_tracer_event == 1) {
				if(state == 0x02) {
					if(++hamster->line_tracer_count > 5) hamster->line_tracer_event = 2;
				} else {
					hamster->line_tracer_event = 2;
				}
			}
			if(hamster->line_tracer_event == 2) {
				if(state != hamster->line_tracer_state || hamster->line_tracer_count > 5) {
					hamster->line_tracer_state = state;
					_device_put(devices[_HAMSTER_S_LINE_TRACER_STATE_INDEX], state << 5, 1);
					if(state == 0x02) {
						hamster->line_tracer_event = 0;
						hamster->line_tracer_count = 0;
					}
				}
			}
		}
		state = (value2 >> 1) & 0x03;
		if(state == 0) state = 2;
		else if(state >= 2) state = 0;
		if(state != hamster->event_battery_state) {
			_device_put(devices[_HAMSTER_S_BATTERY_STATE_INDEX], state, (hamster->event_battery_state == -1) ? 0 : 1);
			hamster->event_battery_state = state;
		}
		id = (value2 >> 5) & 0x01;
		if(id != hamster->port_ack_id) {
			if(hamster->port_ack_id != -1) {
				switch(hamster->io_mode_a) {
					case _HAMSTER_S_IO_MODE_SERIAL_9600:
					case _HAMSTER_S_IO_MODE_SERIAL_14400:
					case _HAMSTER_S_IO_MODE_SERIAL_19200:
					case _HAMSTER_S_IO_MODE_SERIAL_28800:
					case _HAMSTER_S_IO_MODE_SERIAL_38400:
					case _HAMSTER_S_IO_MODE_SERIAL_57600:
					case _HAMSTER_S_IO_MODE_SERIAL_76800:
					case _HAMSTER_S_IO_MODE_SERIAL_115200:
						hamster->port_serial = 1;
						break;
					default:
						hamster->port_serial = 0;
						break;
				}
			}
			hamster->port_ack_id = id;
		}
		hamster->packet_received = 1;
	} else if(value == 2) {
		id = _hex_to_value(buffer, 1, 2);
		if(id != hamster->event_serial_id) {
			if(hamster->event_serial_id != -1) {
				len = _hex_to_value(buffer, 2, 4);
				if(len > 18) len = 18;
				hamster->read_serial_data[0] = len;
				for(i = 1, j = 4; i <= len; ++i, j += 2) {
					hamster->read_serial_data[i] = _hex_to_value(buffer, j, j + 2);
				}
				for(i = len + 1; i <= 18; ++i) {
					hamster->read_serial_data[i] = 0;
				}
				_device_put_array(devices[_HAMSTER_S_READ_SERIAL_INDEX], hamster->read_serial_data, 19, 1);
			}
			hamster->event_serial_id = id;
		}
		hamster->packet_received = 3;
	}
	if(hamster->serial_send_id != hamster->serial_send_prev_id) {
		hamster->serial_send_prev_id = hamster->serial_send_id;
		_device_put_empty(devices[_HAMSTER_S_SERIAL_STATE_INDEX], 1);
	}
	if(_device_e(devices[_HAMSTER_S_READ_SERIAL_INDEX]) == 1) {
		_read_queue_push(&hamster->read_queue, hamster->read_serial_data, 0);
	}
	if(_device_e(devices[_HAMSTER_S_SERIAL_STATE_INDEX]) == 1) {
		temp = _write_queue_pop(&hamster->write_queue);
		if(temp == NULL) {
			hamster->write_serial_id = (hamster->write_serial_id % 255) + 1;
		} else {
			_device_write_array(devices[_HAMSTER_S_WRITE_SERIAL_INDEX], temp, 19);
		}
	}
	return 1;
}

int _hamster_s_receive(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		int read_bytes = _connector_read(connector);
		if(read_bytes != 0) {
			if(_hamster_s_decode_sensory_packet(robot, connector->buffer) == 1) {
				if(robot->ready == 0) {
					robot->ready = 1;
					_runner_register_checked();
				}
			}
			return 1;
		}
	}
	return 0;
}

void _hamster_s_send(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		_hamster_s_encode_motoring_packet(robot);
	}
}

void _hamster_s_update_sensory_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _HAMSTER_S_SIGNAL_STRENGTH_INDEX; i <= _HAMSTER_S_SERIAL_STATE_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _hamster_s_update_motoring_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _HAMSTER_S_LEFT_WHEEL_INDEX; i <= _HAMSTER_S_CONFIG_BAND_WIDTH_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _hamster_s_reset(struct _robot* robot) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_reset(robot);
	
	hamster->left_wheel = 0;
	hamster->right_wheel = 0;
	hamster->left_red = 0;
	hamster->left_green = 0;
	hamster->left_blue = 0;
	hamster->right_red = 0;
	hamster->right_green = 0;
	hamster->right_blue = 0;
	hamster->buzzer = 0;
	hamster->output_a = 0;
	hamster->output_b = 0;
	hamster->pulse = 0;
	hamster->note = 0;
	hamster->sound = 0;
	hamster->line_tracer_mode = 0;
	hamster->line_tracer_gain = 4;
	hamster->line_tracer_speed = 5;
	hamster->io_mode_a = 0;
	hamster->io_mode_b = 0;
	hamster->motor_mode = 0;
	hamster->config_proximity = 2;
	hamster->config_gravity = 0;
	hamster->config_band_width = 0;
	
	hamster->light = 0;
	hamster->temperature = 0;
	
	//hamster->wheel_id = 0;
	hamster->wheel_pulse = 0;
	hamster->wheel_pulse_prev = -1;
	hamster->wheel_event = 0;
	hamster->wheel_state = 0;
	hamster->wheel_count = 0;
	hamster->wheel_move = 0;
	hamster->wheel_move_count = 0;
	hamster->pulse_written = 0;
	
	hamster->current_sound = 0;
	hamster->sound_repeat = 1;
	//hamster->sound_flag = 0;
	hamster->sound_event = 0;
	hamster->sound_state = 0;
	hamster->sound_count = 0;
	hamster->sound_written = 0;
	
	//hamster->line_tracer_id = 0;
	hamster->line_tracer_event = 0;
	hamster->line_tracer_state = 0;
	hamster->line_tracer_count = 0;
	hamster->line_tracer_mode_written = 0;
	
	hamster->event_free_fall_id = -1;
	hamster->event_tap_id = -1;
	hamster->event_tilt = -4;
	hamster->event_serial_id = -1;
	hamster->event_pulse_count = -1;
	hamster->event_battery_state = -1;
	
	//hamster->command_serial_id = 0;
	hamster->write_serial_written = 0;
	
	hamster->port_ack_id = -1;
	hamster->port_serial = 0;
	hamster->serial_send_id = 0;
	hamster->serial_send_prev_id = 0;
	hamster->packet_sent = 0;
	hamster->packet_received = 0;
	
	//hamster->write_serial_id = 0;
	hamster->bpm = 60;
	hamster->gain = -1;
	hamster->speed = 5;
	hamster->board_count = 0;
	hamster->board_state = 0;
	hamster->serial_baud = _HAMSTER_S_IO_MODE_SERIAL_9600;
	
	_write_queue_reset(&hamster->write_queue);
	_read_queue_reset(&hamster->read_queue);
}

void _hamster_s_dispose(struct _robot* robot) {
	_robot_group_remove_robot(_GROUP_HAMSTER_S, robot);
	_runner_remove_robot(robot);

	robot->running = 0;
	if(robot->thread_alive == 1) {
		WaitForSingleObject(robot->thread_handle, 1000);
	}
	Sleep(100);
	_robot_dispose(robot);
	free((struct _hamster_s_robot*)robot);
}

unsigned WINAPI _hamster_s_thread_proc(void* arg) {
	struct _robot* robot = (struct _robot*)arg;
	int shutdown = 0, shutdown_count = 0;
	
	if(robot == NULL) return 0;

	robot->thread_alive = 1;
	while(1) {
		if(_hamster_s_receive(robot) == 1) {
			_hamster_s_send(robot);
			if(shutdown == 1) {
				break;
			}
		}
		if(shutdown == 1) {
			if(++shutdown_count > 5) {
				break;
			}
		}
		if(robot->running == 0) {
			shutdown = 1;
		}
		Sleep(5);
	}
	robot->thread_alive = 0;
	return 0;
}

struct _hamster_s_robot* _hamster_s_create(const char* port_name) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)malloc(sizeof(struct _hamster_s_robot));
	struct _robot* robot = (struct _robot*)hamster;
	struct _device** devices;
	struct _connector* connector;
	unsigned int thread_id;
	
	_robot_init(robot, _robot_group_count_robots(_GROUP_HAMSTER_S), "HamsterS", 55);
	
	hamster->left_wheel = 0;
	hamster->right_wheel = 0;
	hamster->left_red = 0;
	hamster->left_green = 0;
	hamster->left_blue = 0;
	hamster->right_red = 0;
	hamster->right_green = 0;
	hamster->right_blue = 0;
	hamster->buzzer = 0;
	hamster->output_a = 0;
	hamster->output_b = 0;
	hamster->pulse = 0;
	hamster->note = 0;
	hamster->sound = 0;
	hamster->line_tracer_mode = 0;
	hamster->line_tracer_gain = 4;
	hamster->line_tracer_speed = 5;
	hamster->io_mode_a = 0;
	hamster->io_mode_b = 0;
	hamster->motor_mode = 0;
	hamster->config_proximity = 2;
	hamster->config_gravity = 0;
	hamster->config_band_width = 0;
	
	hamster->light = 0;
	hamster->temperature = 0;
	
	hamster->wheel_id = 0;
	hamster->wheel_pulse = 0;
	hamster->wheel_pulse_prev = -1;
	hamster->wheel_event = 0;
	hamster->wheel_state = 0;
	hamster->wheel_count = 0;
	hamster->wheel_move = 0;
	hamster->wheel_move_count = 0;
	hamster->pulse_written = 0;

	hamster->current_sound = 0;
	hamster->sound_repeat = 1;
	hamster->sound_flag = 0;
	hamster->sound_event = 0;
	hamster->sound_state = 0;
	hamster->sound_count = 0;
	hamster->sound_written = 0;
	
	hamster->line_tracer_id = 0;
	hamster->line_tracer_event = 0;
	hamster->line_tracer_state = 0;
	hamster->line_tracer_count = 0;
	hamster->line_tracer_mode_written = 0;
	
	hamster->event_free_fall_id = -1;
	hamster->event_tap_id = -1;
	hamster->event_tilt = -4;
	hamster->event_serial_id = -1;
	hamster->event_pulse_count = -1;
	hamster->event_battery_state = -1;
	
	hamster->command_serial_id = 0;
	hamster->write_serial_written = 0;
	
	hamster->port_ack_id = -1;
	hamster->port_serial = 0;
	hamster->serial_send_id = 0;
	hamster->serial_send_prev_id = 0;
	hamster->packet_sent = 0;
	hamster->packet_received = 0;
	
	hamster->write_serial_id = 0;
	hamster->bpm = 60;
	hamster->gain = -1;
	hamster->speed = 5;
	hamster->board_count = 0;
	hamster->board_state = 0;
	hamster->serial_baud = _HAMSTER_S_IO_MODE_SERIAL_9600;
	
	_write_queue_reset(&hamster->write_queue);
	_read_queue_reset(&hamster->read_queue);
	
	robot->find_device = _hamster_s_find_device;
	robot->request_motoring_data = _hamster_s_request_motoring_data;
	robot->update_sensory_device_state = _hamster_s_update_sensory_device_state;
	robot->update_motoring_device_state = _hamster_s_update_motoring_device_state;
	robot->reset = _hamster_s_reset;
	robot->dispose = _hamster_s_dispose;
	
	robot->devices_size = 40;
	devices = (struct _device**)malloc(sizeof(struct _device*) * 40);
	devices[0] = _device_create(HAMSTER_S_LEFT_WHEEL, "LeftWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, -100, 100, 0);
	devices[1] = _device_create(HAMSTER_S_RIGHT_WHEEL, "RightWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, -100, 100, 0);
	devices[2] = _device_create(HAMSTER_S_LEFT_RGB, "LeftRgb", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 3, 0, 255, 0);
	devices[3] = _device_create(HAMSTER_S_RIGHT_RGB, "RightRgb", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 3, 0, 255, 0);
	devices[4] = _device_create(HAMSTER_S_BUZZER, "Buzzer", DEVICE_TYPE_EFFECTOR, DATA_TYPE_FLOAT, 1, 0, 167746.0f, 0);
	devices[5] = _device_create(HAMSTER_S_OUTPUT_A, "OutputA", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[6] = _device_create(HAMSTER_S_OUTPUT_B, "OutputB", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[7] = _device_create(HAMSTER_S_PULSE, "Pulse", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[8] = _device_create(HAMSTER_S_NOTE, "Note", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 88, 0);
	devices[9] = _device_create(HAMSTER_S_SOUND, "Sound", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 127, 0);
	devices[10] = _device_create(HAMSTER_S_LINE_TRACER_MODE, "LineTracerMode", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 15, 0);
	devices[11] = _device_create(HAMSTER_S_LINE_TRACER_GAIN, "LineTracerGain", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 10, 4);
	devices[12] = _device_create(HAMSTER_S_LINE_TRACER_SPEED, "LineTracerSpeed", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 10, 5);
	devices[13] = _device_create(HAMSTER_S_IO_MODE_A, "IoModeA", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[14] = _device_create(HAMSTER_S_IO_MODE_B, "IoModeB", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[15] = _device_create(HAMSTER_S_WRITE_SERIAL, "WriteSerial", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 19, 0, 255, 0);
	devices[16] = _device_create(HAMSTER_S_MOTOR_MODE, "MotorMode", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[17] = _device_create(HAMSTER_S_CONFIG_PROXIMITY, "ConfigProximity", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 7, 2);
	devices[18] = _device_create(HAMSTER_S_CONFIG_GRAVITY, "ConfigGravity", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[19] = _device_create(HAMSTER_S_CONFIG_BAND_WIDTH, "ConfigBandWidth", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 7, 0);
	devices[20] = _device_create(HAMSTER_S_SIGNAL_STRENGTH, "SignalStrength", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -128, 0, 0);
	devices[21] = _device_create(HAMSTER_S_LEFT_PROXIMITY, "LeftProximity", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[22] = _device_create(HAMSTER_S_RIGHT_PROXIMITY, "RightProximity", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[23] = _device_create(HAMSTER_S_LEFT_FLOOR, "LeftFloor", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[24] = _device_create(HAMSTER_S_RIGHT_FLOOR, "RightFloor", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[25] = _device_create(HAMSTER_S_ACCELERATION, "Acceleration", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 3, -32768, 32767, 0);
	devices[26] = _device_create(HAMSTER_S_LIGHT, "Light", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[27] = _device_create(HAMSTER_S_TEMPERATURE, "Temperature", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -41, 87, 0);
	devices[28] = _device_create(HAMSTER_S_INPUT_A, "InputA", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[29] = _device_create(HAMSTER_S_INPUT_B, "InputB", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[30] = _device_create(HAMSTER_S_FREE_FALL, "FreeFall", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	devices[31] = _device_create(HAMSTER_S_TAP, "Tap", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	devices[32] = _device_create(HAMSTER_S_TILT, "Tilt", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, -3, 3, 0);
	devices[33] = _device_create(HAMSTER_S_READ_SERIAL, "ReadSerial", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 19, 0, 255, 0);
	devices[34] = _device_create(HAMSTER_S_PULSE_COUNT, "PulseCount", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[35] = _device_create(HAMSTER_S_WHEEL_STATE, "WheelState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[36] = _device_create(HAMSTER_S_SOUND_STATE, "SoundState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 1, 0);
	devices[37] = _device_create(HAMSTER_S_LINE_TRACER_STATE, "LineTracerState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[38] = _device_create(HAMSTER_S_BATTERY_STATE, "BatteryState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 2, 2);
	devices[39] = _device_create(HAMSTER_S_SERIAL_STATE, "SerialState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	robot->devices = devices;
	robot->connector = NULL;
	robot->alive = 1;

	_robot_group_add_robot(_GROUP_HAMSTER_S, robot);
	_runner_add_robot(robot);
	_runner_start();
	
	_runner_register_required();
	robot->running = 1;
	robot->thread_handle = (HANDLE)_beginthreadex(NULL,
		0,
		_hamster_s_thread_proc,
		robot,
		CREATE_SUSPENDED,
		&thread_id);
	if(robot->thread_handle != 0) {
		ResumeThread(robot->thread_handle);
	}

	connector = _connector_create("HamsterS", robot->index, _VALID_PACKET_LENGTH, _CR);
	robot->connector = connector;
	if(connector != NULL) {
		int result;
		connector->arg = (void*)robot;
		connector->check_connection = _hamster_s_check_connection;
		result = _connector_open(connector, port_name, _SERIAL_BAUDRATE_115200, _SERIAL_FLOWCONTROL_RTSCTS_IN | _SERIAL_FLOWCONTROL_RTSCTS_OUT);
		if(result == _CONNECTION_RESULT_FOUND) {
			while(robot->ready == 0 && robot->running == 1) {
				Sleep(10);
			}
		} else if(result == _CONNECTION_RESULT_NOT_AVAILABLE) {
			_runner_register_checked();
		}
	}
	return hamster;
}

const char* _hamster_s_get_id(void) {
	return HAMSTER_S_ID;
}

int _hamster_s_get_motor_mode_value(const char* mode) {
	if(mode == NULL) return -1;
	if(_STRICMP(mode, HAMSTER_S_MOTOR_AUTO) == 0) return _HAMSTER_S_MOTOR_AUTO;
	if(_STRICMP(mode, HAMSTER_S_MOTOR_WAVE) == 0) return _HAMSTER_S_MOTOR_WAVE;
	if(_STRICMP(mode, HAMSTER_S_MOTOR_HALF) == 0) return _HAMSTER_S_MOTOR_HALF;
	if(_STRICMP(mode, HAMSTER_S_MOTOR_FULL) == 0) return _HAMSTER_S_MOTOR_FULL;
	return -1;
}

void _hamster_s_motor_mode(int hamster_s_index, const char* mode) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_s_get_motor_mode_value(mode);
	if(value >= 0) _robot_write(robot, HAMSTER_S_MOTOR_MODE, value);
}

void _hamster_s_wheels(int hamster_s_index, double left_velocity, double right_velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, _round_double_value(left_velocity));
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, _round_double_value(right_velocity));
}

void _hamster_s_left_wheel(int hamster_s_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, _round_double_value(velocity));
}

void _hamster_s_right_wheel(int hamster_s_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, _round_double_value(velocity));
}

void _hamster_s_stop(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
}

int _hamster_s_wheel_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, HAMSTER_S_WHEEL_STATE) == 1 && _robot_read(robot, HAMSTER_S_WHEEL_STATE) == 2) ? 1 : 0;
}

void _hamster_s_motion(struct _robot* robot, int pulse, double left_velocity, double right_velocity, int (*callback)(void* arg)) {
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	if(pulse > 0) {
		_robot_write(robot, HAMSTER_S_LEFT_WHEEL, _round_double_value(left_velocity));
		_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, _round_double_value(right_velocity));
		_robot_write(robot, HAMSTER_S_PULSE, pulse);
		wait_until(callback, NULL);
		_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
		_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
	} else {
		_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
		_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
		_robot_write(robot, HAMSTER_S_PULSE, 0);
	}
}

void _hamster_s_move_forward(int hamster_s_index, double cm, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(cm < 0) {
		cm = -cm;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_cm_to_pulse(cm), velocity, velocity, callback);
}

void _hamster_s_move_backward(int hamster_s_index, double cm, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(cm < 0) {
		cm = -cm;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_cm_to_pulse(cm), -velocity, -velocity, callback);
}

void _hamster_s_turn_left(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_calc_turn_pulse(degree), -velocity, velocity, callback);
}

void _hamster_s_turn_right(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_calc_turn_pulse(degree), velocity, -velocity, callback);
}

void _hamster_s_pivot_left_wheel(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_calc_pivot_pulse(degree), 0, velocity, callback);
}

void _hamster_s_pivot_right_wheel(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _hamster_s_calc_pivot_pulse(degree), velocity, 0, callback);
}

void _hamster_s_pivot_left_pen(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	left_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	pulse = _hamster_s_calc_pivot_pen_pulse(degree, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, pulse, left_velocity, velocity, callback);
}

void _hamster_s_pivot_right_pen(int hamster_s_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	right_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	pulse = _hamster_s_calc_pivot_pen_pulse(degree, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, pulse, velocity, right_velocity, callback);
}

void _hamster_s_circle_left(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, left_velocity, velocity, callback);
}

void _hamster_s_circle_right(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, velocity, right_velocity, callback);
}

void _hamster_s_left_pen_circle_left(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	radius += _HAMSTER_S_PEN_CENTER_DISTANCE;
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, left_velocity, velocity, callback);
}

void _hamster_s_left_pen_circle_right(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		radius -= _HAMSTER_S_PEN_CENTER_DISTANCE;
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	} else {
		radius = _HAMSTER_S_PEN_CENTER_DISTANCE - radius;
		left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
		right_velocity = velocity;
	}
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, left_velocity, right_velocity, callback);
}

void _hamster_s_right_pen_circle_left(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		radius -= _HAMSTER_S_PEN_CENTER_DISTANCE;
		left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
		right_velocity = velocity;
	} else {
		radius = _HAMSTER_S_PEN_CENTER_DISTANCE - radius;
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	}
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, left_velocity, right_velocity, callback);
}

void _hamster_s_right_pen_circle_right(int hamster_s_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	int pulse;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	radius += _HAMSTER_S_PEN_CENTER_DISTANCE;
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	pulse = _hamster_s_calc_circle_pulse(degree, radius);
	_hamster_s_motion(robot, pulse, velocity, right_velocity, callback);
}

void _hamster_s_motion_sec(struct _robot* robot, double sec, double left_velocity, double right_velocity) {
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	if(sec < 0) {
		sec = -sec;
		left_velocity = -left_velocity;
		right_velocity = -right_velocity;
	}
	if(sec > 0) {
		_robot_write(robot, HAMSTER_S_LEFT_WHEEL, _round_double_value(left_velocity));
		_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, _round_double_value(right_velocity));
		wait(sec * 1000);
	}
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
}

void _hamster_s_move_forward_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, velocity, velocity);
}

void _hamster_s_move_backward_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, -velocity, -velocity);
}

void _hamster_s_turn_left_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, -velocity, velocity);
}

void _hamster_s_turn_right_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, velocity, -velocity);
}

void _hamster_s_pivot_left_wheel_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, 0, velocity);
}

void _hamster_s_pivot_right_wheel_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_hamster_s_motion_sec(robot, sec, velocity, 0);
}

void _hamster_s_pivot_left_pen_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	left_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion_sec(robot, sec, left_velocity, velocity);
}

void _hamster_s_pivot_right_pen_sec(int hamster_s_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	
	if(robot == NULL) return;
	right_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion_sec(robot, sec, velocity, right_velocity);
}

void _hamster_s_circle_left_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion_sec(robot, sec, left_velocity, velocity);
}

void _hamster_s_circle_right_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion_sec(robot, sec, velocity, right_velocity);
}

void _hamster_s_left_pen_circle_left_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	radius += _HAMSTER_S_PEN_CENTER_DISTANCE;
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion_sec(robot, sec, left_velocity, velocity);
}

void _hamster_s_left_pen_circle_right_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		radius -= _HAMSTER_S_PEN_CENTER_DISTANCE;
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	} else {
		radius = _HAMSTER_S_PEN_CENTER_DISTANCE - radius;
		left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
		right_velocity = velocity;
	}
	_hamster_s_motion_sec(robot, sec, left_velocity, right_velocity);
}

void _hamster_s_right_pen_circle_left_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		radius -= _HAMSTER_S_PEN_CENTER_DISTANCE;
		left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
		right_velocity = velocity;
	} else {
		radius = _HAMSTER_S_PEN_CENTER_DISTANCE - radius;
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	}
	_hamster_s_motion_sec(robot, sec, left_velocity, right_velocity);
}

void _hamster_s_right_pen_circle_right_sec(int hamster_s_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	radius += _HAMSTER_S_PEN_CENTER_DISTANCE;
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion_sec(robot, sec, velocity, right_velocity);
}

void _hamster_s_move_forward_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, velocity, callback);
}

void _hamster_s_move_backward_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), -velocity, -velocity, callback);
}

void _hamster_s_turn_left_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), -velocity, velocity, callback);
}

void _hamster_s_turn_right_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, -velocity, callback);
}

void _hamster_s_pivot_left_wheel_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), 0, velocity, callback);
}

void _hamster_s_pivot_right_wheel_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, 0, callback);
}

void _hamster_s_pivot_left_pen_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	left_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, _round_double_value(pulse), left_velocity, velocity, callback);
}

void _hamster_s_pivot_right_pen_pulse(int hamster_s_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	float right_velocity;
	
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	right_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, right_velocity, callback);
}

void _hamster_s_circle_left_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion(robot, _round_double_value(pulse), left_velocity, velocity, callback);
}

void _hamster_s_circle_right_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	float right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius);
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, right_velocity, callback);
}

void _hamster_s_left_pen_circle_left_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	left_velocity = _hamster_s_calc_inner_velocity(velocity, radius + _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, _round_double_value(pulse), left_velocity, velocity, callback);
}

void _hamster_s_left_pen_circle_right_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, radius - _HAMSTER_S_PEN_CENTER_DISTANCE);
	} else {
		left_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE - radius);
		right_velocity = velocity;
	}
	_hamster_s_motion(robot, _round_double_value(pulse), left_velocity, right_velocity, callback);
}

void _hamster_s_right_pen_circle_left_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double left_velocity, right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	if(radius >= _HAMSTER_S_PEN_CENTER_DISTANCE) {
		left_velocity = _hamster_s_calc_inner_velocity(velocity, radius - _HAMSTER_S_PEN_CENTER_DISTANCE);
		right_velocity = velocity;
	} else {
		left_velocity = velocity;
		right_velocity = _hamster_s_calc_inner_velocity(velocity, _HAMSTER_S_PEN_CENTER_DISTANCE - radius);
	}
	_hamster_s_motion(robot, _round_double_value(pulse), left_velocity, right_velocity, callback);
}

void _hamster_s_right_pen_circle_right_pulse(int hamster_s_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	double right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	right_velocity = _hamster_s_calc_inner_velocity(velocity, radius + _HAMSTER_S_PEN_CENTER_DISTANCE);
	_hamster_s_motion(robot, _round_double_value(pulse), velocity, right_velocity, callback);
}

int _hamster_s_line_tracer_mode_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, HAMSTER_S_LINE_TRACER_STATE) == 1 && _robot_read(robot, HAMSTER_S_LINE_TRACER_STATE) == 0x40) ? 1 : 0;
}

void _hamster_s_line_tracer_mode(int hamster_s_index, int mode, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, mode);
	
	switch(mode) {
		case _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_LEFT:
		case _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_RIGHT:
		case _HAMSTER_S_LINE_TRACER_MODE_BLACK_MOVE_FORWARD:
		case _HAMSTER_S_LINE_TRACER_MODE_BLACK_UTURN:
		case _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_LEFT:
		case _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_RIGHT:
		case _HAMSTER_S_LINE_TRACER_MODE_WHITE_MOVE_FORWARD:
		case _HAMSTER_S_LINE_TRACER_MODE_WHITE_UTURN: {
			wait_until(callback, NULL);
			_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
			break;
		}
	}
}

void _hamster_s_line_left(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_LEFT_SENSOR, NULL);
}

void _hamster_s_line_right(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_RIGHT_SENSOR, NULL);
}

void _hamster_s_line_both(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_BOTH_SENSORS, NULL);
}

void _hamster_s_cross_forward(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_MOVE_FORWARD, callback);
}

void _hamster_s_cross_left(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_LEFT, callback);
}

void _hamster_s_cross_right(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_TURN_RIGHT, callback);
}

void _hamster_s_cross_uturn(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_BLACK_UTURN, callback);
}

void _hamster_s_white_line_left(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_LEFT_SENSOR, NULL);
}

void _hamster_s_white_line_right(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_RIGHT_SENSOR, NULL);
}

void _hamster_s_white_line_both(int hamster_s_index) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_BOTH_SENSORS, NULL);
}

void _hamster_s_white_cross_forward(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_MOVE_FORWARD, callback);
}

void _hamster_s_white_cross_left(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_LEFT, callback);
}

void _hamster_s_white_cross_right(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_TURN_RIGHT, callback);
}

void _hamster_s_white_cross_uturn(int hamster_s_index, int (*callback)(void* arg)) {
	_hamster_s_line_tracer_mode(hamster_s_index, _HAMSTER_S_LINE_TRACER_MODE_WHITE_UTURN, callback);
}

void _hamster_s_line_gain_auto(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int _gain;
	
	if(hamster == NULL) return;
	hamster->gain = -1;
	_gain = _hamster_speed_to_gain(hamster->speed);
	if(_gain > 0) {
		_robot_write(robot, HAMSTER_S_LINE_TRACER_GAIN, _gain);
	}
}

void _hamster_s_line_gain(int hamster_s_index, double gain) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int _gain;
	
	if(hamster == NULL) return;
	_gain = _round_double_value(gain);
	if(_gain > 0) {
		hamster->gain = gain;
		_robot_write(robot, HAMSTER_S_LINE_TRACER_GAIN, _gain);
	} else {
		hamster->gain = -1;
		_gain = _hamster_speed_to_gain(hamster->speed);
		if(_gain > 0) {
			_robot_write(robot, HAMSTER_S_LINE_TRACER_GAIN, _gain);
		}
	}
}

void _hamster_s_line_speed(int hamster_s_index, double speed) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int _speed, _gain;
	
	if(hamster == NULL) return;
	_speed = _round_double_value(speed);
	_gain = hamster->gain;
	if(_gain < 0) _gain = _hamster_speed_to_gain(_speed);
	if(_speed > 0 && _gain > 0) {
		hamster->speed = _speed;
		_robot_write(robot, HAMSTER_S_LINE_TRACER_SPEED, _speed);
		_robot_write(robot, HAMSTER_S_LINE_TRACER_GAIN, _gain);
	}
}

int _hamster_s_board_forward_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				int left_floor = _robot_read(robot, HAMSTER_S_LEFT_FLOOR);
				int right_floor = _robot_read(robot, HAMSTER_S_RIGHT_FLOOR);
				int diff = _round_float_value((left_floor - right_floor) * 0.25f);
				if(left_floor < 50 && right_floor < 50) {
					hamster->board_count ++;
				} else {
					hamster->board_count = 0;
				}
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 45 + diff);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 45 - diff);
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(hamster->board_count < 10) {
				int diff = _round_float_value((_robot_read(robot, HAMSTER_S_LEFT_FLOOR) - _robot_read(robot, HAMSTER_S_RIGHT_FLOOR)) * 0.25f);
				hamster->board_count ++;
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 45 + diff);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 45 - diff);
			} else {
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
				return 1;
			}
			break;
		}
	}
	return 0;
}

void _hamster_s_board_forward(int hamster_s_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 45);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

int _hamster_s_board_left_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_S_LEFT_FLOOR) > 50) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(_robot_read(robot, HAMSTER_S_LEFT_FLOOR) < 20) {
				hamster->board_state = 3;
			}
			break;
		}
		case 3: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_S_LEFT_FLOOR) < 20) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 4;
			}
			break;
		}
		case 4: {
			if(_robot_read(robot, HAMSTER_S_LEFT_FLOOR) > 50) {
				hamster->board_state = 5;
			}
			break;
		}
		case 5: {
			int diff = _robot_read(robot, HAMSTER_S_LEFT_FLOOR) - _robot_read(robot, HAMSTER_S_RIGHT_FLOOR);
			if(diff > -15) {
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
				return 1;
			} else {
				diff = _round_float_value(diff * 0.5f);
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, diff);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, -diff);
			}
			break;
		}
	}
	return 0;
}

void _hamster_s_board_left(int hamster_s_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, -45);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

int _hamster_s_board_right_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return 0;
	switch(hamster->board_state) {
		case 1: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_S_RIGHT_FLOOR) > 50) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 2;
			}
			break;
		}
		case 2: {
			if(_robot_read(robot, HAMSTER_S_RIGHT_FLOOR) < 20) {
				hamster->board_state = 3;
			}
			break;
		}
		case 3: {
			if(hamster->board_count < 2) {
				if(_robot_read(robot, HAMSTER_S_RIGHT_FLOOR) < 20) {
					hamster->board_count ++;
				}
			} else {
				hamster->board_count = 0;
				hamster->board_state = 4;
			}
			break;
		}
		case 4: {
			if(_robot_read(robot, HAMSTER_S_RIGHT_FLOOR) > 50) {
				hamster->board_state = 5;
			}
			break;
		}
		case 5: {
			int diff = _robot_read(robot, HAMSTER_S_RIGHT_FLOOR) - _robot_read(robot, HAMSTER_S_LEFT_FLOOR);
			if(diff > -15) {
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 0);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, 0);
				return 1;
			} else {
				diff = _round_float_value(diff * 0.5f);
				_robot_write(robot, HAMSTER_S_LEFT_WHEEL, -diff);
				_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, diff);
			}
			break;
		}
	}
	return 0;
}

void _hamster_s_board_right(int hamster_s_index, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_S_PULSE, 0);
	_robot_write(robot, HAMSTER_S_LINE_TRACER_MODE, _HAMSTER_S_LINE_TRACER_MODE_OFF);
	_robot_write(robot, HAMSTER_S_LEFT_WHEEL, 45);
	_robot_write(robot, HAMSTER_S_RIGHT_WHEEL, -45);
	hamster->board_count = 0;
	hamster->board_state = 1;
	wait_until(callback, NULL);
}

void _hamster_s_rgbs(int hamster_s_index, double left_red, double left_green, double left_blue, double right_red, double right_green, double right_blue) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 0, _round_double_value(left_red));
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 1, _round_double_value(left_green));
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 2, _round_double_value(left_blue));
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 0, _round_double_value(right_red));
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 1, _round_double_value(right_green));
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 2, _round_double_value(right_blue));
}

void _hamster_s_left_rgb(int hamster_s_index, double red, double green, double blue) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 0, _round_double_value(red));
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 1, _round_double_value(green));
	_robot_write_at(robot, HAMSTER_S_LEFT_RGB, 2, _round_double_value(blue));
}

void _hamster_s_right_rgb(int hamster_s_index, double red, double green, double blue) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 0, _round_double_value(red));
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 1, _round_double_value(green));
	_robot_write_at(robot, HAMSTER_S_RIGHT_RGB, 2, _round_double_value(blue));
}

int _HAMSTER_S_COLOR_TO_RGB[][3] = {
	{ 0, 0, 0 },
	{ 255, 0, 0 },
	{ 255, 63, 0 },
	{ 255, 255, 0 },
	{ 0, 255, 0 },
	{ 0, 255, 255 },
	{ 0, 0, 255 },
	{ 63, 0, 255 },
	{ 255, 0, 255 },
	{ 255, 255, 255 }
};

int* _hamster_s_get_rgb(const char* color) {
	if(color == NULL) return NULL;
	if(_STRICMP(color, HAMSTER_S_LED_OFF) == 0) return _HAMSTER_S_COLOR_TO_RGB[0];
	if(_STRICMP(color, HAMSTER_S_LED_RED) == 0) return _HAMSTER_S_COLOR_TO_RGB[1];
	if(_STRICMP(color, HAMSTER_S_LED_ORANGE) == 0) return _HAMSTER_S_COLOR_TO_RGB[2];
	if(_STRICMP(color, HAMSTER_S_LED_YELLOW) == 0) return _HAMSTER_S_COLOR_TO_RGB[3];
	if(_STRICMP(color, HAMSTER_S_LED_GREEN) == 0) return _HAMSTER_S_COLOR_TO_RGB[4];
	if(_STRICMP(color, HAMSTER_S_LED_SKY_BLUE) == 0) return _HAMSTER_S_COLOR_TO_RGB[5];
	if(_STRICMP(color, HAMSTER_S_LED_BLUE) == 0) return _HAMSTER_S_COLOR_TO_RGB[6];
	if(_STRICMP(color, HAMSTER_S_LED_VIOLET) == 0) return _HAMSTER_S_COLOR_TO_RGB[7];
	if(_STRICMP(color, HAMSTER_S_LED_PURPLE) == 0) return _HAMSTER_S_COLOR_TO_RGB[8];
	if(_STRICMP(color, HAMSTER_S_LED_WHITE) == 0) return _HAMSTER_S_COLOR_TO_RGB[9];
	if(_STRICMP(color, "SKY_BLUE") == 0) return _HAMSTER_S_COLOR_TO_RGB[5];
	if(_STRICMP(color, "SKYBLUE") == 0) return _HAMSTER_S_COLOR_TO_RGB[5];
	return NULL;
}

void _hamster_s_leds(int hamster_s_index, const char* left_color, const char* right_color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int* rgb;
	
	if(robot == NULL) return;
	rgb = _hamster_s_get_rgb(left_color);
	if(rgb != NULL) _robot_write_array(robot, HAMSTER_S_LEFT_RGB, rgb, 3);
	rgb = _hamster_s_get_rgb(right_color);
	if(rgb != NULL) _robot_write_array(robot, HAMSTER_S_RIGHT_RGB, rgb, 3);
}

void _hamster_s_left_led(int hamster_s_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int* rgb;
	
	if(robot == NULL) return;
	rgb = _hamster_s_get_rgb(color);
	if(rgb != NULL) _robot_write_array(robot, HAMSTER_S_LEFT_RGB, rgb, 3);
}

void _hamster_s_right_led(int hamster_s_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int* rgb;
	
	if(robot == NULL) return;
	rgb = _hamster_s_get_rgb(color);
	if(rgb != NULL) _robot_write_array(robot, HAMSTER_S_RIGHT_RGB, rgb, 3);
}

void _hamster_s_buzzer(int hamster_s_index, double hz) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_NOTE, 0);
	_hamster_s_run_sound(robot, 0, 1);
	_robot_write_float(robot, HAMSTER_S_BUZZER, (float)hz);
}

void _hamster_s_tempo(int hamster_s_index, double bpm) {
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)_robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(hamster == NULL) return;
	if(bpm > 0) hamster->bpm = bpm;
}

void _hamster_s_pitch(int hamster_s_index, const char* pitch) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int _pitch;

	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_BUZZER, 0);
	_hamster_s_run_sound(robot, 0, 1);
	_pitch = _get_pitch_value(pitch);
	if(_pitch >= 0) _robot_write(robot, HAMSTER_S_NOTE, _pitch);
}

void _hamster_s_note(int hamster_s_index, const char* pitch, double beats) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL) return;
	_robot_write(robot, HAMSTER_S_BUZZER, 0);
	_hamster_s_run_sound(robot, 0, 1);
	if(beats > 0 && hamster->bpm > 0) {
		int _pitch = _get_pitch_value(pitch);
		if(_pitch == 0) {
			_robot_write(robot, HAMSTER_S_NOTE, 0);
			wait((int)(beats * 60 * 1000 / hamster->bpm));
		} else if(_pitch > 0) {
			int timeout = (int)(beats * 60 * 1000 / hamster->bpm);
			int tail = 0;
			if(timeout > 100) {
				tail = 100;
			}
			_robot_write(robot, HAMSTER_S_NOTE, _pitch);
			wait(timeout - tail);
			_robot_write(robot, HAMSTER_S_NOTE, 0);
			if(tail > 0) {
				wait(tail);
			}
		}
	} else {
		_robot_write(robot, HAMSTER_S_NOTE, 0);
	}
}

int _hamster_s_get_sound_value(const char* sound) {
	if(sound == NULL) return -1;
	if(_STRICMP(sound, HAMSTER_S_SOUND_OFF) == 0) return _HAMSTER_S_SOUND_OFF;
	if(_STRICMP(sound, HAMSTER_S_SOUND_BEEP) == 0) return _HAMSTER_S_SOUND_BEEP;
	if(_STRICMP(sound, HAMSTER_S_SOUND_RANDOM_BEEP) == 0) return _HAMSTER_S_SOUND_RANDOM_BEEP;
	if(_STRICMP(sound, HAMSTER_S_SOUND_NOISE) == 0) return _HAMSTER_S_SOUND_NOISE;
	if(_STRICMP(sound, HAMSTER_S_SOUND_SIREN) == 0) return _HAMSTER_S_SOUND_SIREN;
	if(_STRICMP(sound, HAMSTER_S_SOUND_ENGINE) == 0) return _HAMSTER_S_SOUND_ENGINE;
	if(_STRICMP(sound, HAMSTER_S_SOUND_CHOP) == 0) return _HAMSTER_S_SOUND_CHOP;
	if(_STRICMP(sound, HAMSTER_S_SOUND_ROBOT) == 0) return _HAMSTER_S_SOUND_ROBOT;
	if(_STRICMP(sound, HAMSTER_S_SOUND_DIBIDIBIDIP) == 0) return _HAMSTER_S_SOUND_DIBIDIBIDIP;
	if(_STRICMP(sound, HAMSTER_S_SOUND_GOOD_JOB) == 0) return _HAMSTER_S_SOUND_GOOD_JOB;
	if(_STRICMP(sound, HAMSTER_S_SOUND_HAPPY) == 0) return _HAMSTER_S_SOUND_HAPPY;
	if(_STRICMP(sound, HAMSTER_S_SOUND_ANGRY) == 0) return _HAMSTER_S_SOUND_ANGRY;
	if(_STRICMP(sound, HAMSTER_S_SOUND_SAD) == 0) return _HAMSTER_S_SOUND_SAD;
	if(_STRICMP(sound, HAMSTER_S_SOUND_SLEEP) == 0) return _HAMSTER_S_SOUND_SLEEP;
	if(_STRICMP(sound, HAMSTER_S_SOUND_MARCH) == 0) return _HAMSTER_S_SOUND_MARCH;
	if(_STRICMP(sound, HAMSTER_S_SOUND_BIRTHDAY) == 0) return _HAMSTER_S_SOUND_BIRTHDAY;
	if(_STRICMP(sound, "RANDOM_BEEP") == 0) return _HAMSTER_S_SOUND_RANDOM_BEEP;
	if(_STRICMP(sound, "GOOD_JOB") == 0) return _HAMSTER_S_SOUND_GOOD_JOB;
	return -1;
}

void _hamster_s_repeat_sound(int hamster_s_index, const char* sound, double repeat) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int _sound, _repeat;

	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_BUZZER, 0);
	_robot_write(robot, HAMSTER_S_NOTE, 0);
	_sound = _hamster_s_get_sound_value(sound);
	_repeat = _round_double_value(repeat);
	if(_sound > 0 && _repeat != 0) {
		_hamster_s_run_sound(robot, _sound, _repeat);
	} else {
		_hamster_s_run_sound(robot, 0, 1);
	}
}

void _hamster_s_sound(int hamster_s_index, const char* sound) {
	_hamster_s_repeat_sound(hamster_s_index, sound, 1);
}

int _hamster_s_sound_callback(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, HAMSTER_S_SOUND_STATE) == 1 && _robot_read(robot, HAMSTER_S_SOUND_STATE) == 0) ? 1 : 0;
}

void _hamster_s_repeat_sound_until_done(int hamster_s_index, const char* sound, double repeat, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int _sound, _repeat;

	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_BUZZER, 0);
	_robot_write(robot, HAMSTER_S_NOTE, 0);
	_sound = _hamster_s_get_sound_value(sound);
	_repeat = _round_double_value(repeat);
	if(_sound > 0 && _repeat != 0) {
		_hamster_s_run_sound(robot, _sound, _repeat);
		wait_until(callback, NULL);
	} else {
		_hamster_s_run_sound(robot, 0, 1);
	}
}

void _hamster_s_sound_until_done(int hamster_s_index, const char* sound, int (*callback)(void* arg)) {
	_hamster_s_repeat_sound_until_done(hamster_s_index, sound, 1, callback);
}

int _hamster_s_get_io_mode_value(const char* mode) {
	if(mode == NULL) return -1;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_ANALOG_INPUT) == 0) return _HAMSTER_S_IO_MODE_ANALOG_INPUT;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_DIGITAL_INPUT) == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_UP) == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_UP;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_DOWN) == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_DOWN;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_VOLTAGE_INPUT) == 0) return _HAMSTER_S_IO_MODE_VOLTAGE_INPUT;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_SERVO_OUTPUT) == 0) return _HAMSTER_S_IO_MODE_SERVO_OUTPUT;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_PWM_OUTPUT) == 0) return _HAMSTER_S_IO_MODE_PWM_OUTPUT;
	if(_STRICMP(mode, HAMSTER_S_IO_MODE_DIGITAL_OUTPUT) == 0) return _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT;
	if(_STRICMP(mode, "ANALOG_INPUT") == 0) return _HAMSTER_S_IO_MODE_ANALOG_INPUT;
	if(_STRICMP(mode, "DIGITAL_INPUT") == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT;
	if(_STRICMP(mode, "DIGITAL_INPUT_PULL_UP") == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_UP;
	if(_STRICMP(mode, "DIGITAL_INPUT_PULL_DOWN") == 0) return _HAMSTER_S_IO_MODE_DIGITAL_INPUT_PULL_DOWN;
	if(_STRICMP(mode, "VOLTAGE_INPUT") == 0) return _HAMSTER_S_IO_MODE_VOLTAGE_INPUT;
	if(_STRICMP(mode, "SERVO_OUTPUT") == 0) return _HAMSTER_S_IO_MODE_SERVO_OUTPUT;
	if(_STRICMP(mode, "PWM_OUTPUT") == 0) return _HAMSTER_S_IO_MODE_PWM_OUTPUT;
	if(_STRICMP(mode, "DIGITAL_OUTPUT") == 0) return _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT;
	return -1;
}

void _hamster_s_io_mode_a(int hamster_s_index, const char* mode) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_s_get_io_mode_value(mode);
	if(value >= 0) _robot_write(robot, HAMSTER_S_IO_MODE_A, value);
}

void _hamster_s_io_mode_b(int hamster_s_index, const char* mode) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	int value;
	
	if(robot == NULL) return;
	value = _hamster_s_get_io_mode_value(mode);
	if(value >= 0) _robot_write(robot, HAMSTER_S_IO_MODE_B, value);
}

void _hamster_s_output_a(int hamster_s_index, double value) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_OUTPUT_A, _round_double_value(value));
}

void _hamster_s_output_b(int hamster_s_index, double value) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_OUTPUT_B, _round_double_value(value));
}

void _hamster_s_open_gripper(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_OUTPUT_A, 1);
	_robot_write(robot, HAMSTER_S_OUTPUT_B, 0);
	wait(500);
}

void _hamster_s_close_gripper(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_OUTPUT_A, 0);
	_robot_write(robot, HAMSTER_S_OUTPUT_B, 1);
	wait(500);
}

void _hamster_s_release_gripper(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, _HAMSTER_S_IO_MODE_DIGITAL_OUTPUT);
	_robot_write(robot, HAMSTER_S_OUTPUT_A, 0);
	_robot_write(robot, HAMSTER_S_OUTPUT_B, 0);
}

int _hamster_s_write_serial_callback(int hamster_s_index, void* arg) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int id = *(int*)arg;
	
	if(hamster == NULL) return 0;
	return hamster->write_serial_id != id ? 1 : 0;
}

void _hamster_s_write_serial(int hamster_s_index, const char* text, int len, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int* temp;
	
	if(hamster == NULL || text == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, hamster->serial_baud);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, hamster->serial_baud);
	_write_queue_push(&hamster->write_queue, text, len, 0);
	temp = _write_queue_pop(&hamster->write_queue);
	if(temp != NULL) {
		int id = hamster->write_serial_id;
		_robot_write_array(robot, HAMSTER_S_WRITE_SERIAL, temp, 19);
		wait_until(callback, &id);
	}
}

void _hamster_s_write_serial_line(int hamster_s_index, const char* text, int len, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int* temp;
	
	if(hamster == NULL || text == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, hamster->serial_baud);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, hamster->serial_baud);
	_write_queue_push(&hamster->write_queue, text, len, 1);
	temp = _write_queue_pop(&hamster->write_queue);
	if(temp != NULL) {
		int id = hamster->write_serial_id;
		_robot_write_array(robot, HAMSTER_S_WRITE_SERIAL, temp, 19);
		wait_until(callback, &id);
	}
}

void _hamster_s_read_serial(int hamster_s_index, char* text, int len) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL || text == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, hamster->serial_baud);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, hamster->serial_baud);
	
	while(_read_queue_pop(&hamster->read_queue, text, len, 0) <= 0) {
		Sleep(10);
	}
}

void _hamster_s_read_serial_ultil(int hamster_s_index, char* text, int len, char delimiter) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	
	if(hamster == NULL || text == NULL) return;
	_robot_write(robot, HAMSTER_S_IO_MODE_A, hamster->serial_baud);
	_robot_write(robot, HAMSTER_S_IO_MODE_B, hamster->serial_baud);
	
	while(_read_queue_pop(&hamster->read_queue, text, len, delimiter) <= 0) {
		Sleep(10);
	}
}

int _get_serial_rate(int baud) {
	switch(baud) {
		case 9600: return _HAMSTER_S_IO_MODE_SERIAL_9600;
		case 14400: return _HAMSTER_S_IO_MODE_SERIAL_14400;
		case 19200: return _HAMSTER_S_IO_MODE_SERIAL_19200;
		case 28800: return _HAMSTER_S_IO_MODE_SERIAL_28800;
		case 38400: return _HAMSTER_S_IO_MODE_SERIAL_38400;
		case 57600: return _HAMSTER_S_IO_MODE_SERIAL_57600;
		case 76800: return _HAMSTER_S_IO_MODE_SERIAL_76800;
		case 115200: return _HAMSTER_S_IO_MODE_SERIAL_115200;
	}
	return -1;
}

void _hamster_s_serial_rate(int hamster_s_index, double baud) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	struct _hamster_s_robot* hamster = (struct _hamster_s_robot*)robot;
	int _baud;
	
	if(hamster == NULL) return;
	_baud = _get_serial_rate(baud);
	if(_baud >= 0) {
		hamster->serial_baud = _baud;
		_robot_write(robot, HAMSTER_S_IO_MODE_A, _baud);
		_robot_write(robot, HAMSTER_S_IO_MODE_B, _baud);
	}
}

int _hamster_s_signal_strength(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_SIGNAL_STRENGTH);
}

int _hamster_s_left_proximity(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_LEFT_PROXIMITY);
}

int _hamster_s_right_proximity(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_RIGHT_PROXIMITY);
}

int _hamster_s_left_floor(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_LEFT_FLOOR);
}

int _hamster_s_right_floor(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_RIGHT_FLOOR);
}

int _hamster_s_acceleration_x(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_S_ACCELERATION, 0);
}

int _hamster_s_acceleration_y(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_S_ACCELERATION, 1);
}

int _hamster_s_acceleration_z(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, HAMSTER_S_ACCELERATION, 2);
}

int _hamster_s_light(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_LIGHT);
}

int _hamster_s_temperature(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_TEMPERATURE);
}

int _hamster_s_input_a(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_INPUT_A);
}

int _hamster_s_input_b(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_INPUT_B);
}

float _hamster_s_voltage_a(int hamster_s_index) {
	return _hamster_s_input_a(hamster_s_index) * 3.6f / 255;
}

float _hamster_s_voltage_b(int hamster_s_index) {
	return _hamster_s_input_b(hamster_s_index) * 3.6f / 255;
}

int _hamster_s_free_fall(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_e(robot, HAMSTER_S_FREE_FALL);
}

int _hamster_s_tap(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_e(robot, HAMSTER_S_TAP);
}

int _hamster_s_pulse_count(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_PULSE_COUNT);
}

int _hamster_s_tilt(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_TILT);
}

int _hamster_s_battery_state(int hamster_s_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_HAMSTER_S, hamster_s_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, HAMSTER_S_BATTERY_STATE);
}

#define _FUNCTION_HAMSTER_S(n) \
	__inline const char* _hamster_s_get_name_##n(void) { return _robot_group_get_name(_GROUP_HAMSTER_S, n); } \
	__inline void _hamster_s_set_name_##n(const char* name) { _robot_group_set_name(_GROUP_HAMSTER_S, n, name); } \
	__inline const char* _hamster_s_get_id_##n(void) { return HAMSTER_S_ID; } \
	__inline int _hamster_s_get_index_##n(void) { return n; } \
	__inline int _hamster_s_e_##n(int device_id) { return _robot_group_e(_GROUP_HAMSTER_S, n, device_id); } \
	__inline int _hamster_s_read_##n(int device_id) { return _robot_group_read(_GROUP_HAMSTER_S, n, device_id); } \
	__inline int _hamster_s_read_at_##n(int device_id, int index) { return _robot_group_read_at(_GROUP_HAMSTER_S, n, device_id, index); } \
	__inline int _hamster_s_read_array_##n(int device_id, int* data, int length) { return _robot_group_read_array(_GROUP_HAMSTER_S, n, device_id, data, length); } \
	__inline float _hamster_s_read_float_##n(int device_id) { return _robot_group_read_float(_GROUP_HAMSTER_S, n, device_id); } \
	__inline float _hamster_s_read_float_at_##n(int device_id, int index) { return _robot_group_read_float_at(_GROUP_HAMSTER_S, n, device_id, index); } \
	__inline int _hamster_s_read_float_array_##n(int device_id, float* data, int length) { return _robot_group_read_float_array(_GROUP_HAMSTER_S, n, device_id, data, length); } \
	__inline int _hamster_s_write_empty_##n(int device_id) { return _robot_group_write_empty(_GROUP_HAMSTER_S, n, device_id); } \
	__inline int _hamster_s_write_##n(int device_id, int data) { return _robot_group_write(_GROUP_HAMSTER_S, n, device_id, data); } \
	__inline int _hamster_s_write_at_##n(int device_id, int index, int data) { return _robot_group_write_at(_GROUP_HAMSTER_S, n, device_id, index, data); } \
	__inline int _hamster_s_write_array_##n(int device_id, const int* data, int length) { return _robot_group_write_array(_GROUP_HAMSTER_S, n, device_id, data, length); } \
	__inline int _hamster_s_write_float_##n(int device_id, float data) { return _robot_group_write_float(_GROUP_HAMSTER_S, n, device_id, data); } \
	__inline int _hamster_s_write_float_at_##n(int device_id, int index, float data) { return _robot_group_write_float_at(_GROUP_HAMSTER_S, n, device_id, index, data); } \
	__inline int _hamster_s_write_float_array_##n(int device_id, const float* data, int length) { return _robot_group_write_float_array(_GROUP_HAMSTER_S, n, device_id, data, length); } \
	__inline void _hamster_s_reset_##n(void) { _robot_group_reset(_GROUP_HAMSTER_S, n); } \
	__inline void _hamster_s_dispose_##n(void) { _robot_group_dispose(_GROUP_HAMSTER_S, n); } \
	__inline void _hamster_s_motor_mode_##n(const char* mode) { _hamster_s_motor_mode(n, mode); } \
	__inline void _hamster_s_wheels_##n(double left_velocity, double right_velocity) { _hamster_s_wheels(n, left_velocity, right_velocity); } \
	__inline void _hamster_s_left_wheel_##n(double velocity) { _hamster_s_left_wheel(n, velocity); } \
	__inline void _hamster_s_right_wheel_##n(double velocity) { _hamster_s_right_wheel(n, velocity); } \
	__inline void _hamster_s_stop_##n(void) { _hamster_s_stop(n); } \
	__inline int _hamster_s_wheel_callback_##n(void* arg) { return _hamster_s_wheel_callback(n); } \
	__inline void _hamster_s_move_forward_##n(double cm, double velocity) { _hamster_s_move_forward(n, cm, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_move_backward_##n(double cm, double velocity) { _hamster_s_move_backward(n, cm, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_turn_left_##n(double degree, double velocity) { _hamster_s_turn_left(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_turn_right_##n(double degree, double velocity) { _hamster_s_turn_right(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_left_wheel_##n(double degree, double velocity) { _hamster_s_pivot_left_wheel(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_right_wheel_##n(double degree, double velocity) { _hamster_s_pivot_right_wheel(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_left_pen_##n(double degree, double velocity) { _hamster_s_pivot_left_pen(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_right_pen_##n(double degree, double velocity) { _hamster_s_pivot_right_pen(n, degree, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_circle_left_##n(double degree, double radius, double velocity) { _hamster_s_circle_left(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_circle_right_##n(double degree, double radius, double velocity) { _hamster_s_circle_right(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_left_pen_circle_left_##n(double degree, double radius, double velocity) { _hamster_s_left_pen_circle_left(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_left_pen_circle_right_##n(double degree, double radius, double velocity) { _hamster_s_left_pen_circle_right(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_right_pen_circle_left_##n(double degree, double radius, double velocity) { _hamster_s_right_pen_circle_left(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_right_pen_circle_right_##n(double degree, double radius, double velocity) { _hamster_s_right_pen_circle_right(n, degree, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_move_forward_sec_##n(double sec, double velocity) { _hamster_s_move_forward_sec(n, sec, velocity); } \
	__inline void _hamster_s_move_backward_sec_##n(double sec, double velocity) { _hamster_s_move_backward_sec(n, sec, velocity); } \
	__inline void _hamster_s_turn_left_sec_##n(double sec, double velocity) { _hamster_s_turn_left_sec(n, sec, velocity); } \
	__inline void _hamster_s_turn_right_sec_##n(double sec, double velocity) { _hamster_s_turn_right_sec(n, sec, velocity); } \
	__inline void _hamster_s_pivot_left_wheel_sec_##n(double sec, double velocity) { _hamster_s_pivot_left_wheel_sec(n, sec, velocity); } \
	__inline void _hamster_s_pivot_right_wheel_sec_##n(double sec, double velocity) { _hamster_s_pivot_right_wheel_sec(n, sec, velocity); } \
	__inline void _hamster_s_pivot_left_pen_sec_##n(double sec, double velocity) { _hamster_s_pivot_left_pen_sec(n, sec, velocity); } \
	__inline void _hamster_s_pivot_right_pen_sec_##n(double sec, double velocity) { _hamster_s_pivot_right_pen_sec(n, sec, velocity); } \
	__inline void _hamster_s_circle_left_sec_##n(double sec, double radius, double velocity) { _hamster_s_circle_left_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_circle_right_sec_##n(double sec, double radius, double velocity) { _hamster_s_circle_right_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_left_pen_circle_left_sec_##n(double sec, double radius, double velocity) { _hamster_s_left_pen_circle_left_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_left_pen_circle_right_sec_##n(double sec, double radius, double velocity) { _hamster_s_left_pen_circle_right_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_right_pen_circle_left_sec_##n(double sec, double radius, double velocity) { _hamster_s_right_pen_circle_left_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_right_pen_circle_right_sec_##n(double sec, double radius, double velocity) { _hamster_s_right_pen_circle_right_sec(n, sec, radius, velocity); } \
	__inline void _hamster_s_move_forward_pulse_##n(double pulse, double velocity) { _hamster_s_move_forward_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_move_backward_pulse_##n(double pulse, double velocity) { _hamster_s_move_backward_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_turn_left_pulse_##n(double pulse, double velocity) { _hamster_s_turn_left_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_turn_right_pulse_##n(double pulse, double velocity) { _hamster_s_turn_right_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_left_wheel_pulse_##n(double pulse, double velocity) { _hamster_s_pivot_left_wheel_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_right_wheel_pulse_##n(double pulse, double velocity) { _hamster_s_pivot_right_wheel_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_left_pen_pulse_##n(double pulse, double velocity) { _hamster_s_pivot_left_pen_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_pivot_right_pen_pulse_##n(double pulse, double velocity) { _hamster_s_pivot_right_pen_pulse(n, pulse, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_circle_left_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_circle_left_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_circle_right_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_circle_right_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_left_pen_circle_left_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_left_pen_circle_left_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_left_pen_circle_right_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_left_pen_circle_right_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_right_pen_circle_left_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_right_pen_circle_left_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline void _hamster_s_right_pen_circle_right_pulse_##n(double pulse, double radius, double velocity) { _hamster_s_right_pen_circle_right_pulse(n, pulse, radius, velocity, _hamster_s_wheel_callback_##n); } \
	__inline int _hamster_s_line_tracer_mode_callback_##n(void* arg) { return _hamster_s_line_tracer_mode_callback(n); } \
	__inline void _hamster_s_line_tracer_mode_##n(int mode) { _hamster_s_line_tracer_mode(n, mode, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_line_left_##n(void) { _hamster_s_line_left(n); } \
	__inline void _hamster_s_line_right_##n(void) { _hamster_s_line_right(n); } \
	__inline void _hamster_s_line_both_##n(void) { _hamster_s_line_both(n); } \
	__inline void _hamster_s_cross_forward_##n(void) { _hamster_s_cross_forward(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_cross_left_##n(void) { _hamster_s_cross_left(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_cross_right_##n(void) { _hamster_s_cross_right(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_cross_uturn_##n(void) { _hamster_s_cross_uturn(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_white_line_left_##n(void) { _hamster_s_white_line_left(n); } \
	__inline void _hamster_s_white_line_right_##n(void) { _hamster_s_white_line_right(n); } \
	__inline void _hamster_s_white_line_both_##n(void) { _hamster_s_white_line_both(n); } \
	__inline void _hamster_s_white_cross_forward_##n(void) { _hamster_s_white_cross_forward(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_white_cross_left_##n(void) { _hamster_s_white_cross_left(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_white_cross_right_##n(void) { _hamster_s_white_cross_right(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_white_cross_uturn_##n(void) { _hamster_s_white_cross_uturn(n, _hamster_s_line_tracer_mode_callback_##n); } \
	__inline void _hamster_s_line_gain_auto_##n(void) { _hamster_s_line_gain_auto(n); } \
	__inline void _hamster_s_line_gain_##n(double gain) { _hamster_s_line_gain(n, gain); } \
	__inline void _hamster_s_line_speed_##n(double speed) { _hamster_s_line_speed(n, speed); } \
	__inline int _hamster_s_board_forward_callback_##n(void* arg) { return _hamster_s_board_forward_callback(n); } \
	__inline void _hamster_s_board_forward_##n(void) { _hamster_s_board_forward(n, _hamster_s_board_forward_callback_##n); } \
	__inline int _hamster_s_board_left_callback_##n(void* arg) { return _hamster_s_board_left_callback(n); } \
	__inline void _hamster_s_board_left_##n(void) { _hamster_s_board_left(n, _hamster_s_board_left_callback_##n); } \
	__inline int _hamster_s_board_right_callback_##n(void* arg) { return _hamster_s_board_right_callback(n); } \
	__inline void _hamster_s_board_right_##n(void) { _hamster_s_board_right(n, _hamster_s_board_right_callback_##n); } \
	__inline void _hamster_s_rgbs_##n(double left_red, double left_green, double left_blue, double right_red, double right_green, double right_blue) { _hamster_s_rgbs(n, left_red, left_green, left_blue, right_red, right_green, right_blue); } \
	__inline void _hamster_s_left_rgb_##n(double red, double green, double blue) { _hamster_s_left_rgb(n, red, green, blue); } \
	__inline void _hamster_s_right_rgb_##n(double red, double green, double blue) { _hamster_s_right_rgb(n, red, green, blue); } \
	__inline void _hamster_s_leds_##n(const char* left_color, const char* right_color) { _hamster_s_leds(n, left_color, right_color); } \
	__inline void _hamster_s_left_led_##n(const char* color) { _hamster_s_left_led(n, color); } \
	__inline void _hamster_s_right_led_##n(const char* color) { _hamster_s_right_led(n, color); } \
	__inline void _hamster_s_buzzer_##n(double hz) { _hamster_s_buzzer(n, hz); } \
	__inline void _hamster_s_tempo_##n(double bpm) { _hamster_s_tempo(n, bpm); } \
	__inline void _hamster_s_pitch_##n(const char* pitch) { _hamster_s_pitch(n, pitch); } \
	__inline void _hamster_s_note_##n(const char* pitch, double beats) { _hamster_s_note(n, pitch, beats); } \
	__inline void _hamster_s_repeat_sound_##n(const char* sound, double repeat) { _hamster_s_repeat_sound(n, sound, repeat); } \
	__inline void _hamster_s_sound_##n(const char* sound) { _hamster_s_sound(n, sound); } \
	__inline int _hamster_s_sound_callback_##n(void* arg) { return _hamster_s_sound_callback(n); } \
	__inline void _hamster_s_repeat_sound_until_done_##n(const char* sound, double repeat) { _hamster_s_repeat_sound_until_done(n, sound, repeat, _hamster_s_sound_callback_##n); } \
	__inline void _hamster_s_sound_until_done_##n(const char* sound) { _hamster_s_sound_until_done(n, sound, _hamster_s_sound_callback_##n); } \
	__inline void _hamster_s_io_mode_a_##n(const char* mode) { _hamster_s_io_mode_a(n, mode); } \
	__inline void _hamster_s_io_mode_b_##n(const char* mode) { _hamster_s_io_mode_b(n, mode); } \
	__inline void _hamster_s_output_a_##n(double value) { _hamster_s_output_a(n, value); } \
	__inline void _hamster_s_output_b_##n(double value) { _hamster_s_output_b(n, value); } \
	__inline void _hamster_s_open_gripper_##n(void) { _hamster_s_open_gripper(n); } \
	__inline void _hamster_s_close_gripper_##n(void) { _hamster_s_close_gripper(n); } \
	__inline void _hamster_s_release_gripper_##n(void) { _hamster_s_release_gripper(n); } \
	__inline int _hamster_s_write_serial_callback_##n(void* arg) { return _hamster_s_write_serial_callback(n, arg); } \
	__inline void _hamster_s_write_serial_##n(const char* text, int len) { _hamster_s_write_serial(n, text, len, _hamster_s_write_serial_callback_##n); } \
	__inline void _hamster_s_write_serial_line_##n(const char* text, int len) { _hamster_s_write_serial_line(n, text, len, _hamster_s_write_serial_callback_##n); } \
	__inline void _hamster_s_read_serial_##n(char* text, int len) { _hamster_s_read_serial(n, text, len); } \
	__inline void _hamster_s_read_serial_ultil_##n(char* text, int len, char delimiter) { _hamster_s_read_serial_ultil(n, text, len, delimiter); } \
	__inline void _hamster_s_serial_rate_##n(double baud) { _hamster_s_serial_rate(n, baud); } \
	__inline int _hamster_s_signal_strength_##n(void) { return _hamster_s_signal_strength(n); } \
	__inline int _hamster_s_left_proximity_##n(void) { return _hamster_s_left_proximity(n); } \
	__inline int _hamster_s_right_proximity_##n(void) { return _hamster_s_right_proximity(n); } \
	__inline int _hamster_s_left_floor_##n(void) { return _hamster_s_left_floor(n); } \
	__inline int _hamster_s_right_floor_##n(void) { return _hamster_s_right_floor(n); } \
	__inline int _hamster_s_acceleration_x_##n(void) { return _hamster_s_acceleration_x(n); } \
	__inline int _hamster_s_acceleration_y_##n(void) { return _hamster_s_acceleration_y(n); } \
	__inline int _hamster_s_acceleration_z_##n(void) { return _hamster_s_acceleration_z(n); } \
	__inline int _hamster_s_light_##n(void) { return _hamster_s_light(n); } \
	__inline int _hamster_s_temperature_##n(void) { return _hamster_s_temperature(n); } \
	__inline int _hamster_s_input_a_##n(void) { return _hamster_s_input_a(n); } \
	__inline int _hamster_s_input_b_##n(void) { return _hamster_s_input_b(n); } \
	__inline float _hamster_s_voltage_a_##n(void) { return _hamster_s_voltage_a(n); } \
	__inline float _hamster_s_voltage_b_##n(void) { return _hamster_s_voltage_b(n); } \
	__inline int _hamster_s_free_fall_##n(void) { return _hamster_s_free_fall(n); } \
	__inline int _hamster_s_tap_##n(void) { return _hamster_s_tap(n); } \
	__inline int _hamster_s_pulse_count_##n(void) { return _hamster_s_pulse_count(n); } \
	__inline int _hamster_s_tilt_##n(void) { return _hamster_s_tilt(n); } \
	__inline int _hamster_s_battery_state_##n(void) { return _hamster_s_battery_state(n); }

#define _FUNCTION_PTR_HAMSTER_S(name, n) \
	name->get_name = _hamster_s_get_name_##n; \
	name->set_name = _hamster_s_set_name_##n; \
	name->get_id = _hamster_s_get_id_##n; \
	name->get_index = _hamster_s_get_index_##n; \
	name->e = _hamster_s_e_##n; \
	name->read = _hamster_s_read_##n; \
	name->read_at = _hamster_s_read_at_##n; \
	name->read_array = _hamster_s_read_array_##n; \
	name->read_float = _hamster_s_read_float_##n; \
	name->read_float_at = _hamster_s_read_float_at_##n; \
	name->read_float_array = _hamster_s_read_float_array_##n; \
	name->write_empty = _hamster_s_write_empty_##n; \
	name->write = _hamster_s_write_##n; \
	name->write_at = _hamster_s_write_at_##n; \
	name->write_array = _hamster_s_write_array_##n; \
	name->write_float = _hamster_s_write_float_##n; \
	name->write_float_at = _hamster_s_write_float_at_##n; \
	name->write_float_array = _hamster_s_write_float_array_##n; \
	name->reset = _hamster_s_reset_##n; \
	name->dispose = _hamster_s_dispose_##n; \
	name->motor_mode = _hamster_s_motor_mode_##n; \
	name->wheels = _hamster_s_wheels_##n; \
	name->left_wheel = _hamster_s_left_wheel_##n; \
	name->right_wheel = _hamster_s_right_wheel_##n; \
	name->stop = _hamster_s_stop_##n; \
	name->move_forward = _hamster_s_move_forward_##n; \
	name->move_backward = _hamster_s_move_backward_##n; \
	name->turn_left = _hamster_s_turn_left_##n; \
	name->turn_right = _hamster_s_turn_right_##n; \
	name->pivot_left_wheel = _hamster_s_pivot_left_wheel_##n; \
	name->pivot_right_wheel = _hamster_s_pivot_right_wheel_##n; \
	name->pivot_left_pen = _hamster_s_pivot_left_pen_##n; \
	name->pivot_right_pen = _hamster_s_pivot_right_pen_##n; \
	name->circle_left = _hamster_s_circle_left_##n; \
	name->circle_right = _hamster_s_circle_right_##n; \
	name->left_pen_circle_left = _hamster_s_left_pen_circle_left_##n; \
	name->left_pen_circle_right = _hamster_s_left_pen_circle_right_##n; \
	name->right_pen_circle_left = _hamster_s_right_pen_circle_left_##n; \
	name->right_pen_circle_right = _hamster_s_right_pen_circle_right_##n; \
	name->move_forward_sec = _hamster_s_move_forward_sec_##n; \
	name->move_backward_sec = _hamster_s_move_backward_sec_##n; \
	name->turn_left_sec = _hamster_s_turn_left_sec_##n; \
	name->turn_right_sec = _hamster_s_turn_right_sec_##n; \
	name->pivot_left_wheel_sec = _hamster_s_pivot_left_wheel_sec_##n; \
	name->pivot_right_wheel_sec = _hamster_s_pivot_right_wheel_sec_##n; \
	name->pivot_left_pen_sec = _hamster_s_pivot_left_pen_sec_##n; \
	name->pivot_right_pen_sec = _hamster_s_pivot_right_pen_sec_##n; \
	name->circle_left_sec = _hamster_s_circle_left_sec_##n; \
	name->circle_right_sec = _hamster_s_circle_right_sec_##n; \
	name->left_pen_circle_left_sec = _hamster_s_left_pen_circle_left_sec_##n; \
	name->left_pen_circle_right_sec = _hamster_s_left_pen_circle_right_sec_##n; \
	name->right_pen_circle_left_sec = _hamster_s_right_pen_circle_left_sec_##n; \
	name->right_pen_circle_right_sec = _hamster_s_right_pen_circle_right_sec_##n; \
	name->move_forward_pulse = _hamster_s_move_forward_pulse_##n; \
	name->move_backward_pulse = _hamster_s_move_backward_pulse_##n; \
	name->turn_left_pulse = _hamster_s_turn_left_pulse_##n; \
	name->turn_right_pulse = _hamster_s_turn_right_pulse_##n; \
	name->pivot_left_wheel_pulse = _hamster_s_pivot_left_wheel_pulse_##n; \
	name->pivot_right_wheel_pulse = _hamster_s_pivot_right_wheel_pulse_##n; \
	name->pivot_left_pen_pulse = _hamster_s_pivot_left_pen_pulse_##n; \
	name->pivot_right_pen_pulse = _hamster_s_pivot_right_pen_pulse_##n; \
	name->circle_left_pulse = _hamster_s_circle_left_pulse_##n; \
	name->circle_right_pulse = _hamster_s_circle_right_pulse_##n; \
	name->left_pen_circle_left_pulse = _hamster_s_left_pen_circle_left_pulse_##n; \
	name->left_pen_circle_right_pulse = _hamster_s_left_pen_circle_right_pulse_##n; \
	name->right_pen_circle_left_pulse = _hamster_s_right_pen_circle_left_pulse_##n; \
	name->right_pen_circle_right_pulse = _hamster_s_right_pen_circle_right_pulse_##n; \
	name->line_tracer_mode = _hamster_s_line_tracer_mode_##n; \
	name->line_left = _hamster_s_line_left_##n; \
	name->line_right = _hamster_s_line_right_##n; \
	name->line_both = _hamster_s_line_both_##n; \
	name->cross_forward = _hamster_s_cross_forward_##n; \
	name->cross_left = _hamster_s_cross_left_##n; \
	name->cross_right = _hamster_s_cross_right_##n; \
	name->cross_uturn = _hamster_s_cross_uturn_##n; \
	name->white_line_left = _hamster_s_white_line_left_##n; \
	name->white_line_right = _hamster_s_white_line_right_##n; \
	name->white_line_both = _hamster_s_white_line_both_##n; \
	name->white_cross_forward = _hamster_s_white_cross_forward_##n; \
	name->white_cross_left = _hamster_s_white_cross_left_##n; \
	name->white_cross_right = _hamster_s_white_cross_right_##n; \
	name->white_cross_uturn = _hamster_s_white_cross_uturn_##n; \
	name->line_gain_auto = _hamster_s_line_gain_auto_##n; \
	name->line_gain = _hamster_s_line_gain_##n; \
	name->line_speed = _hamster_s_line_speed_##n; \
	name->board_forward = _hamster_s_board_forward_##n; \
	name->board_left = _hamster_s_board_left_##n; \
	name->board_right = _hamster_s_board_right_##n; \
	name->rgbs = _hamster_s_rgbs_##n; \
	name->left_rgb = _hamster_s_left_rgb_##n; \
	name->right_rgb = _hamster_s_right_rgb_##n; \
	name->leds = _hamster_s_leds_##n; \
	name->left_led = _hamster_s_left_led_##n; \
	name->right_led = _hamster_s_right_led_##n; \
	name->buzzer = _hamster_s_buzzer_##n; \
	name->tempo = _hamster_s_tempo_##n; \
	name->pitch = _hamster_s_pitch_##n; \
	name->note = _hamster_s_note_##n; \
	name->sound = _hamster_s_sound_##n; \
	name->sound_until_done = _hamster_s_sound_until_done_##n; \
	name->repeat_sound = _hamster_s_repeat_sound_##n; \
	name->repeat_sound_until_done = _hamster_s_repeat_sound_until_done_##n; \
	name->io_mode_a = _hamster_s_io_mode_a_##n; \
	name->io_mode_b = _hamster_s_io_mode_b_##n; \
	name->output_a = _hamster_s_output_a_##n; \
	name->output_b = _hamster_s_output_b_##n; \
	name->open_gripper = _hamster_s_open_gripper_##n; \
	name->close_gripper = _hamster_s_close_gripper_##n; \
	name->release_gripper = _hamster_s_release_gripper_##n; \
	name->write_serial = _hamster_s_write_serial_##n; \
	name->write_serial_line = _hamster_s_write_serial_line_##n; \
	name->read_serial = _hamster_s_read_serial_##n; \
	name->read_serial_ultil = _hamster_s_read_serial_ultil_##n; \
	name->serial_rate = _hamster_s_serial_rate_##n; \
	name->signal_strength = _hamster_s_signal_strength_##n; \
	name->left_proximity = _hamster_s_left_proximity_##n; \
	name->right_proximity = _hamster_s_right_proximity_##n; \
	name->left_floor = _hamster_s_left_floor_##n; \
	name->right_floor = _hamster_s_right_floor_##n; \
	name->acceleration_x = _hamster_s_acceleration_x_##n; \
	name->acceleration_y = _hamster_s_acceleration_y_##n; \
	name->acceleration_z = _hamster_s_acceleration_z_##n; \
	name->light = _hamster_s_light_##n; \
	name->temperature = _hamster_s_temperature_##n; \
	name->input_a = _hamster_s_input_a_##n; \
	name->input_b = _hamster_s_input_b_##n; \
	name->voltage_a = _hamster_s_voltage_a_##n; \
	name->voltage_b = _hamster_s_voltage_b_##n; \
	name->free_fall = _hamster_s_free_fall_##n; \
	name->tap = _hamster_s_tap_##n; \
	name->pulse_count = _hamster_s_pulse_count_##n; \
	name->tilt = _hamster_s_tilt_##n; \
	name->battery_state = _hamster_s_battery_state_##n;

_FUNCTION_HAMSTER_S(0);
_FUNCTION_HAMSTER_S(1);
_FUNCTION_HAMSTER_S(2);
_FUNCTION_HAMSTER_S(3);
_FUNCTION_HAMSTER_S(4);
_FUNCTION_HAMSTER_S(5);
_FUNCTION_HAMSTER_S(6);
_FUNCTION_HAMSTER_S(7);
_FUNCTION_HAMSTER_S(8);
_FUNCTION_HAMSTER_S(9);

HamsterS* hamster_s_create_port(const char* port_name) {
	if(_robot_group_count_robots(_GROUP_HAMSTER_S) < _MAX_NUM_ROBOTS) {
		struct hamster_s* hamster = (struct hamster_s*)malloc(sizeof(struct hamster_s));
		struct _robot* robot = (struct _robot*)_hamster_s_create(port_name);
	
		switch(robot->index) {
			case 0: _FUNCTION_PTR_HAMSTER_S(hamster, 0); break;
			case 1: _FUNCTION_PTR_HAMSTER_S(hamster, 1); break;
			case 2: _FUNCTION_PTR_HAMSTER_S(hamster, 2); break;
			case 3: _FUNCTION_PTR_HAMSTER_S(hamster, 3); break;
			case 4: _FUNCTION_PTR_HAMSTER_S(hamster, 4); break;
			case 5: _FUNCTION_PTR_HAMSTER_S(hamster, 5); break;
			case 6: _FUNCTION_PTR_HAMSTER_S(hamster, 6); break;
			case 7: _FUNCTION_PTR_HAMSTER_S(hamster, 7); break;
			case 8: _FUNCTION_PTR_HAMSTER_S(hamster, 8); break;
			case 9: _FUNCTION_PTR_HAMSTER_S(hamster, 9); break;
		}
		return hamster;
	}
	return NULL;
}

HamsterS* hamster_s_create(void) {
	return hamster_s_create_port(NULL);
}

const char* hamster_s_get_name(void) {
	return _robot_group_get_name(_GROUP_HAMSTER_S, 0);
}

void hamster_s_set_name(const char* name) {
	_robot_group_set_name(_GROUP_HAMSTER_S, 0, name);
}

const char* hamster_s_get_id(void) {
	return HAMSTER_S_ID;
}

int hamster_s_e(int device_id) {
	return _robot_group_e(_GROUP_HAMSTER_S, 0, device_id);
}

int hamster_s_read(int device_id) {
	return _robot_group_read(_GROUP_HAMSTER_S, 0, device_id);
}

int hamster_s_read_at(int device_id, int index) {
	return _robot_group_read_at(_GROUP_HAMSTER_S, 0, device_id, index);
}

int hamster_s_read_array(int device_id, int* data, int length) {
	return _robot_group_read_array(_GROUP_HAMSTER_S, 0, device_id, data, length);
}

float hamster_s_read_float(int device_id) {
	return _robot_group_read_float(_GROUP_HAMSTER_S, 0, device_id);
}

float hamster_s_read_float_at(int device_id, int index) {
	return _robot_group_read_float_at(_GROUP_HAMSTER_S, 0, device_id, index);
}

int hamster_s_read_float_array(int device_id, float* data, int length) {
	return _robot_group_read_float_array(_GROUP_HAMSTER_S, 0, device_id, data, length);
}

int hamster_s_write_empty(int device_id) {
	return _robot_group_write_empty(_GROUP_HAMSTER_S, 0, device_id);
}

int hamster_s_write(int device_id, int data) {
	return _robot_group_write(_GROUP_HAMSTER_S, 0, device_id, data);
}

int hamster_s_write_at(int device_id, int index, int data) {
	return _robot_group_write_at(_GROUP_HAMSTER_S, 0, device_id, index, data);
}

int hamster_s_write_array(int device_id, const int* data, int length) {
	return _robot_group_write_array(_GROUP_HAMSTER_S, 0, device_id, data, length);
}

int hamster_s_write_float(int device_id, float data) {
	return _robot_group_write_float(_GROUP_HAMSTER_S, 0, device_id, data);
}

int hamster_s_write_float_at(int device_id, int index, float data) {
	return _robot_group_write_float_at(_GROUP_HAMSTER_S, 0, device_id, index, data);
}

int hamster_s_write_float_array(int device_id, const float* data, int length) {
	return _robot_group_write_float_array(_GROUP_HAMSTER_S, 0, device_id, data, length);
}

void hamster_s_reset(void) {
	_robot_group_reset(_GROUP_HAMSTER_S, 0);
}

void hamster_s_dispose(void) {
	_robot_group_dispose(_GROUP_HAMSTER_S, 0);
}

void hamster_s_motor_mode(const char* mode) {
	_hamster_s_motor_mode(0, mode);
}

void hamster_s_wheels(double left_velocity, double right_velocity) {
	_hamster_s_wheels(0, left_velocity, right_velocity);
}

void hamster_s_left_wheel(double velocity) {
	_hamster_s_left_wheel(0, velocity);
}

void hamster_s_right_wheel(double velocity) {
	_hamster_s_right_wheel(0, velocity);
}

void hamster_s_stop(void) {
	_hamster_s_stop(0);
}

int _hamster_s_wheel_evaluate(void* arg) {
	return _hamster_s_wheel_callback(0);
}

void hamster_s_move_forward(double cm, double velocity) {
	_hamster_s_move_forward(0, cm, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_move_backward(double cm, double velocity) {
	_hamster_s_move_backward(0, cm, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_turn_left(double degree, double velocity) {
	_hamster_s_turn_left(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_turn_right(double degree, double velocity) {
	_hamster_s_turn_right(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_left_wheel(double degree, double velocity) {
	_hamster_s_pivot_left_wheel(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_right_wheel(double degree, double velocity) {
	_hamster_s_pivot_right_wheel(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_left_pen(double degree, double velocity) {
	_hamster_s_pivot_left_pen(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_right_pen(double degree, double velocity) {
	_hamster_s_pivot_right_pen(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_circle_left(double degree, double radius, double velocity) {
	_hamster_s_circle_left(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_circle_right(double degree, double radius, double velocity) {
	_hamster_s_circle_right(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_left_pen_circle_left(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_left(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_left_pen_circle_right(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_right(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_right_pen_circle_left(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_left(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_right_pen_circle_right(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_right(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_move_forward_sec(double cm, double velocity) {
	_hamster_s_move_forward_sec(0, cm, velocity);
}

void hamster_s_move_backward_sec(double cm, double velocity) {
	_hamster_s_move_backward_sec(0, cm, velocity);
}

void hamster_s_turn_left_sec(double degree, double velocity) {
	_hamster_s_turn_left_sec(0, degree, velocity);
}

void hamster_s_turn_right_sec(double degree, double velocity) {
	_hamster_s_turn_right_sec(0, degree, velocity);
}

void hamster_s_pivot_left_wheel_sec(double degree, double velocity) {
	_hamster_s_pivot_left_wheel_sec(0, degree, velocity);
}

void hamster_s_pivot_right_wheel_sec(double degree, double velocity) {
	_hamster_s_pivot_right_wheel_sec(0, degree, velocity);
}

void hamster_s_pivot_left_pen_sec(double degree, double velocity) {
	_hamster_s_pivot_left_pen_sec(0, degree, velocity);
}

void hamster_s_pivot_right_pen_sec(double degree, double velocity) {
	_hamster_s_pivot_right_pen_sec(0, degree, velocity);
}

void hamster_s_circle_left_sec(double degree, double radius, double velocity) {
	_hamster_s_circle_left_sec(0, degree, radius, velocity);
}

void hamster_s_circle_right_sec(double degree, double radius, double velocity) {
	_hamster_s_circle_right_sec(0, degree, radius, velocity);
}

void hamster_s_left_pen_circle_left_sec(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_left_sec(0, degree, radius, velocity);
}

void hamster_s_left_pen_circle_right_sec(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_right_sec(0, degree, radius, velocity);
}

void hamster_s_right_pen_circle_left_sec(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_left_sec(0, degree, radius, velocity);
}

void hamster_s_right_pen_circle_right_sec(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_right_sec(0, degree, radius, velocity);
}

void hamster_s_move_forward_pulse(double cm, double velocity) {
	_hamster_s_move_forward_pulse(0, cm, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_move_backward_pulse(double cm, double velocity) {
	_hamster_s_move_backward_pulse(0, cm, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_turn_left_pulse(double degree, double velocity) {
	_hamster_s_turn_left_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_turn_right_pulse(double degree, double velocity) {
	_hamster_s_turn_right_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_left_wheel_pulse(double degree, double velocity) {
	_hamster_s_pivot_left_wheel_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_right_wheel_pulse(double degree, double velocity) {
	_hamster_s_pivot_right_wheel_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_left_pen_pulse(double degree, double velocity) {
	_hamster_s_pivot_left_pen_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_pivot_right_pen_pulse(double degree, double velocity) {
	_hamster_s_pivot_right_pen_pulse(0, degree, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_circle_left_pulse(double degree, double radius, double velocity) {
	_hamster_s_circle_left_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_circle_right_pulse(double degree, double radius, double velocity) {
	_hamster_s_circle_right_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_left_pen_circle_left_pulse(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_left_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_left_pen_circle_right_pulse(double degree, double radius, double velocity) {
	_hamster_s_left_pen_circle_right_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_right_pen_circle_left_pulse(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_left_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

void hamster_s_right_pen_circle_right_pulse(double degree, double radius, double velocity) {
	_hamster_s_right_pen_circle_right_pulse(0, degree, radius, velocity, _hamster_s_wheel_evaluate);
}

int _hamster_s_line_tracer_mode_evaluate(void* arg) {
	return _hamster_s_line_tracer_mode_callback(0);
}

void hamster_s_line_tracer_mode(int mode) {
	_hamster_s_line_tracer_mode(0, mode, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_line_left(void) {
	_hamster_s_line_left(0);
}

void hamster_s_line_right(void) {
	_hamster_s_line_right(0);
}

void hamster_s_line_both(void) {
	_hamster_s_line_both(0);
}

void hamster_s_cross_forward(void) {
	_hamster_s_cross_forward(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_cross_left(void) {
	_hamster_s_cross_left(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_cross_right(void) {
	_hamster_s_cross_right(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_cross_uturn(void) {
	_hamster_s_cross_uturn(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_white_line_left(void) {
	_hamster_s_white_line_left(0);
}

void hamster_s_white_line_right(void) {
	_hamster_s_white_line_right(0);
}

void hamster_s_white_line_both(void) {
	_hamster_s_white_line_both(0);
}

void hamster_s_white_cross_forward(void) {
	_hamster_s_white_cross_forward(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_white_cross_left(void) {
	_hamster_s_white_cross_left(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_white_cross_right(void) {
	_hamster_s_white_cross_right(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_white_cross_uturn(void) {
	_hamster_s_white_cross_uturn(0, _hamster_s_line_tracer_mode_evaluate);
}

void hamster_s_line_gain_auto(void) {
	_hamster_s_line_gain_auto(0);
}

void hamster_s_line_gain(double gain) {
	_hamster_s_line_gain(0, gain);
}

void hamster_s_line_speed(double speed) {
	_hamster_s_line_speed(0, speed);
}

int _hamster_s_board_forward_evaluate(void* arg) {
	return _hamster_s_board_forward_callback(0);
}

void hamster_s_board_forward(void) {
	_hamster_s_board_forward(0, _hamster_s_board_forward_evaluate);
}

int _hamster_s_board_left_evaluate(void* arg) {
	return _hamster_s_board_left_callback(0);
}

void hamster_s_board_left(void) {
	_hamster_s_board_left(0, _hamster_s_board_left_evaluate);
}

int _hamster_s_board_right_evaluate(void* arg) {
	return _hamster_s_board_right_callback(0);
}

void hamster_s_board_right(void) {
	_hamster_s_board_right(0, _hamster_s_board_right_evaluate);
}

void hamster_s_rgbs(double left_red, double left_green, double left_blue, double right_red, double right_green, double right_blue) {
	_hamster_s_rgbs(0, left_red, left_green, left_blue, right_red, right_green, right_blue);
}

void hamster_s_left_rgb(double red, double green, double blue) {
	_hamster_s_left_rgb(0, red, green, blue);
}

void hamster_s_right_rgb(double red, double green, double blue) {
	_hamster_s_right_rgb(0, red, green, blue);
}

void hamster_s_leds(const char* left_color, const char* right_color) {
	_hamster_s_leds(0, left_color, right_color);
}

void hamster_s_left_led(const char* color) {
	_hamster_s_left_led(0, color);
}

void hamster_s_right_led(const char* color) {
	_hamster_s_right_led(0, color);
}

void hamster_s_buzzer(double hz) {
	_hamster_s_buzzer(0, hz);
}

void hamster_s_tempo(double bpm) {
	_hamster_s_tempo(0, bpm);
}

void hamster_s_pitch(const char* pitch) {
	_hamster_s_pitch(0, pitch);
}

void hamster_s_note(const char* pitch, double beats) {
	_hamster_s_note(0, pitch, beats);
}

void hamster_s_sound(const char* sound) {
	_hamster_s_sound(0, sound);
}

void hamster_s_repeat_sound(const char* sound, double repeat) {
	_hamster_s_repeat_sound(0, sound, repeat);
}

int _hamster_s_sound_evaluate(void* arg) {
	return _hamster_s_sound_callback(0);
}

void hamster_s_sound_until_done(const char* sound) {
	_hamster_s_sound_until_done(0, sound, _hamster_s_sound_evaluate);
}

void hamster_s_repeat_sound_until_done(const char* sound, double repeat) {
	_hamster_s_repeat_sound_until_done(0, sound, repeat, _hamster_s_sound_evaluate);
}

void hamster_s_io_mode_a(const char* mode) {
	_hamster_s_io_mode_a(0, mode);
}

void hamster_s_io_mode_b(const char* mode) {
	_hamster_s_io_mode_b(0, mode);
}

void hamster_s_output_a(double value) {
	_hamster_s_output_a(0, value);
}

void hamster_s_output_b(double value) {
	_hamster_s_output_b(0, value);
}

void hamster_s_open_gripper(void) {
	_hamster_s_open_gripper(0);
}

void hamster_s_close_gripper(void) {
	_hamster_s_close_gripper(0);
}

void hamster_s_release_gripper(void) {
	_hamster_s_release_gripper(0);
}

int _hamster_s_write_serial_evaluate(void* arg) {
	return _hamster_s_write_serial_callback(0, arg);
}

void hamster_s_write_serial(const char* text, int len) {
	_hamster_s_write_serial(0, text, len, _hamster_s_write_serial_evaluate);
}

void hamster_s_write_serial_line(const char* text, int len) {
	_hamster_s_write_serial_line(0, text, len, _hamster_s_write_serial_evaluate);
}

void hamster_s_read_serial(char* text, int len) {
	_hamster_s_read_serial(0, text, len);
}

void hamster_s_read_serial_ultil(char* text, int len, char delimiter) {
	_hamster_s_read_serial_ultil(0, text, len, delimiter);
}

void hamster_s_serial_rate(double baud) {
	_hamster_s_serial_rate(0, baud);
}

int hamster_s_signal_strength(void) {
	return _hamster_s_signal_strength(0);
}

int hamster_s_left_proximity(void) {
	return _hamster_s_left_proximity(0);
}

int hamster_s_right_proximity(void) {
	return _hamster_s_right_proximity(0);
}

int hamster_s_left_floor(void) {
	return _hamster_s_left_floor(0);
}

int hamster_s_right_floor(void) {
	return _hamster_s_right_floor(0);
}

int hamster_s_acceleration_x(void) {
	return _hamster_s_acceleration_x(0);
}

int hamster_s_acceleration_y(void) {
	return _hamster_s_acceleration_y(0);
}

int hamster_s_acceleration_z(void) {
	return _hamster_s_acceleration_z(0);
}

int hamster_s_light(void) {
	return _hamster_s_light(0);
}

int hamster_s_temperature(void) {
	return _hamster_s_temperature(0);
}

int hamster_s_input_a(void) {
	return _hamster_s_input_a(0);
}

int hamster_s_input_b(void) {
	return _hamster_s_input_b(0);
}

float hamster_s_voltage_a(void) {
	return _hamster_s_voltage_a(0);
}

float hamster_s_voltage_b(void) {
	return _hamster_s_voltage_b(0);
}

int hamster_s_free_fall(void) {
	return _hamster_s_free_fall(0);
}

int hamster_s_tap(void) {
	return _hamster_s_tap(0);
}

int hamster_s_pulse_count(void) {
	return _hamster_s_pulse_count(0);
}

int hamster_s_tilt(void) {
	return _hamster_s_tilt(0);
}

int hamster_s_battery_state(void) {
	return _hamster_s_battery_state(0);
}

/*------------------------------
  TURTLE
------------------------------*/

#define _TURTLE_LEFT_WHEEL_INDEX 0
#define _TURTLE_RIGHT_WHEEL_INDEX 1
#define _TURTLE_LED_INDEX 2
#define _TURTLE_BUZZER_INDEX 3
#define _TURTLE_PULSE_INDEX 4
#define _TURTLE_NOTE_INDEX 5
#define _TURTLE_SOUND_INDEX 6
#define _TURTLE_LINE_TRACER_MODE_INDEX 7
#define _TURTLE_LINE_TRACER_GAIN_INDEX 8
#define _TURTLE_LINE_TRACER_SPEED_INDEX 9
#define _TURTLE_LAMP_INDEX 10
#define _TURTLE_LOCK_INDEX 11
#define _TURTLE_SIGNAL_STRENGTH_INDEX 12
#define _TURTLE_COLOR_INDEX 13
#define _TURTLE_FLOOR_INDEX 14
#define _TURTLE_ACCELERATION_INDEX 15
#define _TURTLE_TEMPERATURE_INDEX 16
#define _TURTLE_BUTTON_INDEX 17
#define _TURTLE_CLICKED_INDEX 18
#define _TURTLE_DOUBLE_CLICKED_INDEX 19
#define _TURTLE_LONG_PRESSED_INDEX 20
#define _TURTLE_COLOR_NUMBER_INDEX 21
#define _TURTLE_COLOR_PATTERN_INDEX 22
#define _TURTLE_PULSE_COUNT_INDEX 23
#define _TURTLE_WHEEL_STATE_INDEX 24
#define _TURTLE_SOUND_STATE_INDEX 25
#define _TURTLE_LINE_TRACER_STATE_INDEX 26
#define _TURTLE_TILT_INDEX 27
#define _TURTLE_BATTERY_STATE_INDEX 28

#define _GROUP_TURTLE 2

#define _TURTLE_LINE_TRACER_MODE_OFF 0
#define _TURTLE_LINE_TRACER_MODE_BLACK 10
#define _TURTLE_LINE_TRACER_MODE_RED 11
#define _TURTLE_LINE_TRACER_MODE_GREEN 13
#define _TURTLE_LINE_TRACER_MODE_BLUE 15
#define _TURTLE_LINE_TRACER_MODE_ANY 17
#define _TURTLE_LINE_TRACER_MODE_TURN_LEFT 20
#define _TURTLE_LINE_TRACER_MODE_TURN_RIGHT 30
#define _TURTLE_LINE_TRACER_MODE_CROSS 40
#define _TURTLE_LINE_TRACER_MODE_UTURN 50
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_RED 61
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_YELLOW 62
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_GREEN 63
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_CYAN 64
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_BLUE 65
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA 66
#define _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_ANY 67
#define _TURTLE_LINE_TRACER_MODE_RED_UNTIL_BLACK 71
#define _TURTLE_LINE_TRACER_MODE_GREEN_UNTIL_BLACK 73
#define _TURTLE_LINE_TRACER_MODE_BLUE_UNTIL_BLACK 75
#define _TURTLE_LINE_TRACER_MODE_ANY_UNTIL_BLACK 77

#define _TURTLE_SOUND_OFF 0
#define _TURTLE_SOUND_BEEP 1
#define _TURTLE_SOUND_RANDOM_BEEP 2
#define _TURTLE_SOUND_SIREN 3
#define _TURTLE_SOUND_ENGINE 4
#define _TURTLE_SOUND_ROBOT 5
#define _TURTLE_SOUND_MARCH 6
#define _TURTLE_SOUND_BIRTHDAY 7
#define _TURTLE_SOUND_DIBIDIBIDIP 8
#define _TURTLE_SOUND_GOOD_JOB 9

#define _TURTLE_COLOR_BLACK "BLACK"
#define _TURTLE_COLOR_RED "RED"
#define _TURTLE_COLOR_ORANGE "ORANGE"
#define _TURTLE_COLOR_YELLOW "YELLOW"
#define _TURTLE_COLOR_GREEN "GREEN"
#define _TURTLE_COLOR_SKY_BLUE "SKY_BLUE"
#define _TURTLE_COLOR_CYAN "CYAN"
#define _TURTLE_COLOR_BLUE "BLUE"
#define _TURTLE_COLOR_PURPLE "PURPLE"
#define _TURTLE_COLOR_MAGENTA "MAGENTA"
#define _TURTLE_COLOR_WHITE "WHITE"

#define _TURTLE_CM_TO_PULSE 106.8f
#define _TURTLE_DEG_TO_PULSE 1570.0f
#define _TURTLE_WHEEL_CENTER_DISTANCE 2.3125f

struct _turtle_robot {
	struct _robot robot;
	float left_wheel;
	float right_wheel;
	int led_red;
	int led_green;
	int led_blue;
	float buzzer;
	int pulse;
	int note;
	int sound;
	int line_tracer_mode;
	int line_tracer_gain;
	int line_tracer_speed;
	int lamp;
	int lock;
	int acceleration_x[10];
	int acceleration_y[10];
	int acceleration_z[10];
	float acceleration_sum_x;
	float acceleration_sum_y;
	float acceleration_sum_z;
	int acceleration_index;
	int acceleration_count;
	int clicked_id;
	int long_pressed_id;
	int wheel_id;
	int wheel_pulse;
	int wheel_pulse_prev;
	int wheel_event;
	int wheel_state;
	int wheel_count;
	int pulse_written;
	int current_sound;
	int sound_repeat;
	int sound_flag;
	int sound_event;
	int sound_state;
	int sound_count;
	int sound_written;
	int line_tracer_flag;
	int line_tracer_event;
	int line_tracer_state;
	int line_tracer_count;
	int line_tracer_mode_written;
	int event_button;
	int event_color_number;
	int event_color_pattern;
	int event_pulse_count;
	int event_tilt;
	int event_battery_state;
	double spin_left;
	double spin_right;
	double pivot_left;
	double pivot_right;
	double wheel_center_distance;
	double bpm;
};

void _turtle_parse(struct _turtle_robot* turtle, char* buffer) {
	int value = _hex_to_value(buffer, 2, 4);
	if(value > 0x7F) value -= 0x100;
	turtle->spin_right = value + 1;
	value = _hex_to_value(buffer, 4, 6);
	if(value > 0x7F) value -= 0x100;
	turtle->pivot_right = value + 1;
	value = _hex_to_value(buffer, 6, 8);
	if(value > 0x7F) value -= 0x100;
	turtle->spin_left = value + 1;
	value = _hex_to_value(buffer, 8, 10);
	if(value > 0x7F) value -= 0x100;
	turtle->pivot_left = value + 1;
	value = _hex_to_value(buffer, 10, 14);
	if(value > 0x7FFF) value -= 0x10000;
	turtle->wheel_center_distance = _TURTLE_WHEEL_CENTER_DISTANCE + (value + 1) / 10000.0;
	if(turtle->wheel_center_distance <= 0) turtle->wheel_center_distance = _TURTLE_WHEEL_CENTER_DISTANCE;
}

void _turtle_load_param(struct _connector* connector, struct _serial* serial) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)connector->arg;
	int i, loaded = 0, count = 0, read_bytes;
	char write_buffer[54];
	char* buf;
	const char* address;
	
	strcpy(write_buffer, "7000000000000000000000000000000000004400-");
	buf = &write_buffer[41];
	address = _connector_get_address(connector);
	for(i = 0; i < 12; ++i) {
		*(buf++) = address[i];
	}
	*buf = '\r';
	
	while(1) {
		read_bytes = _connector_read_packet(connector, serial, NULL);
		if(read_bytes > 0) {
			buf = connector->buffer;
			if(loaded == 1) {
				if(buf[0] == 'B' && buf[1] == '0') {
					_turtle_parse(turtle, buf);
					write_buffer[0] = '7';
					_serial_write(serial, write_buffer, 54);
				} else if(buf[0] == '1') {
					return;
				}
			} else {
				if(buf[0] == 'B' && buf[1] == '0') {
					_turtle_parse(turtle, buf);
					write_buffer[0] = '7';
					_serial_write(serial, write_buffer, 54);
					loaded = 1;
				} else {
					if(count % 10 == 0) {
						write_buffer[0] = 'B';
						_serial_write(serial, write_buffer, 54);
					}
					count ++;
				}
			}
		}
		Sleep(5);
	}
}

int _turtle_check_connection(struct _connector* connector, struct _serial* serial) {
	int i, read_bytes;

	for(i = 0; i < _RETRY; ++i) {
		_serial_write(serial, "FF\r", 3);
		read_bytes = _connector_read_packet(connector, serial, "FF");
		if(read_bytes != 0) {
			char tmp[5][256];
			int index = 0, len;
			char* buf;
			char* comma;

			buf = connector->buffer;
			comma = strchr(buf, ',');
			while(comma != NULL && index < 5) {
				len = comma - buf;
				_STRNCPY(tmp[index], 256, buf, len);
				tmp[index][len] = '\0';
				index ++;
				buf = comma + 1;
				comma = strchr(buf, ',');
			}
			if(index >= 4 && buf != NULL && strlen(buf) >= 12) {
				if(_STRICMP(tmp[1], "Turtle") == 0 && _STRICMP(tmp[2], "09") == 0) {
					buf[12] = '\0';
					_connector_set_address(connector, buf);
					_turtle_load_param(connector, serial);
					_connector_set_connection_state(connector, _CONNECTION_STATE_CONNECTED);
					return _CONNECTION_RESULT_FOUND;
				}
			}
			return _CONNECTION_RESULT_NOT_AVAILABLE;
		}
	}
	return _CONNECTION_RESULT_NOT_AVAILABLE;
}

struct _device* _turtle_find_device(struct _robot* robot, int device_id) {
	if(robot == NULL) return NULL;
	if(robot->devices == NULL) return NULL;
	switch(device_id) {
		case TURTLE_LEFT_WHEEL: return robot->devices[_TURTLE_LEFT_WHEEL_INDEX];
		case TURTLE_RIGHT_WHEEL: return robot->devices[_TURTLE_RIGHT_WHEEL_INDEX];
		case TURTLE_LED: return robot->devices[_TURTLE_LED_INDEX];
		case TURTLE_BUZZER: return robot->devices[_TURTLE_BUZZER_INDEX];
		case TURTLE_PULSE: return robot->devices[_TURTLE_PULSE_INDEX];
		case TURTLE_NOTE: return robot->devices[_TURTLE_NOTE_INDEX];
		case TURTLE_SOUND: return robot->devices[_TURTLE_SOUND_INDEX];
		case TURTLE_LINE_TRACER_MODE: return robot->devices[_TURTLE_LINE_TRACER_MODE_INDEX];
		case TURTLE_LINE_TRACER_GAIN: return robot->devices[_TURTLE_LINE_TRACER_GAIN_INDEX];
		case TURTLE_LINE_TRACER_SPEED: return robot->devices[_TURTLE_LINE_TRACER_SPEED_INDEX];
		case TURTLE_LAMP: return robot->devices[_TURTLE_LAMP_INDEX];
		case TURTLE_LOCK: return robot->devices[_TURTLE_LOCK_INDEX];
		case TURTLE_SIGNAL_STRENGTH: return robot->devices[_TURTLE_SIGNAL_STRENGTH_INDEX];
		case TURTLE_COLOR: return robot->devices[_TURTLE_COLOR_INDEX];
		case TURTLE_FLOOR: return robot->devices[_TURTLE_FLOOR_INDEX];
		case TURTLE_ACCELERATION: return robot->devices[_TURTLE_ACCELERATION_INDEX];
		case TURTLE_TEMPERATURE: return robot->devices[_TURTLE_TEMPERATURE_INDEX];
		case TURTLE_BUTTON: return robot->devices[_TURTLE_BUTTON_INDEX];
		case TURTLE_CLICKED: return robot->devices[_TURTLE_CLICKED_INDEX];
		case TURTLE_DOUBLE_CLICKED: return robot->devices[_TURTLE_DOUBLE_CLICKED_INDEX];
		case TURTLE_LONG_PRESSED: return robot->devices[_TURTLE_LONG_PRESSED_INDEX];
		case TURTLE_COLOR_NUMBER: return robot->devices[_TURTLE_COLOR_NUMBER_INDEX];
		case TURTLE_COLOR_PATTERN: return robot->devices[_TURTLE_COLOR_PATTERN_INDEX];
		case TURTLE_PULSE_COUNT: return robot->devices[_TURTLE_PULSE_COUNT_INDEX];
		case TURTLE_WHEEL_STATE: return robot->devices[_TURTLE_WHEEL_STATE_INDEX];
		case TURTLE_SOUND_STATE: return robot->devices[_TURTLE_SOUND_STATE_INDEX];
		case TURTLE_LINE_TRACER_STATE: return robot->devices[_TURTLE_LINE_TRACER_STATE_INDEX];
		case TURTLE_TILT: return robot->devices[_TURTLE_TILT_INDEX];
		case TURTLE_BATTERY_STATE: return robot->devices[_TURTLE_BATTERY_STATE_INDEX];
	}
	return NULL;
}

void _turtle_request_motoring_data(struct _robot* robot) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	struct _device** devices = robot->devices;
	
	turtle->left_wheel = _device_read_float(devices[_TURTLE_LEFT_WHEEL_INDEX]);
	turtle->right_wheel = _device_read_float(devices[_TURTLE_RIGHT_WHEEL_INDEX]);
	turtle->led_red = _device_read_at(devices[_TURTLE_LED_INDEX], 0);
	turtle->led_green = _device_read_at(devices[_TURTLE_LED_INDEX], 1);
	turtle->led_blue = _device_read_at(devices[_TURTLE_LED_INDEX], 2);
	turtle->buzzer = _device_read_float(devices[_TURTLE_BUZZER_INDEX]);
	if(devices[_TURTLE_PULSE_INDEX]->written == 1) {
		turtle->pulse = _device_read(devices[_TURTLE_PULSE_INDEX]);
		turtle->pulse_written = 1;
	}
	if(devices[_TURTLE_NOTE_INDEX]->written == 1) {
		turtle->note = _device_read(devices[_TURTLE_NOTE_INDEX]);
	}
	if(devices[_TURTLE_SOUND_INDEX]->written == 1) {
		turtle->sound = _device_read(devices[_TURTLE_SOUND_INDEX]);
		turtle->sound_written = 1;
	}
	if(devices[_TURTLE_LINE_TRACER_MODE_INDEX]->written == 1) {
		turtle->line_tracer_mode = _device_read(devices[_TURTLE_LINE_TRACER_MODE_INDEX]);
		turtle->line_tracer_mode_written = 1;
	}
	if(devices[_TURTLE_LINE_TRACER_GAIN_INDEX]->written == 1) {
		turtle->line_tracer_gain = _device_read(devices[_TURTLE_LINE_TRACER_GAIN_INDEX]);
	}
	if(devices[_TURTLE_LINE_TRACER_SPEED_INDEX]->written == 1) {
		turtle->line_tracer_speed = _device_read(devices[_TURTLE_LINE_TRACER_SPEED_INDEX]);
	}
	if(devices[_TURTLE_LAMP_INDEX]->written == 1) {
		turtle->lamp = _device_read(devices[_TURTLE_LAMP_INDEX]);
	}
	if(devices[_TURTLE_LOCK_INDEX]->written == 1) {
		turtle->lock = _device_read(devices[_TURTLE_LOCK_INDEX]);
	}
	_robot_clear_written(robot);
}

int _turtle_cm_to_pulse(double cm) {
	return _round_double_value(cm * _TURTLE_CM_TO_PULSE);
}

int _turtle_calc_turn_left_pulse(struct _turtle_robot* turtle, double degree) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE + turtle->spin_left) / 360.0);
}

int _turtle_calc_turn_right_pulse(struct _turtle_robot* turtle, double degree) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE + turtle->spin_right) / 360.0);
}

int _turtle_calc_pivot_left_pulse(struct _turtle_robot* turtle, double degree) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE * 2 + turtle->pivot_left) / 360.0);
}

int _turtle_calc_pivot_right_pulse(struct _turtle_robot* turtle, double degree) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE * 2 + turtle->pivot_right) / 360.0);
}

double _turtle_calc_inner_velocity(struct _turtle_robot* turtle, double outer_velocity, double radius) {
	return outer_velocity * (radius - turtle->wheel_center_distance) / (radius + turtle->wheel_center_distance);
}

int _turtle_calc_circle_left_pulse(struct _turtle_robot* turtle, double degree, double radius) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE + turtle->spin_left) * (radius + turtle->wheel_center_distance) / 360.0 / turtle->wheel_center_distance);
}

int _turtle_calc_circle_right_pulse(struct _turtle_robot* turtle, double degree, double radius) {
	return _round_double_value(degree * (_TURTLE_DEG_TO_PULSE + turtle->spin_right) * (radius + turtle->wheel_center_distance) / 360.0 / turtle->wheel_center_distance);
}

void _turtle_encode_motoring_packet(struct _robot* robot) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	char* buffer = robot->write_buffer;
	int index = 0, temp, t, i, len;
	const char* address;

	buffer[index++] = '1';
	buffer[index++] = '0';
	index = _value_to_hex(buffer, index, _round_float_value(turtle->left_wheel * 10.68f), 2);
	index = _value_to_hex(buffer, index, _round_float_value(turtle->right_wheel * 10.68f), 2);
	turtle->wheel_pulse = turtle->pulse;
	if(turtle->pulse_written == 1) {
		if(turtle->wheel_pulse != 0 || turtle->wheel_pulse_prev != 0) {
			turtle->wheel_id = (turtle->wheel_id % 255) + 1;
		}
		turtle->wheel_count = 0;
		if(turtle->wheel_pulse > 0) {
			turtle->wheel_event = 1;
		} else {
			turtle->wheel_event = 0;
		}
		turtle->wheel_pulse_prev = turtle->wheel_pulse;
		turtle->pulse_written = 0;
	}
	index = _value_to_hex(buffer, index, turtle->wheel_id, 1);
	index = _value_to_hex(buffer, index, turtle->wheel_pulse, 2);
	
	temp = 0;
	switch(turtle->line_tracer_mode) {
		case _TURTLE_LINE_TRACER_MODE_OFF: temp = 0x00; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK: temp = 0x08; break;
		case _TURTLE_LINE_TRACER_MODE_RED: temp = 0x09; break;
		case _TURTLE_LINE_TRACER_MODE_GREEN: temp = 0x0b; break;
		case _TURTLE_LINE_TRACER_MODE_BLUE: temp = 0x0d; break;
		case _TURTLE_LINE_TRACER_MODE_ANY: temp = 0x0f; break;
		case _TURTLE_LINE_TRACER_MODE_TURN_LEFT: temp = 0x10; break;
		case _TURTLE_LINE_TRACER_MODE_TURN_RIGHT: temp = 0x18; break;
		case _TURTLE_LINE_TRACER_MODE_CROSS: temp = 0x20; break;
		case _TURTLE_LINE_TRACER_MODE_UTURN: temp = 0x28; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_RED: temp = 0x31; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_YELLOW: temp = 0x32; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_GREEN: temp = 0x33; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_CYAN: temp = 0x34; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_BLUE: temp = 0x35; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA: temp = 0x36; break;
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_ANY: temp = 0x37; break;
		case _TURTLE_LINE_TRACER_MODE_RED_UNTIL_BLACK: temp = 0x39; break;
		case _TURTLE_LINE_TRACER_MODE_GREEN_UNTIL_BLACK: temp = 0x3b; break;
		case _TURTLE_LINE_TRACER_MODE_BLUE_UNTIL_BLACK: temp = 0x3d; break;
		case _TURTLE_LINE_TRACER_MODE_ANY_UNTIL_BLACK: temp = 0x3f; break;
	}
	if(turtle->line_tracer_mode_written == 1) {
		turtle->line_tracer_count = 0;
		if(temp > 0) {
			turtle->line_tracer_flag ^= 0x80;
			turtle->line_tracer_event = 1;
		} else {
			turtle->line_tracer_event = 0;
		}
		turtle->line_tracer_mode_written = 0;
	}
	temp |= turtle->line_tracer_flag & 0x80;
	index = _value_to_hex(buffer, index, temp, 1);
	index = _value_to_hex(buffer, index, turtle->led_red, 1);
	index = _value_to_hex(buffer, index, turtle->led_green, 1);
	index = _value_to_hex(buffer, index, turtle->led_blue, 1);
	temp = (turtle->lock & 0x01) << 4;
	temp |= (1 - turtle->lamp) & 0x01;
	index = _value_to_hex(buffer, index, temp, 1);
	index = _value_to_hex(buffer, index, _round_float_value(turtle->buzzer * 100), 3);
	index = _value_to_hex(buffer, index, turtle->note, 1);
	temp = 0;
	switch(turtle->sound) {
		case _TURTLE_SOUND_OFF: temp = 0x00; break;
		case _TURTLE_SOUND_BEEP: temp = 0x01; break;
		case _TURTLE_SOUND_RANDOM_BEEP: temp = 0x05; break;
		case _TURTLE_SOUND_SIREN: temp = 0x10; break;
		case _TURTLE_SOUND_ENGINE: temp = 0x20; break;
		case _TURTLE_SOUND_ROBOT: temp = 0x30; break;
		case _TURTLE_SOUND_MARCH: temp = 0x40; break;
		case _TURTLE_SOUND_BIRTHDAY: temp = 0x41; break;
		case _TURTLE_SOUND_DIBIDIBIDIP: temp = 0x42; break;
		case _TURTLE_SOUND_GOOD_JOB: temp = 0x43; break;
	}
	if(turtle->sound_written == 1) {
		turtle->sound_count = 0;
		if(temp > 0) {
			turtle->sound_flag ^= 0x80;
			turtle->sound_event = 1;
		} else {
			turtle->sound_event = 0;
		}
		turtle->sound_written = 0;
	}
	temp |= turtle->sound_flag & 0x80;
	index = _value_to_hex(buffer, index, temp, 1);
	temp = ((turtle->line_tracer_gain - 1) & 0x0f) << 4;
	temp |= (turtle->line_tracer_speed - 1) & 0x0f;
	index = _value_to_hex(buffer, index, temp, 1);
	buffer[index++] = '0';
	buffer[index++] = '0';
	buffer[index++] = '-';
	address = _connector_get_address(robot->connector);
	for(i = 0; i < 12; ++i) {
		buffer[index++] = address[i];
	}
	buffer[index++] = '\r';
	_connector_write(robot->connector, buffer, _MOTORING_PACKET_LENGTH);
}

void _turtle_run_sound(struct _robot* robot, int sound, int repeat) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	
	if(turtle == NULL) return;
	if(repeat < 0) repeat = -1;
	if(repeat != 0) {
		turtle->current_sound = sound;
		turtle->sound_repeat = repeat;
		_robot_write(robot, TURTLE_SOUND, sound);
	}
}

int _turtle_decode_sensory_packet(struct _robot* robot, char* packet) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	struct _device** devices = robot->devices;
	char* buffer = packet;
	int value, red, green, blue, r, g, b, color_number, clicked_id, long_pressed_id, pattern, pattern1, pattern2, acc_x, acc_y, acc_z, state;
	
	value = _hex_to_value(buffer, 0, 1);
	if(value != 1) return 0;
	
	red = _hex_to_value(buffer, 2, 6);
	green = _hex_to_value(buffer, 6, 10);
	blue = _hex_to_value(buffer, 10, 14);
	value = _hex_to_value(buffer, 14, 16);
	
	r = _round_float_value(red * 255.0f / 1023);
	g = _round_float_value(green * 255.0f / 1023);
	b = _round_float_value(blue * 255.0f / 1023);
	if(r > 255) r = 255;
	else if(r < 0) r = 0;
	if(g > 255) g = 255;
	else if(g < 0) g = 0;
	if(b > 255) b = 255;
	else if(b < 0) b = 0;
	_device_put_at(devices[_TURTLE_COLOR_INDEX], 0, r, 1);
	_device_put_at(devices[_TURTLE_COLOR_INDEX], 1, g, 1);
	_device_put_at(devices[_TURTLE_COLOR_INDEX], 2, b, 1);
	_device_put_at(devices[_TURTLE_COLOR_INDEX], 3, value, 1);
	color_number = TURTLE_COLOR_NONE;
	if(value < 75) {
		if(red > 600 && green > 600 && blue > 600) {
			color_number = TURTLE_COLOR_WHITE;
		} else if(green < red * 3/10 && blue < red * 3/10 && red > 350) {
			color_number = TURTLE_COLOR_RED;
		} else if(blue < red * 5/10 && blue < green * 6/10 && red > 300 && green > 250) {
			if(green < red * 15/20) {
				color_number = TURTLE_COLOR_ORANGE;
			} else {
				color_number = TURTLE_COLOR_YELLOW;
			}
		} else if(red < green * 6/10 && blue < green * 6/10 && green > 300) {
			color_number = TURTLE_COLOR_GREEN;
		} else if(red < blue * 6/10 && green < blue * 7/10 && blue > 250) {
			color_number = TURTLE_COLOR_BLUE;
		} else if(red < green * 6/10 && red < blue * 6/10 && blue > 250 && green > 250) {
			color_number = TURTLE_COLOR_CYAN;
		} else if(green < red * 7/10 && green < blue * 7/10 && red > 250) {
			color_number = TURTLE_COLOR_MAGENTA;
		} else if(red < 300 && green < 300 && blue < 250) {
			color_number = TURTLE_COLOR_BLACK;
		}
	}
	if(color_number != turtle->event_color_number) {
		_device_put(devices[_TURTLE_COLOR_NUMBER_INDEX], color_number, (turtle->event_color_number == -2) ? 0 : 1);
		turtle->event_color_number = color_number;
	}
	value = _hex_to_value(buffer, 16, 18);
	clicked_id = (value >> 2) & 0x03;
	long_pressed_id = (value >> 4) & 0x03;
	if(turtle->clicked_id < 0) {
		turtle->clicked_id = clicked_id;
	} else if(clicked_id != turtle->clicked_id) {
		turtle->clicked_id = clicked_id;
		value = value & 0x03;
		if(value == 1) {
			_device_put_empty(devices[_TURTLE_CLICKED_INDEX], 1);
		} else if(value == 2) {
			_device_put_empty(devices[_TURTLE_DOUBLE_CLICKED_INDEX], 1);
		}
	}
	if(turtle->long_pressed_id < 0) {
		turtle->long_pressed_id = long_pressed_id;
	} else if(long_pressed_id != turtle->long_pressed_id) {
		turtle->long_pressed_id = long_pressed_id;
		_device_put_empty(devices[_TURTLE_LONG_PRESSED_INDEX], 1);
	}
	value = _hex_to_value(buffer, 18, 20);
	_device_put(devices[_TURTLE_FLOOR_INDEX], value, 1);
	value = _hex_to_value(buffer, 20, 22);
	pattern = (value >> 1) & 0x7F;
	if(turtle->event_color_pattern < 0) {
		turtle->event_color_pattern = pattern;
	} else if(pattern != turtle->event_color_pattern) {
		turtle->event_color_pattern = pattern;
		pattern1 = (pattern >> 3) & 0x07;
		if(pattern1 > 1) pattern1++;
		pattern2 = pattern & 0x07;
		if(pattern2 > 1) pattern2++;
		_device_put(devices[_TURTLE_COLOR_PATTERN_INDEX], pattern1 * 10 + pattern2, 1);
	}
	value = value & 0x01;
	if(value != turtle->event_button) {
		_device_put(devices[_TURTLE_BUTTON_INDEX], value, (turtle->event_button == -1) ? 0 : 1);
		turtle->event_button = value;
	}
	value = _hex_to_value(buffer, 22, 26);
	if(value != turtle->event_pulse_count) {
		_device_put(devices[_TURTLE_PULSE_COUNT_INDEX], value, (turtle->event_pulse_count == -1) ? 0 : 1);
		turtle->event_pulse_count = value;
	}
	if(turtle->acceleration_count < 10) {
		++turtle->acceleration_count;
	} else {
		turtle->acceleration_index %= 10;
		turtle->acceleration_sum_x -= turtle->acceleration_x[turtle->acceleration_index];
		turtle->acceleration_sum_y -= turtle->acceleration_y[turtle->acceleration_index];
		turtle->acceleration_sum_z -= turtle->acceleration_z[turtle->acceleration_index];
	}
	value = _hex_to_value(buffer, 26, 28);
	if(value > 0x7F) value -= 0x100;
	value *= 256;
	turtle->acceleration_sum_x += value;
	turtle->acceleration_x[turtle->acceleration_index] = value;
	value = _hex_to_value(buffer, 28, 30);
	if(value > 0x7F) value -= 0x100;
	value *= 256;
	turtle->acceleration_sum_y += value;
	turtle->acceleration_y[turtle->acceleration_index] = value;
	value = _hex_to_value(buffer, 30, 32);
	if(value > 0x7F) value -= 0x100;
	value *= 256;
	turtle->acceleration_sum_z += value;
	turtle->acceleration_z[turtle->acceleration_index] = value;
	++turtle->acceleration_index;
	acc_x = _round_float_value(turtle->acceleration_sum_x / turtle->acceleration_count);
	acc_y = _round_float_value(turtle->acceleration_sum_y / turtle->acceleration_count);
	acc_z = _round_float_value(turtle->acceleration_sum_z / turtle->acceleration_count);
	_device_put_at(devices[_TURTLE_ACCELERATION_INDEX], 0, acc_x, 1);
	_device_put_at(devices[_TURTLE_ACCELERATION_INDEX], 1, acc_y, 1);
	_device_put_at(devices[_TURTLE_ACCELERATION_INDEX], 2, acc_z, 1);
	if(acc_z < 8192 && acc_x > 8192 && acc_y > -4096 && acc_y < 4096) value = 1;
	else if(acc_z < 8192 && acc_x < -8192 && acc_y > -4096 && acc_y < 4096) value = -1;
	else if(acc_z < 8192 && acc_y > 8192 && acc_x > -4096 && acc_x < 4096) value = 2;
	else if(acc_z < 8192 && acc_y < -8192 && acc_x > -4096 && acc_x < 4096) value = -2;
	else if(acc_z > 12288 && acc_x > -8192 && acc_x < 8192 && acc_y > -8192 && acc_y < 8192) value = 3;
	else if(acc_z < -12288 && acc_x > -4096 && acc_x < 4096 && acc_y > -4096 && acc_y < 4096) value = -3;
	else value = 0;
	if(value != turtle->event_tilt) {
		_device_put(devices[_TURTLE_TILT_INDEX], value, (turtle->event_tilt == -4) ? 0 : 1);
		turtle->event_tilt = value;
	}
	value = _hex_to_value(buffer, 32, 34);
	if(value > 0x7F) value -= 0x100;
	_device_put(devices[_TURTLE_TEMPERATURE_INDEX], _round_float_value(value / 2.0f + 24), 1);
	value = _hex_to_value(buffer, 34, 36);
	value -= 0x100;
	_device_put(devices[_TURTLE_SIGNAL_STRENGTH_INDEX], value, 1);
	value = _hex_to_value(buffer, 36, 38) + 200;
	if(value < 365) state = 0;
	else if(value < 375) state = 1;
	else state = 2;
	if(state != turtle->event_battery_state) {
		_device_put(devices[_TURTLE_BATTERY_STATE_INDEX], state, (turtle->event_battery_state == -1) ? 0 : 1);
		turtle->event_battery_state = state;
	}
	value = _hex_to_value(buffer, 38, 40);
	state = (value >> 4) & 0x01;
	if(turtle->wheel_event == 1) {
		if(state == 0) {
			if(turtle->wheel_pulse > 0 && turtle->wheel_pulse < 20) {
				if(++turtle->wheel_count > 8) turtle->wheel_event = 2;
			}
		} else {
			turtle->wheel_event = 2;
		}
	}
	if(turtle->wheel_event == 2) {
		if(state != turtle->wheel_state || turtle->wheel_count > 8) {
			turtle->wheel_state = state;
			_device_put(devices[_TURTLE_WHEEL_STATE_INDEX], state, 1);
			if(state == 0) {
				turtle->wheel_event = 0;
				turtle->wheel_count = 0;
			}
		}
	}
	state = value & 0x03;
	if(state > 1) state = 1;
	if(turtle->sound_event == 1) {
		if(state == 0) {
			if(++turtle->sound_count > 5) turtle->sound_event = 2;
		} else {
			turtle->sound_event = 2;
		}
	}
	if(turtle->sound_event == 2) {
		if(state != turtle->sound_state || turtle->sound_count > 5) {
			turtle->sound_state = state;
			if(state == 0) {
				turtle->sound_event = 0;
				turtle->sound_count = 0;
				if(turtle->current_sound > 0) {
					if(turtle->sound_repeat < 0) {
						_turtle_run_sound(robot, turtle->current_sound, -1);
					} else if(turtle->sound_repeat > 1) {
						turtle->sound_repeat --;
						_turtle_run_sound(robot, turtle->current_sound, turtle->sound_repeat);
					} else {
						turtle->current_sound = 0;
						turtle->sound_repeat = 1;
						_device_put(devices[_TURTLE_SOUND_STATE_INDEX], state, 1);
					}
				} else {
					turtle->current_sound = 0;
					turtle->sound_repeat = 1;
				}
			}
		}
	}
	state = (value >> 2) & 0x03;
	if((state & 0x02) != 0) {
		if(turtle->line_tracer_event == 1) {
			if(state == 0x02) {
				if(++turtle->line_tracer_count > 5) turtle->line_tracer_event = 2;
			} else {
				turtle->line_tracer_event = 2;
			}
		}
		if(turtle->line_tracer_event == 2) {
			if(state != turtle->line_tracer_state || turtle->line_tracer_count > 5) {
				turtle->line_tracer_state = state;
				_device_put(devices[_TURTLE_LINE_TRACER_STATE_INDEX], state, 1);
				if(state == 0x02) {
					turtle->line_tracer_event = 0;
					turtle->line_tracer_count = 0;
				}
			}
		}
	}
	return 1;
}

int _turtle_receive(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		int read_bytes = _connector_read(connector);
		if(read_bytes != 0) {
			if(_turtle_decode_sensory_packet(robot, connector->buffer) == 1) {
				if(robot->ready == 0) {
					robot->ready = 1;
					_runner_register_checked();
				}
			}
			return 1;
		}
	}
	return 0;
}

void _turtle_send(struct _robot* robot) {
	struct _connector* connector = robot->connector;
	
	if(connector != NULL) {
		_turtle_encode_motoring_packet(robot);
	}
}

void _turtle_update_sensory_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _TURTLE_SIGNAL_STRENGTH_INDEX; i <= _TURTLE_BATTERY_STATE_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _turtle_update_motoring_device_state(struct _robot* robot) {
	struct _device** devices = robot->devices;
	int i;

	for(i = _TURTLE_LEFT_WHEEL_INDEX; i <= _TURTLE_LOCK_INDEX; ++i) {
		_device_update_device_state(devices[i]);
	}
}

void _turtle_reset(struct _robot* robot) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	
	if(turtle == NULL) return;
	_robot_reset(robot);
	
	turtle->left_wheel = 0;
	turtle->right_wheel = 0;
	turtle->led_red = 0;
	turtle->led_green = 0;
	turtle->led_blue = 0;
	turtle->buzzer = 0;
	turtle->pulse = 0;
	turtle->note = 0;
	turtle->sound = 0;
	turtle->line_tracer_mode = 0;
	turtle->line_tracer_gain = 5;
	turtle->line_tracer_speed = 5;
	turtle->lamp = 1;
	turtle->lock = 0;
	
	turtle->acceleration_sum_x = 0;
	turtle->acceleration_sum_y = 0;
	turtle->acceleration_sum_z = 0;
	turtle->acceleration_index = 0;
	turtle->acceleration_count = 0;
	
	turtle->clicked_id = -1;
	turtle->long_pressed_id = -1;
	
	//turtle->wheel_id = 0;
	turtle->wheel_pulse = 0;
	turtle->wheel_pulse_prev = -1;
	turtle->wheel_event = 0;
	turtle->wheel_state = 0;
	turtle->wheel_count = 0;
	turtle->pulse_written = 0;
	
	turtle->current_sound = 0;
	turtle->sound_repeat = 1;
	//turtle->sound_flag = 0;
	turtle->sound_event = 0;
	turtle->sound_state = 0;
	turtle->sound_count = 0;
	turtle->sound_written = 0;
	
	//turtle->line_tracer_flag = 0;
	turtle->line_tracer_event = 0;
	turtle->line_tracer_state = 0;
	turtle->line_tracer_count = 0;
	turtle->line_tracer_mode_written = 0;
	
	turtle->event_button = -1;
	turtle->event_color_number = -2;
	turtle->event_color_pattern = -1;
	turtle->event_pulse_count = -1;
	turtle->event_tilt = -4;
	turtle->event_battery_state = -1;
	
	turtle->bpm = 60;
}

void _turtle_dispose(struct _robot* robot) {
	_robot_group_remove_robot(_GROUP_TURTLE, robot);
	_runner_remove_robot(robot);

	robot->running = 0;
	if(robot->thread_alive == 1) {
		WaitForSingleObject(robot->thread_handle, 1000);
	}
	Sleep(100);
	_robot_dispose(robot);
	free((struct _turtle_robot*)robot);
}

unsigned WINAPI _turtle_thread_proc(void* arg) {
	struct _robot* robot = (struct _robot*)arg;
	int shutdown = 0, shutdown_count = 0;
	
	if(robot == NULL) return 0;

	robot->thread_alive = 1;
	while(1) {
		if(_turtle_receive(robot) == 1) {
			_turtle_send(robot);
			if(shutdown == 1) {
				break;
			}
		}
		if(shutdown == 1) {
			if(++shutdown_count > 5) {
				break;
			}
		}
		if(robot->running == 0) {
			shutdown = 1;
		}
		Sleep(5);
	}
	robot->thread_alive = 0;
	return 0;
}

struct _turtle_robot* _turtle_create(const char* port_name) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)malloc(sizeof(struct _turtle_robot));
	struct _robot* robot = (struct _robot*)turtle;
	struct _device** devices;
	struct _connector* connector;
	unsigned int thread_id;
	
	_robot_init(robot, _robot_group_count_robots(_GROUP_TURTLE), "Turtle", 55);
	
	turtle->left_wheel = 0;
	turtle->right_wheel = 0;
	turtle->led_red = 0;
	turtle->led_green = 0;
	turtle->led_blue = 0;
	turtle->buzzer = 0;
	turtle->pulse = 0;
	turtle->note = 0;
	turtle->sound = 0;
	turtle->line_tracer_mode = 0;
	turtle->line_tracer_gain = 5;
	turtle->line_tracer_speed = 5;
	turtle->lamp = 1;
	turtle->lock = 0;
	
	turtle->acceleration_sum_x = 0;
	turtle->acceleration_sum_y = 0;
	turtle->acceleration_sum_z = 0;
	turtle->acceleration_index = 0;
	turtle->acceleration_count = 0;
	
	turtle->clicked_id = -1;
	turtle->long_pressed_id = -1;
	
	turtle->wheel_id = 0;
	turtle->wheel_pulse = 0;
	turtle->wheel_pulse_prev = -1;
	turtle->wheel_event = 0;
	turtle->wheel_state = 0;
	turtle->wheel_count = 0;
	turtle->pulse_written = 0;
	
	turtle->current_sound = 0;
	turtle->sound_repeat = 1;
	turtle->sound_flag = 0;
	turtle->sound_event = 0;
	turtle->sound_state = 0;
	turtle->sound_count = 0;
	turtle->sound_written = 0;
	
	turtle->line_tracer_flag = 0;
	turtle->line_tracer_event = 0;
	turtle->line_tracer_state = 0;
	turtle->line_tracer_count = 0;
	turtle->line_tracer_mode_written = 0;
	
	turtle->event_button = -1;
	turtle->event_color_number = -2;
	turtle->event_color_pattern = -1;
	turtle->event_pulse_count = -1;
	turtle->event_tilt = -4;
	turtle->event_battery_state = -1;
	
	turtle->spin_left = 0;
	turtle->spin_right = 0;
	turtle->pivot_left = 0;
	turtle->pivot_right = 0;
	turtle->wheel_center_distance = _TURTLE_WHEEL_CENTER_DISTANCE;
	turtle->bpm = 60;
	
	robot->find_device = _turtle_find_device;
	robot->request_motoring_data = _turtle_request_motoring_data;
	robot->update_sensory_device_state = _turtle_update_sensory_device_state;
	robot->update_motoring_device_state = _turtle_update_motoring_device_state;
	robot->reset = _turtle_reset;
	robot->dispose = _turtle_dispose;
	
	robot->devices_size = 29;
	devices = (struct _device**)malloc(sizeof(struct _device*) * 29);
	devices[0] = _device_create(TURTLE_LEFT_WHEEL, "LeftWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_FLOAT, 1, -400, 400, 0);
	devices[1] = _device_create(TURTLE_RIGHT_WHEEL, "RightWheel", DEVICE_TYPE_EFFECTOR, DATA_TYPE_FLOAT, 1, -400, 400, 0);
	devices[2] = _device_create(TURTLE_LED, "Led", DEVICE_TYPE_EFFECTOR, DATA_TYPE_INTEGER, 3, 0, 255, 0);
	devices[3] = _device_create(TURTLE_BUZZER, "Buzzer", DEVICE_TYPE_EFFECTOR, DATA_TYPE_FLOAT, 1, 0, 167772.15, 0);
	devices[4] = _device_create(TURTLE_PULSE, "Pulse", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[5] = _device_create(TURTLE_NOTE, "Note", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 88, 0);
	devices[6] = _device_create(TURTLE_SOUND, "Sound", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 127, 0);
	devices[7] = _device_create(TURTLE_LINE_TRACER_MODE, "LineTracerMode", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 127, 0);
	devices[8] = _device_create(TURTLE_LINE_TRACER_GAIN, "LineTracerGain", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 8, 5);
	devices[9] = _device_create(TURTLE_LINE_TRACER_SPEED, "LineTracerSpeed", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 1, 8, 5);
	devices[10] = _device_create(TURTLE_LAMP, "Lamp", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 1, 1);
	devices[11] = _device_create(TURTLE_LOCK, "Lock", DEVICE_TYPE_COMMAND, DATA_TYPE_INTEGER, 1, 0, 1, 0);
	devices[12] = _device_create(TURTLE_SIGNAL_STRENGTH, "SignalStrength", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -128, 0, 0);
	devices[13] = _device_create(TURTLE_COLOR, "Color", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 4, 0, 255, 0);
	devices[14] = _device_create(TURTLE_FLOOR, "Floor", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, 0, 255, 0);
	devices[15] = _device_create(TURTLE_ACCELERATION, "Acceleration", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 3, -32768, 32767, 0);
	devices[16] = _device_create(TURTLE_TEMPERATURE, "Temperature", DEVICE_TYPE_SENSOR, DATA_TYPE_INTEGER, 1, -40, 88, 0);
	devices[17] = _device_create(TURTLE_BUTTON, "Button", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 1, 0);
	devices[18] = _device_create(TURTLE_CLICKED, "Clicked", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	devices[19] = _device_create(TURTLE_DOUBLE_CLICKED, "DoubleClicked", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	devices[20] = _device_create(TURTLE_LONG_PRESSED, "LongPressed", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 0, 0, 0, 0);
	devices[21] = _device_create(TURTLE_COLOR_NUMBER, "ColorNumber", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, -1, 8, -1);
	devices[22] = _device_create(TURTLE_COLOR_PATTERN, "ColorPattern", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, -1, 88, -1);
	devices[23] = _device_create(TURTLE_PULSE_COUNT, "PulseCount", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 65535, 0);
	devices[24] = _device_create(TURTLE_WHEEL_STATE, "WheelState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 1, 0);
	devices[25] = _device_create(TURTLE_SOUND_STATE, "SoundState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 1, 0);
	devices[26] = _device_create(TURTLE_LINE_TRACER_STATE, "LineTracerState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 3, 0);
	devices[27] = _device_create(TURTLE_TILT, "Tilt", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, -3, 3, 0);
	devices[28] = _device_create(TURTLE_BATTERY_STATE, "BatteryState", DEVICE_TYPE_EVENT, DATA_TYPE_INTEGER, 1, 0, 2, 2);

	robot->devices = devices;
	robot->connector = NULL;
	robot->alive = 1;

	_robot_group_add_robot(_GROUP_TURTLE, robot);
	_runner_add_robot(robot);
	_runner_start();
	
	_runner_register_required();
	robot->running = 1;
	robot->thread_handle = (HANDLE)_beginthreadex(NULL,
		0,
		_turtle_thread_proc,
		robot,
		CREATE_SUSPENDED,
		&thread_id);
	if(robot->thread_handle != 0) {
		ResumeThread(robot->thread_handle);
	}

	connector = _connector_create("Turtle", robot->index, _VALID_PACKET_LENGTH, _CR);
	robot->connector = connector;
	if(connector != NULL) {
		int result;
		connector->arg = (void*)robot;
		connector->check_connection = _turtle_check_connection;
		result = _connector_open(connector, port_name, _SERIAL_BAUDRATE_115200, _SERIAL_FLOWCONTROL_RTSCTS_IN | _SERIAL_FLOWCONTROL_RTSCTS_OUT);
		if(result == _CONNECTION_RESULT_FOUND) {
			while(robot->ready == 0 && robot->running == 1) {
				Sleep(10);
			}
		} else if(result == _CONNECTION_RESULT_NOT_AVAILABLE) {
			_runner_register_checked();
		}
	}
	return turtle;
}

const char* _turtle_get_id(void) {
	return TURTLE_ID;
}

void _turtle_wheels(int turtle_index, double left_velocity, double right_velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	_robot_write_float(robot, TURTLE_LEFT_WHEEL, (float)left_velocity);
	_robot_write_float(robot, TURTLE_RIGHT_WHEEL, (float)right_velocity);
}

void _turtle_left_wheel(int turtle_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	_robot_write_float(robot, TURTLE_LEFT_WHEEL, (float)velocity);
}

void _turtle_right_wheel(int turtle_index, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	_robot_write_float(robot, TURTLE_RIGHT_WHEEL, (float)velocity);
}

void _turtle_stop(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	_robot_write(robot, TURTLE_LEFT_WHEEL, 0);
	_robot_write(robot, TURTLE_RIGHT_WHEEL, 0);
}

int _turtle_wheel_callback(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, TURTLE_WHEEL_STATE) == 1 && _robot_read(robot, TURTLE_WHEEL_STATE) == 0) ? 1 : 0;
}

void _turtle_motion(struct _robot* robot, int pulse, double left_velocity, double right_velocity, int (*callback)(void* arg)) {
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	_robot_write(robot, TURTLE_LEFT_WHEEL, 0);
	_robot_write(robot, TURTLE_RIGHT_WHEEL, 0);
	_robot_write(robot, TURTLE_PULSE, 0);
	wait(100);
	if(pulse > 0) {
		_robot_write_float(robot, TURTLE_LEFT_WHEEL, (float)left_velocity);
		_robot_write_float(robot, TURTLE_RIGHT_WHEEL, (float)right_velocity);
		_robot_write(robot, TURTLE_PULSE, pulse);
		wait_until(callback, NULL);
	}
	_robot_write(robot, TURTLE_LEFT_WHEEL, 0);
	_robot_write(robot, TURTLE_RIGHT_WHEEL, 0);
	_robot_write(robot, TURTLE_PULSE, 0);
	wait(100);
}

void _turtle_move_forward(int turtle_index, double cm, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(cm < 0) {
		cm = -cm;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_cm_to_pulse(cm), velocity, velocity, callback);
}

void _turtle_move_backward(int turtle_index, double cm, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(cm < 0) {
		cm = -cm;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_cm_to_pulse(cm), -velocity, -velocity, callback);
}

void _turtle_turn_left(int turtle_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_calc_turn_left_pulse((struct _turtle_robot*)robot, degree), -velocity, velocity, callback);
}

void _turtle_turn_right(int turtle_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_calc_turn_right_pulse((struct _turtle_robot*)robot, degree), velocity, -velocity, callback);
}

void _turtle_pivot_left(int turtle_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_calc_pivot_left_pulse((struct _turtle_robot*)robot, degree), 0, velocity, callback);
}

void _turtle_pivot_right(int turtle_index, double degree, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	_turtle_motion(robot, _turtle_calc_pivot_right_pulse((struct _turtle_robot*)robot, degree), velocity, 0, callback);
}

void _turtle_circle_left(int turtle_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	left_velocity = _turtle_calc_inner_velocity(turtle, velocity, radius);
	_turtle_motion(robot, _turtle_calc_circle_left_pulse(turtle, degree, radius), left_velocity, velocity, callback);
}

void _turtle_circle_right(int turtle_index, double degree, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	double right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) degree = 0;
	if(degree < 0) {
		degree = -degree;
		velocity = -velocity;
	}
	right_velocity = _turtle_calc_inner_velocity(turtle, velocity, radius);
	_turtle_motion(robot, _turtle_calc_circle_right_pulse(turtle, degree, radius), velocity, right_velocity, callback);
}

void _turtle_motion_sec(struct _robot* robot, double sec, double left_velocity, double right_velocity) {
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
	if(sec < 0) {
		sec = -sec;
		left_velocity = -left_velocity;
		right_velocity = -right_velocity;
	}
	if(sec > 0) {
		_robot_write_float(robot, TURTLE_LEFT_WHEEL, (float)left_velocity);
		_robot_write_float(robot, TURTLE_RIGHT_WHEEL, (float)right_velocity);
		wait(sec * 1000);
	}
	_robot_write(robot, TURTLE_LEFT_WHEEL, 0);
	_robot_write(robot, TURTLE_RIGHT_WHEEL, 0);
}

void _turtle_move_forward_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, velocity, velocity);
}

void _turtle_move_backward_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, -velocity, -velocity);
}

void _turtle_turn_left_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, -velocity, velocity);
}

void _turtle_turn_right_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, velocity, -velocity);
}

void _turtle_pivot_left_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, 0, velocity);
}

void _turtle_pivot_right_sec(int turtle_index, double sec, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	_turtle_motion_sec(robot, sec, velocity, 0);
}

void _turtle_circle_left_sec(int turtle_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	left_velocity = _turtle_calc_inner_velocity((struct _turtle_robot*)robot, velocity, radius);
	_turtle_motion_sec(robot, sec, left_velocity, velocity);
}

void _turtle_circle_right_sec(int turtle_index, double sec, double radius, double velocity) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	double right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) sec = 0;
	right_velocity = _turtle_calc_inner_velocity((struct _turtle_robot*)robot, velocity, radius);
	_turtle_motion_sec(robot, sec, velocity, right_velocity);
}

void _turtle_move_forward_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), velocity, velocity, callback);
}

void _turtle_move_backward_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), -velocity, -velocity, callback);
}

void _turtle_turn_left_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), -velocity, velocity, callback);
}

void _turtle_turn_right_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), velocity, -velocity, callback);
}

void _turtle_pivot_left_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), 0, velocity, callback);
}

void _turtle_pivot_right_pulse(int turtle_index, double pulse, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	if(robot == NULL) return;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	_turtle_motion(robot, _round_double_value(pulse), velocity, 0, callback);
}

void _turtle_circle_left_pulse(int turtle_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	double left_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	left_velocity = _turtle_calc_inner_velocity((struct _turtle_robot*)robot, velocity, radius);
	_turtle_motion(robot, _round_double_value(pulse), left_velocity, velocity, callback);
}

void _turtle_circle_right_pulse(int turtle_index, double pulse, double radius, double velocity, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	float right_velocity;
	
	if(robot == NULL) return;
	if(radius < 0) pulse = 0;
	if(pulse < 0) {
		pulse = -pulse;
		velocity = -velocity;
	}
	right_velocity = _turtle_calc_inner_velocity((struct _turtle_robot*)robot, velocity, radius);
	_turtle_motion(robot, _round_double_value(pulse), velocity, right_velocity, callback);
}

int _turtle_line_tracer_mode_callback(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
    
	if(robot == NULL) return 0;
	return (_robot_e(robot, TURTLE_LINE_TRACER_STATE) == 1 && _robot_read(robot, TURTLE_LINE_TRACER_STATE) == 0x02) ? 1 : 0;
}

void _turtle_line_tracer_mode(int turtle_index, int mode, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_LEFT_WHEEL, 0);
	_robot_write(robot, TURTLE_RIGHT_WHEEL, 0);
	_robot_write(robot, TURTLE_PULSE, 0);
	_robot_write(robot, TURTLE_LINE_TRACER_MODE, mode);
	
	switch(mode) {
		case _TURTLE_LINE_TRACER_MODE_TURN_LEFT:
		case _TURTLE_LINE_TRACER_MODE_TURN_RIGHT:
		case _TURTLE_LINE_TRACER_MODE_CROSS:
		case _TURTLE_LINE_TRACER_MODE_UTURN:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_RED:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_YELLOW:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_GREEN:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_CYAN:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_BLUE:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA:
		case _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_ANY:
		case _TURTLE_LINE_TRACER_MODE_RED_UNTIL_BLACK:
		case _TURTLE_LINE_TRACER_MODE_GREEN_UNTIL_BLACK:
		case _TURTLE_LINE_TRACER_MODE_BLUE_UNTIL_BLACK:
		case _TURTLE_LINE_TRACER_MODE_ANY_UNTIL_BLACK:
			wait_until(callback, NULL);
			_robot_write(robot, TURTLE_LINE_TRACER_MODE, _TURTLE_LINE_TRACER_MODE_OFF);
			break;
	}
}

void _turtle_black_line(int turtle_index) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK, NULL);
}

void _turtle_red_line(int turtle_index) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_RED, NULL);
}

void _turtle_green_line(int turtle_index) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_GREEN, NULL);
}

void _turtle_blue_line(int turtle_index) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLUE, NULL);
}

void _turtle_any_line(int turtle_index) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_ANY, NULL);
}

void _turtle_black_line_until_red(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_RED, callback);
}

void _turtle_black_line_until_yellow(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_YELLOW, callback);
}

void _turtle_black_line_until_green(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_GREEN, callback);
}

void _turtle_black_line_until_sky_blue(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_CYAN, callback);
}

void _turtle_black_line_until_blue(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_BLUE, callback);
}

void _turtle_black_line_until_purple(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_MAGENTA, callback);
}

void _turtle_black_line_until_any(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLACK_UNTIL_ANY, callback);
}

void _turtle_red_line_until_black(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_RED_UNTIL_BLACK, callback);
}

void _turtle_green_line_until_black(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_GREEN_UNTIL_BLACK, callback);
}

void _turtle_blue_line_until_black(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_BLUE_UNTIL_BLACK, callback);
}

void _turtle_any_line_until_black(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_ANY_UNTIL_BLACK, callback);
}

void _turtle_cross_forward(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_CROSS, callback);
}

void _turtle_cross_left(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_TURN_LEFT, callback);
}

void _turtle_cross_right(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_TURN_RIGHT, callback);
}

void _turtle_cross_uturn(int turtle_index, int (*callback)(void* arg)) {
	_turtle_line_tracer_mode(turtle_index, _TURTLE_LINE_TRACER_MODE_UTURN, callback);
}

void _turtle_line_gain(int turtle_index, double gain) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_LINE_TRACER_GAIN, _round_double_value(gain));
}

void _turtle_line_speed(int turtle_index, double speed) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_LINE_TRACER_SPEED, _round_double_value(speed));
}

void _turtle_rgb(int turtle_index, double red, double green, double blue) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write_at(robot, TURTLE_LED, 0, _round_double_value(red));
	_robot_write_at(robot, TURTLE_LED, 1, _round_double_value(green));
	_robot_write_at(robot, TURTLE_LED, 2, _round_double_value(blue));
}

int _TURTLE_COLOR_TO_RGB[][3] = {
	{ 0, 0, 0 },
	{ 255, 0, 0 },
	{ 255, 63, 0 },
	{ 255, 255, 0 },
	{ 0, 255, 0 },
	{ 0, 255, 255 },
	{ 0, 0, 255 },
	{ 63, 0, 255 },
	{ 255, 0, 255 },
	{ 255, 255, 255 }
};

int* _turtle_get_rgb(const char* color) {
	if(color == NULL) return NULL;
	if(_STRICMP(color, TURTLE_LED_OFF) == 0) return _TURTLE_COLOR_TO_RGB[0];
	if(_STRICMP(color, TURTLE_LED_RED) == 0) return _TURTLE_COLOR_TO_RGB[1];
	if(_STRICMP(color, TURTLE_LED_ORANGE) == 0) return _TURTLE_COLOR_TO_RGB[2];
	if(_STRICMP(color, TURTLE_LED_YELLOW) == 0) return _TURTLE_COLOR_TO_RGB[3];
	if(_STRICMP(color, TURTLE_LED_GREEN) == 0) return _TURTLE_COLOR_TO_RGB[4];
	if(_STRICMP(color, TURTLE_LED_SKY_BLUE) == 0) return _TURTLE_COLOR_TO_RGB[5];
	if(_STRICMP(color, TURTLE_LED_BLUE) == 0) return _TURTLE_COLOR_TO_RGB[6];
	if(_STRICMP(color, TURTLE_LED_VIOLET) == 0) return _TURTLE_COLOR_TO_RGB[7];
	if(_STRICMP(color, TURTLE_LED_PURPLE) == 0) return _TURTLE_COLOR_TO_RGB[8];
	if(_STRICMP(color, TURTLE_LED_WHITE) == 0) return _TURTLE_COLOR_TO_RGB[9];
	if(_STRICMP(color, "SKY_BLUE") == 0) return _TURTLE_COLOR_TO_RGB[5];
	if(_STRICMP(color, "SKYBLUE") == 0) return _TURTLE_COLOR_TO_RGB[5];
	return NULL;
}

void _turtle_led(int turtle_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int* rgb;
	
	if(robot == NULL) return;
	rgb = _turtle_get_rgb(color);
	if(rgb != NULL) _robot_write_array(robot, TURTLE_LED, rgb, 3);
}

void _turtle_buzzer(int turtle_index, double hz) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_NOTE, 0);
	_turtle_run_sound(robot, 0, 1);
	_robot_write_float(robot, TURTLE_BUZZER, (float)hz);
}

void _turtle_tempo(int turtle_index, double bpm) {
	struct _turtle_robot* turtle = (struct _turtle_robot*)_robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(turtle == NULL) return;
	if(bpm > 0) turtle->bpm = bpm;
}

void _turtle_pitch(int turtle_index, const char* pitch) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int _pitch;
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_BUZZER, 0);
	_turtle_run_sound(robot, 0, 1);
	_pitch = _get_pitch_value(pitch);
	if(_pitch >= 0) _robot_write(robot, TURTLE_NOTE, _pitch);
}

void _turtle_note(int turtle_index, const char* pitch, double beats) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	struct _turtle_robot* turtle = (struct _turtle_robot*)robot;
	
	if(turtle == NULL) return;
	_robot_write(robot, TURTLE_BUZZER, 0);
	_turtle_run_sound(robot, 0, 1);
	if(beats > 0 && turtle->bpm > 0) {
		int _pitch = _get_pitch_value(pitch);
		if(_pitch == 0) {
			_robot_write(robot, TURTLE_NOTE, 0);
			wait((int)(beats * 60 * 1000 / turtle->bpm));
		} else if(_pitch > 0) {
			int timeout = (int)(beats * 60 * 1000 / turtle->bpm);
			int tail = 0;
			if(timeout > 100) {
				tail = 100;
			}
			_robot_write(robot, TURTLE_NOTE, _pitch);
			wait(timeout - tail);
			_robot_write(robot, TURTLE_NOTE, 0);
			if(tail > 0) {
				wait(tail);
			}
		}
	} else {
		_robot_write(robot, TURTLE_NOTE, 0);
	}
}

int _turtle_get_sound_value(const char* sound) {
	if(sound == NULL) return -1;
	if(_STRICMP(sound, TURTLE_SOUND_OFF) == 0) return _TURTLE_SOUND_OFF;
	if(_STRICMP(sound, TURTLE_SOUND_BEEP) == 0) return _TURTLE_SOUND_BEEP;
	if(_STRICMP(sound, TURTLE_SOUND_RANDOM_BEEP) == 0) return _TURTLE_SOUND_RANDOM_BEEP;
	if(_STRICMP(sound, TURTLE_SOUND_SIREN) == 0) return _TURTLE_SOUND_SIREN;
	if(_STRICMP(sound, TURTLE_SOUND_ENGINE) == 0) return _TURTLE_SOUND_ENGINE;
	if(_STRICMP(sound, TURTLE_SOUND_ROBOT) == 0) return _TURTLE_SOUND_ROBOT;
	if(_STRICMP(sound, TURTLE_SOUND_MARCH) == 0) return _TURTLE_SOUND_MARCH;
	if(_STRICMP(sound, TURTLE_SOUND_BIRTHDAY) == 0) return _TURTLE_SOUND_BIRTHDAY;
	if(_STRICMP(sound, TURTLE_SOUND_DIBIDIBIDIP) == 0) return _TURTLE_SOUND_DIBIDIBIDIP;
	if(_STRICMP(sound, TURTLE_SOUND_GOOD_JOB) == 0) return _TURTLE_SOUND_GOOD_JOB;
	if(_STRICMP(sound, "RANDOM_BEEP") == 0) return _TURTLE_SOUND_RANDOM_BEEP;
	if(_STRICMP(sound, "GOOD_JOB") == 0) return _TURTLE_SOUND_GOOD_JOB;
	return -1;
}

void _turtle_repeat_sound(int turtle_index, const char* sound, double repeat) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int _sound, _repeat;

	if(robot == NULL) return;
	_robot_write(robot, TURTLE_BUZZER, 0);
	_robot_write(robot, TURTLE_NOTE, 0);
	_sound = _turtle_get_sound_value(sound);
	_repeat = _round_double_value(repeat);
	if(_sound > 0 && _repeat != 0) {
		_turtle_run_sound(robot, _sound, _repeat);
	} else {
		_turtle_run_sound(robot, 0, 1);
	}
}

void _turtle_sound(int turtle_index, const char* sound) {
	_turtle_repeat_sound(turtle_index, sound, 1);
}

int _turtle_sound_callback(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return (_robot_e(robot, TURTLE_SOUND_STATE) == 1 && _robot_read(robot, TURTLE_SOUND_STATE) == 0) ? 1 : 0;
}

void _turtle_repeat_sound_until_done(int turtle_index, const char* sound, double repeat, int (*callback)(void* arg)) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int _sound, _repeat;

	if(robot == NULL) return;
	_robot_write(robot, TURTLE_BUZZER, 0);
	_robot_write(robot, TURTLE_NOTE, 0);
	_sound = _turtle_get_sound_value(sound);
	_repeat = _round_double_value(repeat);
	if(_sound > 0 && _repeat != 0) {
		_turtle_run_sound(robot, _sound, _repeat);
		wait_until(callback, NULL);
	} else {
		_turtle_run_sound(robot, 0, 1);
	}
}

void _turtle_sound_until_done(int turtle_index, const char* sound, int (*callback)(void* arg)) {
	_turtle_repeat_sound_until_done(turtle_index, sound, 1, callback);
}

void _turtle_lamp(int turtle_index, int on) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_LAMP, on);
}

void _turtle_lock(int turtle_index, int on) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return;
	_robot_write(robot, TURTLE_LOCK, on);
}

int _turtle_signal_strength(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_SIGNAL_STRENGTH);
}

int _turtle_floor(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_FLOOR);
}

int _turtle_acceleration_x(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, TURTLE_ACCELERATION, 0);
}

int _turtle_acceleration_y(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, TURTLE_ACCELERATION, 1);
}

int _turtle_acceleration_z(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read_at(robot, TURTLE_ACCELERATION, 2);
}

int _turtle_temperature(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_TEMPERATURE);
}

int _turtle_button(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_BUTTON);
}

int _turtle_clicked(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_e(robot, TURTLE_CLICKED);
}

int _turtle_double_clicked(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_e(robot, TURTLE_DOUBLE_CLICKED);
}

int _turtle_long_pressed(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_e(robot, TURTLE_LONG_PRESSED);
}

int _turtle_color_number(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return -1;
	return _robot_read(robot, TURTLE_COLOR_NUMBER);
}

int _turtle_get_color_number(const char* color) {
	if(color == NULL) return -1;
	if(_STRICMP(color, _TURTLE_COLOR_BLACK) == 0) return TURTLE_COLOR_BLACK;
	if(_STRICMP(color, _TURTLE_COLOR_RED) == 0) return TURTLE_COLOR_RED;
	if(_STRICMP(color, _TURTLE_COLOR_ORANGE) == 0) return TURTLE_COLOR_ORANGE;
	if(_STRICMP(color, _TURTLE_COLOR_YELLOW) == 0) return TURTLE_COLOR_YELLOW;
	if(_STRICMP(color, _TURTLE_COLOR_GREEN) == 0) return TURTLE_COLOR_GREEN;
	if(_STRICMP(color, _TURTLE_COLOR_SKY_BLUE) == 0) return TURTLE_COLOR_SKY_BLUE;
	if(_STRICMP(color, _TURTLE_COLOR_BLUE) == 0) return TURTLE_COLOR_BLUE;
	if(_STRICMP(color, _TURTLE_COLOR_PURPLE) == 0) return TURTLE_COLOR_PURPLE;
	if(_STRICMP(color, _TURTLE_COLOR_WHITE) == 0) return TURTLE_COLOR_WHITE;
	if(_STRICMP(color, _TURTLE_COLOR_CYAN) == 0) return TURTLE_COLOR_CYAN;
	if(_STRICMP(color, _TURTLE_COLOR_MAGENTA) == 0) return TURTLE_COLOR_MAGENTA;
	if(_STRICMP(color, "SKY_BLUE") == 0) return TURTLE_COLOR_SKY_BLUE;
	if(_STRICMP(color, "SKYBLUE") == 0) return TURTLE_COLOR_SKY_BLUE;
	return -1;
}

int _turtle_is_color(int turtle_index, const char* color) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int color_number;
	
	if(robot == NULL) return 0;
	color_number = _turtle_get_color_number(color);
	if(color_number >= 0) {
		return (color_number == _robot_read(robot, TURTLE_COLOR_NUMBER)) ? 1 : 0;
	}
	return 0;
}

int _turtle_color_pattern(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return -1;
	if(_robot_e(robot, TURTLE_COLOR_PATTERN) == 1) {
		return _robot_read(robot, TURTLE_COLOR_PATTERN);
	}
	return -1;
}

int _turtle_is_color_pattern(int turtle_index, const char* color1, const char* color2) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	int color_number1, color_number2;
	
	if(robot == NULL) return 0;
	if(_robot_e(robot, TURTLE_COLOR_PATTERN) == 1) {
		color_number1 = _turtle_get_color_number(color1);
		color_number2 = _turtle_get_color_number(color2);
		if(color_number1 >= 0 && color_number2 >= 0) {
			return (color_number1 * 10 + color_number2 == _robot_read(robot, TURTLE_COLOR_PATTERN)) ? 1 : 0;
		}
	}
	return 0;
}

int _turtle_pulse_count(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_PULSE_COUNT);
}

int _turtle_tilt(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_TILT);
}

int _turtle_battery_state(int turtle_index) {
	struct _robot* robot = _robot_group_get_robot(_GROUP_TURTLE, turtle_index);
	
	if(robot == NULL) return 0;
	return _robot_read(robot, TURTLE_BATTERY_STATE);
}

#define _FUNCTION_TURTLE(n) \
	__inline const char* _turtle_get_name_##n(void) { return _robot_group_get_name(_GROUP_TURTLE, n); } \
	__inline void _turtle_set_name_##n(const char* name) { _robot_group_set_name(_GROUP_TURTLE, n, name); } \
	__inline const char* _turtle_get_id_##n(void) { return TURTLE_ID; } \
	__inline int _turtle_get_index_##n(void) { return n; } \
	__inline int _turtle_e_##n(int device_id) { return _robot_group_e(_GROUP_TURTLE, n, device_id); } \
	__inline int _turtle_read_##n(int device_id) { return _robot_group_read(_GROUP_TURTLE, n, device_id); } \
	__inline int _turtle_read_at_##n(int device_id, int index) { return _robot_group_read_at(_GROUP_TURTLE, n, device_id, index); } \
	__inline int _turtle_read_array_##n(int device_id, int* data, int length) { return _robot_group_read_array(_GROUP_TURTLE, n, device_id, data, length); } \
	__inline float _turtle_read_float_##n(int device_id) { return _robot_group_read_float(_GROUP_TURTLE, n, device_id); } \
	__inline float _turtle_read_float_at_##n(int device_id, int index) { return _robot_group_read_float_at(_GROUP_TURTLE, n, device_id, index); } \
	__inline int _turtle_read_float_array_##n(int device_id, float* data, int length) { return _robot_group_read_float_array(_GROUP_TURTLE, n, device_id, data, length); } \
	__inline int _turtle_write_empty_##n(int device_id) { return _robot_group_write_empty(_GROUP_TURTLE, n, device_id); } \
	__inline int _turtle_write_##n(int device_id, int data) { return _robot_group_write(_GROUP_TURTLE, n, device_id, data); } \
	__inline int _turtle_write_at_##n(int device_id, int index, int data) { return _robot_group_write_at(_GROUP_TURTLE, n, device_id, index, data); } \
	__inline int _turtle_write_array_##n(int device_id, const int* data, int length) { return _robot_group_write_array(_GROUP_TURTLE, n, device_id, data, length); } \
	__inline int _turtle_write_float_##n(int device_id, float data) { return _robot_group_write_float(_GROUP_TURTLE, n, device_id, data); } \
	__inline int _turtle_write_float_at_##n(int device_id, int index, float data) { return _robot_group_write_float_at(_GROUP_TURTLE, n, device_id, index, data); } \
	__inline int _turtle_write_float_array_##n(int device_id, const float* data, int length) { return _robot_group_write_float_array(_GROUP_TURTLE, n, device_id, data, length); } \
	__inline void _turtle_reset_##n(void) { _robot_group_reset(_GROUP_TURTLE, n); } \
	__inline void _turtle_dispose_##n(void) { _robot_group_dispose(_GROUP_TURTLE, n); } \
	__inline void _turtle_wheels_##n(double left_velocity, double right_velocity) { _turtle_wheels(n, left_velocity, right_velocity); } \
	__inline void _turtle_left_wheel_##n(double velocity) { _turtle_left_wheel(n, velocity); } \
	__inline void _turtle_right_wheel_##n(double velocity) { _turtle_right_wheel(n, velocity); } \
	__inline void _turtle_stop_##n(void) { _turtle_stop(n); } \
	__inline int _turtle_wheel_callback_##n(void* arg) { return _turtle_wheel_callback(n); } \
	__inline void _turtle_move_forward_##n(double cm, double velocity) { _turtle_move_forward(n, cm, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_move_backward_##n(double cm, double velocity) { _turtle_move_backward(n, cm, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_turn_left_##n(double degree, double velocity) { _turtle_turn_left(n, degree, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_turn_right_##n(double degree, double velocity) { _turtle_turn_right(n, degree, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_pivot_left_##n(double degree, double velocity) { _turtle_pivot_left(n, degree, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_pivot_right_##n(double degree, double velocity) { _turtle_pivot_right(n, degree, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_circle_left_##n(double degree, double radius, double velocity) { _turtle_circle_left(n, degree, radius, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_circle_right_##n(double degree, double radius, double velocity) { _turtle_circle_right(n, degree, radius, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_move_forward_sec_##n(double sec, double velocity) { _turtle_move_forward_sec(n, sec, velocity); } \
	__inline void _turtle_move_backward_sec_##n(double sec, double velocity) { _turtle_move_backward_sec(n, sec, velocity); } \
	__inline void _turtle_turn_left_sec_##n(double sec, double velocity) { _turtle_turn_left_sec(n, sec, velocity); } \
	__inline void _turtle_turn_right_sec_##n(double sec, double velocity) { _turtle_turn_right_sec(n, sec, velocity); } \
	__inline void _turtle_pivot_left_sec_##n(double sec, double velocity) { _turtle_pivot_left_sec(n, sec, velocity); } \
	__inline void _turtle_pivot_right_sec_##n(double sec, double velocity) { _turtle_pivot_right_sec(n, sec, velocity); } \
	__inline void _turtle_circle_left_sec_##n(double sec, double radius, double velocity) { _turtle_circle_left_sec(n, sec, radius, velocity); } \
	__inline void _turtle_circle_right_sec_##n(double sec, double radius, double velocity) { _turtle_circle_right_sec(n, sec, radius, velocity); } \
	__inline void _turtle_move_forward_pulse_##n(double pulse, double velocity) { _turtle_move_forward_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_move_backward_pulse_##n(double pulse, double velocity) { _turtle_move_backward_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_turn_left_pulse_##n(double pulse, double velocity) { _turtle_turn_left_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_turn_right_pulse_##n(double pulse, double velocity) { _turtle_turn_right_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_pivot_left_pulse_##n(double pulse, double velocity) { _turtle_pivot_left_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_pivot_right_pulse_##n(double pulse, double velocity) { _turtle_pivot_right_pulse(n, pulse, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_circle_left_pulse_##n(double pulse, double radius, double velocity) { _turtle_circle_left_pulse(n, pulse, radius, velocity, _turtle_wheel_callback_##n); } \
	__inline void _turtle_circle_right_pulse_##n(double pulse, double radius, double velocity) { _turtle_circle_right_pulse(n, pulse, radius, velocity, _turtle_wheel_callback_##n); } \
	__inline int _turtle_line_tracer_mode_callback_##n(void* arg) { return _turtle_line_tracer_mode_callback(n); } \
	__inline void _turtle_line_tracer_mode_##n(int mode) { _turtle_line_tracer_mode(n, mode, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_##n(void) { _turtle_black_line(n); } \
	__inline void _turtle_red_line_##n(void) { _turtle_red_line(n); } \
	__inline void _turtle_green_line_##n(void) { _turtle_green_line(n); } \
	__inline void _turtle_blue_line_##n(void) { _turtle_blue_line(n); } \
	__inline void _turtle_any_line_##n(void) { _turtle_any_line(n); } \
	__inline void _turtle_black_line_until_red_##n(void) { _turtle_black_line_until_red(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_yellow_##n(void) { _turtle_black_line_until_yellow(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_green_##n(void) { _turtle_black_line_until_green(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_sky_blue_##n(void) { _turtle_black_line_until_sky_blue(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_blue_##n(void) { _turtle_black_line_until_blue(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_purple_##n(void) { _turtle_black_line_until_purple(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_black_line_until_any_##n(void) { _turtle_black_line_until_any(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_red_line_until_black_##n(void) { _turtle_red_line_until_black(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_green_line_until_black_##n(void) { _turtle_green_line_until_black(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_blue_line_until_black_##n(void) { _turtle_blue_line_until_black(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_any_line_until_black_##n(void) { _turtle_any_line_until_black(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_cross_forward_##n(void) { _turtle_cross_forward(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_cross_left_##n(void) { _turtle_cross_left(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_cross_right_##n(void) { _turtle_cross_right(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_cross_uturn_##n(void) { _turtle_cross_uturn(n, _turtle_line_tracer_mode_callback_##n); } \
	__inline void _turtle_line_gain_##n(double gain) { _turtle_line_gain(n, gain); } \
	__inline void _turtle_line_speed_##n(double speed) { _turtle_line_speed(n, speed); } \
	__inline void _turtle_rgb_##n(double red, double green, double blue) { _turtle_rgb(n, red, green, blue); } \
	__inline void _turtle_led_##n(const char* color) { _turtle_led(n, color); } \
	__inline void _turtle_buzzer_##n(double hz) { _turtle_buzzer(n, hz); } \
	__inline void _turtle_tempo_##n(double bpm) { _turtle_tempo(n, bpm); } \
	__inline void _turtle_pitch_##n(const char* pitch) { _turtle_pitch(n, pitch); } \
	__inline void _turtle_note_##n(const char* pitch, double beats) { _turtle_note(n, pitch, beats); } \
	__inline void _turtle_repeat_sound_##n(const char* sound, double repeat) { _turtle_repeat_sound(n, sound, repeat); } \
	__inline void _turtle_sound_##n(const char* sound) { _turtle_sound(n, sound); } \
	__inline int _turtle_sound_callback_##n(void* arg) { return _turtle_sound_callback(n); } \
	__inline void _turtle_repeat_sound_until_done_##n(const char* sound, double repeat) { _turtle_repeat_sound_until_done(n, sound, repeat, _turtle_sound_callback_##n); } \
	__inline void _turtle_sound_until_done_##n(const char* sound) { _turtle_sound_until_done(n, sound, _turtle_sound_callback_##n); } \
	__inline void _turtle_lamp_##n(int on) { _turtle_lamp(n, on); } \
	__inline void _turtle_lock_##n(int on) { _turtle_lock(n, on); } \
	__inline int _turtle_signal_strength_##n(void) { return _turtle_signal_strength(n); } \
	__inline int _turtle_floor_##n(void) { return _turtle_floor(n); } \
	__inline int _turtle_acceleration_x_##n(void) { return _turtle_acceleration_x(n); } \
	__inline int _turtle_acceleration_y_##n(void) { return _turtle_acceleration_y(n); } \
	__inline int _turtle_acceleration_z_##n(void) { return _turtle_acceleration_z(n); } \
	__inline int _turtle_temperature_##n(void) { return _turtle_temperature(n); } \
	__inline int _turtle_button_##n(void) { return _turtle_button(n); } \
	__inline int _turtle_clicked_##n(void) { return _turtle_clicked(n); } \
	__inline int _turtle_double_clicked_##n(void) { return _turtle_double_clicked(n); } \
	__inline int _turtle_long_pressed_##n(void) { return _turtle_long_pressed(n); } \
	__inline int _turtle_color_number_##n(void) { return _turtle_color_number(n); } \
	__inline int _turtle_is_color_##n(const char* color) { return _turtle_is_color(n, color); } \
	__inline int _turtle_color_pattern_##n(void) { return _turtle_color_pattern(n); } \
	__inline int _turtle_is_color_pattern_##n(const char* color1, const char* color2) { return _turtle_is_color_pattern(n, color1, color2); } \
	__inline int _turtle_pulse_count_##n(void) { return _turtle_pulse_count(n); } \
	__inline int _turtle_tilt_##n(void) { return _turtle_tilt(n); } \
	__inline int _turtle_battery_state_##n(void) { return _turtle_battery_state(n); }

#define _FUNCTION_PTR_TURTLE(name, n) \
	name->get_name = _turtle_get_name_##n; \
	name->set_name = _turtle_set_name_##n; \
	name->get_id = _turtle_get_id_##n; \
	name->get_index = _turtle_get_index_##n; \
	name->e = _turtle_e_##n; \
	name->read = _turtle_read_##n; \
	name->read_at = _turtle_read_at_##n; \
	name->read_array = _turtle_read_array_##n; \
	name->read_float = _turtle_read_float_##n; \
	name->read_float_at = _turtle_read_float_at_##n; \
	name->read_float_array = _turtle_read_float_array_##n; \
	name->write_empty = _turtle_write_empty_##n; \
	name->write = _turtle_write_##n; \
	name->write_at = _turtle_write_at_##n; \
	name->write_array = _turtle_write_array_##n; \
	name->write_float = _turtle_write_float_##n; \
	name->write_float_at = _turtle_write_float_at_##n; \
	name->write_float_array = _turtle_write_float_array_##n; \
	name->reset = _turtle_reset_##n; \
	name->dispose = _turtle_dispose_##n; \
	name->wheels = _turtle_wheels_##n; \
	name->left_wheel = _turtle_left_wheel_##n; \
	name->right_wheel = _turtle_right_wheel_##n; \
	name->stop = _turtle_stop_##n; \
	name->move_forward = _turtle_move_forward_##n; \
	name->move_backward = _turtle_move_backward_##n; \
	name->turn_left = _turtle_turn_left_##n; \
	name->turn_right = _turtle_turn_right_##n; \
	name->pivot_left = _turtle_pivot_left_##n; \
	name->pivot_right = _turtle_pivot_right_##n; \
	name->circle_left = _turtle_circle_left_##n; \
	name->circle_right = _turtle_circle_right_##n; \
	name->move_forward_sec = _turtle_move_forward_sec_##n; \
	name->move_backward_sec = _turtle_move_backward_sec_##n; \
	name->turn_left_sec = _turtle_turn_left_sec_##n; \
	name->turn_right_sec = _turtle_turn_right_sec_##n; \
	name->pivot_left_sec = _turtle_pivot_left_sec_##n; \
	name->pivot_right_sec = _turtle_pivot_right_sec_##n; \
	name->circle_left_sec = _turtle_circle_left_sec_##n; \
	name->circle_right_sec = _turtle_circle_right_sec_##n; \
	name->move_forward_pulse = _turtle_move_forward_pulse_##n; \
	name->move_backward_pulse = _turtle_move_backward_pulse_##n; \
	name->turn_left_pulse = _turtle_turn_left_pulse_##n; \
	name->turn_right_pulse = _turtle_turn_right_pulse_##n; \
	name->pivot_left_pulse = _turtle_pivot_left_pulse_##n; \
	name->pivot_right_pulse = _turtle_pivot_right_pulse_##n; \
	name->circle_left_pulse = _turtle_circle_left_pulse_##n; \
	name->circle_right_pulse = _turtle_circle_right_pulse_##n; \
	name->line_tracer_mode = _turtle_line_tracer_mode_##n; \
	name->black_line = _turtle_black_line_##n; \
	name->red_line = _turtle_red_line_##n; \
	name->green_line = _turtle_green_line_##n; \
	name->blue_line = _turtle_blue_line_##n; \
	name->any_line = _turtle_any_line_##n; \
	name->black_line_until_red = _turtle_black_line_until_red_##n; \
	name->black_line_until_yellow = _turtle_black_line_until_yellow_##n; \
	name->black_line_until_green = _turtle_black_line_until_green_##n; \
	name->black_line_until_sky_blue = _turtle_black_line_until_sky_blue_##n; \
	name->black_line_until_blue = _turtle_black_line_until_blue_##n; \
	name->black_line_until_purple = _turtle_black_line_until_purple_##n; \
	name->black_line_until_any = _turtle_black_line_until_any_##n; \
	name->red_line_until_black = _turtle_red_line_until_black_##n; \
	name->green_line_until_black = _turtle_green_line_until_black_##n; \
	name->blue_line_until_black = _turtle_blue_line_until_black_##n; \
	name->any_line_until_black = _turtle_any_line_until_black_##n; \
	name->cross_forward = _turtle_cross_forward_##n; \
	name->cross_left = _turtle_cross_left_##n; \
	name->cross_right = _turtle_cross_right_##n; \
	name->cross_uturn = _turtle_cross_uturn_##n; \
	name->line_gain = _turtle_line_gain_##n; \
	name->line_speed = _turtle_line_speed_##n; \
	name->rgb = _turtle_rgb_##n; \
	name->led = _turtle_led_##n; \
	name->buzzer = _turtle_buzzer_##n; \
	name->tempo = _turtle_tempo_##n; \
	name->pitch = _turtle_pitch_##n; \
	name->note = _turtle_note_##n; \
	name->sound = _turtle_sound_##n; \
	name->sound_until_done = _turtle_sound_until_done_##n; \
	name->repeat_sound = _turtle_repeat_sound_##n; \
	name->repeat_sound_until_done = _turtle_repeat_sound_until_done_##n; \
	name->lamp = _turtle_lamp_##n; \
	name->lock = _turtle_lock_##n; \
	name->signal_strength = _turtle_signal_strength_##n; \
	name->floor = _turtle_floor_##n; \
	name->acceleration_x = _turtle_acceleration_x_##n; \
	name->acceleration_y = _turtle_acceleration_y_##n; \
	name->acceleration_z = _turtle_acceleration_z_##n; \
	name->temperature = _turtle_temperature_##n; \
	name->button = _turtle_button_##n; \
	name->clicked = _turtle_clicked_##n; \
	name->double_clicked = _turtle_double_clicked_##n; \
	name->long_pressed = _turtle_long_pressed_##n; \
	name->color_number = _turtle_color_number_##n; \
	name->is_color = _turtle_is_color_##n; \
	name->color_pattern = _turtle_color_pattern_##n; \
	name->is_color_pattern = _turtle_is_color_pattern_##n; \
	name->pulse_count = _turtle_pulse_count_##n; \
	name->tilt = _turtle_tilt_##n; \
	name->battery_state = _turtle_battery_state_##n;

_FUNCTION_TURTLE(0);
_FUNCTION_TURTLE(1);
_FUNCTION_TURTLE(2);
_FUNCTION_TURTLE(3);
_FUNCTION_TURTLE(4);
_FUNCTION_TURTLE(5);
_FUNCTION_TURTLE(6);
_FUNCTION_TURTLE(7);
_FUNCTION_TURTLE(8);
_FUNCTION_TURTLE(9);

Turtle* turtle_create_port(const char* port_name) {
	if(_robot_group_count_robots(_GROUP_TURTLE) < _MAX_NUM_ROBOTS) {
		struct turtle* turtle = (struct turtle*)malloc(sizeof(struct turtle));
		struct _robot* robot = (struct _robot*)_turtle_create(port_name);
	
		switch(robot->index) {
			case 0: _FUNCTION_PTR_TURTLE(turtle, 0); break;
			case 1: _FUNCTION_PTR_TURTLE(turtle, 1); break;
			case 2: _FUNCTION_PTR_TURTLE(turtle, 2); break;
			case 3: _FUNCTION_PTR_TURTLE(turtle, 3); break;
			case 4: _FUNCTION_PTR_TURTLE(turtle, 4); break;
			case 5: _FUNCTION_PTR_TURTLE(turtle, 5); break;
			case 6: _FUNCTION_PTR_TURTLE(turtle, 6); break;
			case 7: _FUNCTION_PTR_TURTLE(turtle, 7); break;
			case 8: _FUNCTION_PTR_TURTLE(turtle, 8); break;
			case 9: _FUNCTION_PTR_TURTLE(turtle, 9); break;
		}
		return turtle;
	}
	return NULL;
}

Turtle* turtle_create(void) {
	return turtle_create_port(NULL);
}

const char* turtle_get_name(void) {
	return _robot_group_get_name(_GROUP_TURTLE, 0);
}

void turtle_set_name(const char* name) {
	_robot_group_set_name(_GROUP_TURTLE, 0, name);
}

const char* turtle_get_id(void) {
	return TURTLE_ID;
}

int turtle_e(int device_id) {
	return _robot_group_e(_GROUP_TURTLE, 0, device_id);
}

int turtle_read(int device_id) {
	return _robot_group_read(_GROUP_TURTLE, 0, device_id);
}

int turtle_read_at(int device_id, int index) {
	return _robot_group_read_at(_GROUP_TURTLE, 0, device_id, index);
}

int turtle_read_array(int device_id, int* data, int length) {
	return _robot_group_read_array(_GROUP_TURTLE, 0, device_id, data, length);
}

float turtle_read_float(int device_id) {
	return _robot_group_read_float(_GROUP_TURTLE, 0, device_id);
}

float turtle_read_float_at(int device_id, int index) {
	return _robot_group_read_float_at(_GROUP_TURTLE, 0, device_id, index);
}

int turtle_read_float_array(int device_id, float* data, int length) {
	return _robot_group_read_float_array(_GROUP_TURTLE, 0, device_id, data, length);
}

int turtle_write_empty(int device_id) {
	return _robot_group_write_empty(_GROUP_TURTLE, 0, device_id);
}

int turtle_write(int device_id, int data) {
	return _robot_group_write(_GROUP_TURTLE, 0, device_id, data);
}

int turtle_write_at(int device_id, int index, int data) {
	return _robot_group_write_at(_GROUP_TURTLE, 0, device_id, index, data);
}

int turtle_write_array(int device_id, const int* data, int length) {
	return _robot_group_write_array(_GROUP_TURTLE, 0, device_id, data, length);
}

int turtle_write_float(int device_id, float data) {
	return _robot_group_write_float(_GROUP_TURTLE, 0, device_id, data);
}

int turtle_write_float_at(int device_id, int index, float data) {
	return _robot_group_write_float_at(_GROUP_TURTLE, 0, device_id, index, data);
}

int turtle_write_float_array(int device_id, const float* data, int length) {
	return _robot_group_write_float_array(_GROUP_TURTLE, 0, device_id, data, length);
}

void turtle_reset(void) {
	_robot_group_reset(_GROUP_TURTLE, 0);
}

void turtle_dispose(void) {
	_robot_group_dispose(_GROUP_TURTLE, 0);
}

void turtle_wheels(double left_velocity, double right_velocity) {
	_turtle_wheels(0, left_velocity, right_velocity);
}

void turtle_left_wheel(double velocity) {
	_turtle_left_wheel(0, velocity);
}

void turtle_right_wheel(double velocity) {
	_turtle_right_wheel(0, velocity);
}

void turtle_stop(void) {
	_turtle_stop(0);
}

int _turtle_wheel_evaluate(void* arg) {
	return _turtle_wheel_callback(0);
}

void turtle_move_forward(double cm, double velocity) {
	_turtle_move_forward(0, cm, velocity, _turtle_wheel_evaluate);
}

void turtle_move_backward(double cm, double velocity) {
	_turtle_move_backward(0, cm, velocity, _turtle_wheel_evaluate);
}

void turtle_turn_left(double degree, double velocity) {
	_turtle_turn_left(0, degree, velocity, _turtle_wheel_evaluate);
}

void turtle_turn_right(double degree, double velocity) {
	_turtle_turn_right(0, degree, velocity, _turtle_wheel_evaluate);
}

void turtle_pivot_left(double degree, double velocity) {
	_turtle_pivot_left(0, degree, velocity, _turtle_wheel_evaluate);
}

void turtle_pivot_right(double degree, double velocity) {
	_turtle_pivot_right(0, degree, velocity, _turtle_wheel_evaluate);
}

void turtle_circle_left(double degree, double radius, double velocity) {
	_turtle_circle_left(0, degree, radius, velocity, _turtle_wheel_evaluate);
}

void turtle_circle_right(double degree, double radius, double velocity) {
	_turtle_circle_right(0, degree, radius, velocity, _turtle_wheel_evaluate);
}

void turtle_move_forward_sec(double sec, double velocity) {
	_turtle_move_forward_sec(0, sec, velocity);
}

void turtle_move_backward_sec(double sec, double velocity) {
	_turtle_move_backward_sec(0, sec, velocity);
}

void turtle_turn_left_sec(double sec, double velocity) {
	_turtle_turn_left_sec(0, sec, velocity);
}

void turtle_turn_right_sec(double sec, double velocity) {
	_turtle_turn_right_sec(0, sec, velocity);
}

void turtle_pivot_left_sec(double sec, double velocity) {
	_turtle_pivot_left_sec(0, sec, velocity);
}

void turtle_pivot_right_sec(double sec, double velocity) {
	_turtle_pivot_right_sec(0, sec, velocity);
}

void turtle_circle_left_sec(double sec, double radius, double velocity) {
	_turtle_circle_left_sec(0, sec, radius, velocity);
}

void turtle_circle_right_sec(double sec, double radius, double velocity) {
	_turtle_circle_right_sec(0, sec, radius, velocity);
}

void turtle_move_forward_pulse(double pulse, double velocity) {
	_turtle_move_forward_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_move_backward_pulse(double pulse, double velocity) {
	_turtle_move_backward_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_turn_left_pulse(double pulse, double velocity) {
	_turtle_turn_left_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_turn_right_pulse(double pulse, double velocity) {
	_turtle_turn_right_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_pivot_left_pulse(double pulse, double velocity) {
	_turtle_pivot_left_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_pivot_right_pulse(double pulse, double velocity) {
	_turtle_pivot_right_pulse(0, pulse, velocity, _turtle_wheel_evaluate);
}

void turtle_circle_left_pulse(double pulse, double radius, double velocity) {
	_turtle_circle_left_pulse(0, pulse, radius, velocity, _turtle_wheel_evaluate);
}

void turtle_circle_right_pulse(double pulse, double radius, double velocity) {
	_turtle_circle_right_pulse(0, pulse, radius, velocity, _turtle_wheel_evaluate);
}

int _turtle_line_tracer_mode_evaluate(void* arg) {
	return _turtle_line_tracer_mode_callback(0);
}

void turtle_line_tracer_mode(int mode) {
	_turtle_line_tracer_mode(0, mode, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line(void) {
	_turtle_black_line(0);
}

void turtle_red_line(void) {
	_turtle_red_line(0);
}

void turtle_green_line(void) {
	_turtle_green_line(0);
}

void turtle_blue_line(void) {
	_turtle_blue_line(0);
}

void turtle_any_line(void) {
	_turtle_any_line(0);
}

void turtle_black_line_until_red(void) {
	_turtle_black_line_until_red(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_yellow(void) {
	_turtle_black_line_until_yellow(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_green(void) {
	_turtle_black_line_until_green(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_sky_blue(void) {
	_turtle_black_line_until_sky_blue(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_blue(void) {
	_turtle_black_line_until_blue(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_purple(void) {
	_turtle_black_line_until_purple(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_black_line_until_any(void) {
	_turtle_black_line_until_any(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_red_line_until_black(void) {
	_turtle_red_line_until_black(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_green_line_until_black(void) {
	_turtle_green_line_until_black(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_blue_line_until_black(void) {
	_turtle_blue_line_until_black(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_any_line_until_black(void) {
	_turtle_any_line_until_black(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_cross_forward(void) {
	_turtle_cross_forward(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_cross_left(void) {
	_turtle_cross_left(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_cross_right(void) {
	_turtle_cross_right(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_cross_uturn(void) {
	_turtle_cross_uturn(0, _turtle_line_tracer_mode_evaluate);
}

void turtle_line_gain(double gain) {
	_turtle_line_gain(0, gain);
}

void turtle_line_speed(double speed) {
	_turtle_line_speed(0, speed);
}

void turtle_rgb(double red, double green, double blue) {
	_turtle_rgb(0, red, green, blue);
}

void turtle_led(const char* color) {
	_turtle_led(0, color);
}

void turtle_buzzer(double hz) {
	_turtle_buzzer(0, hz);
}

void turtle_tempo(double bpm) {
	_turtle_tempo(0, bpm);
}

void turtle_pitch(const char* pitch) {
	_turtle_pitch(0, pitch);
}

void turtle_note(const char* pitch, double beats) {
	_turtle_note(0, pitch, beats);
}

int _turtle_sound_evaluate(void* arg) {
	return _turtle_sound_callback(0);
}

void turtle_sound(const char* sound) {
	_turtle_sound(0, sound);
}

void turtle_sound_until_done(const char* sound) {
	_turtle_sound_until_done(0, sound, _turtle_sound_evaluate);
}

void turtle_repeat_sound(const char* sound, double repeat) {
	_turtle_repeat_sound(0, sound, repeat);
}

void turtle_repeat_sound_until_done(const char* sound, double repeat) {
	_turtle_repeat_sound_until_done(0, sound, repeat, _turtle_sound_evaluate);
}

void turtle_lamp(int on) {
	_turtle_lamp(0, on);
}

void turtle_lock(int on) {
	_turtle_lock(0, on);
}

int turtle_signal_strength(void) {
	return _turtle_signal_strength(0);
}

int turtle_floor(void) {
	return _turtle_floor(0);
}

int turtle_acceleration_x(void) {
	return _turtle_acceleration_x(0);
}

int turtle_acceleration_y(void) {
	return _turtle_acceleration_y(0);
}

int turtle_acceleration_z(void) {
	return _turtle_acceleration_z(0);
}

int turtle_temperature(void) {
	return _turtle_temperature(0);
}

int turtle_button(void) {
	return _turtle_button(0);
}

int turtle_clicked(void) {
	return _turtle_clicked(0);
}

int turtle_double_clicked(void) {
	return _turtle_double_clicked(0);
}

int turtle_long_pressed(void) {
	return _turtle_long_pressed(0);
}

int turtle_color_number(void) {
	return _turtle_color_number(0);
}

int turtle_is_color(const char* color) {
	return _turtle_is_color(0, color);
}

int turtle_color_pattern(void) {
	return _turtle_color_pattern(0);
}

int turtle_is_color_pattern(const char* color1, const char* color2) {
	return _turtle_is_color_pattern(0, color1, color2);
}

int turtle_pulse_count(void) {
	return _turtle_pulse_count(0);
}

int turtle_tilt(void) {
	return _turtle_tilt(0);
}

int turtle_battery_state(void) {
	return _turtle_battery_state(0);
}
